/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


package edu.cmu.taghelpertools.gui;

import edu.cmu.taghelpertools.middle_layer.Driver;
import edu.cmu.taghelpertools.middle_layer.InputFileReader;
import edu.cmu.taghelpertools.middle_layer.OptionsModel;
import gnu.trove.THashMap;
import gnu.trove.THashSet;

import java.awt.Dimension;
import java.awt.Font;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.File;
import java.util.Iterator;

import javax.swing.DefaultListModel;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JOptionPane;
import javax.swing.JScrollPane;
import javax.swing.JTextField;
import javax.swing.ListSelectionModel;
import javax.swing.UIManager;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;


import jxl.Cell;
import jxl.Sheet;
import jxl.Workbook;

/**
 * GUI component that is first displayed to the user. The user interacts with
 * the FileSelector to select all the files they wish to process and to set all
 * file-specific feature selection and re-training parameters.
 * 
 * @author Jaime Arguello
 * 
 */
public class FileSelector extends JFrame implements ActionListener, MouseListener {


    public final static String FILE_EXT = "xls";
    private JList fileList;
    private DefaultListModel fileListModel;
    private JButton selectFileButton;
    private JButton deleteFileButton;
    private JButton goButton;
    private JButton optionsButton;
    private JTextField outputPrefix;
    private JLabel outputPrefixLabel;
    //added by yichia
    private JCheckBox uploadCheckBox;
    
    
    
    private THashSet fileSet;
    private THashMap dimMap;
    private Driver driver;

    /**
     * Set the look and feel of the GUI to the OP specific look and feel,
     * instead of the Java default look and feel
     */
    private void setLookAndFeel() {
        try {
            UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
        } catch (Exception e) {
            try {
                UIManager.setLookAndFeel(UIManager.getCrossPlatformLookAndFeelClassName());
            } catch (Exception e1) {
                e1.printStackTrace();
                System.exit(1);
            }
        }
    }

    /**
     * Creates the GUI components and creates the GUI, with the help of
     * FileSelectorLayout, a customized layout manager, created with
     * "JavaGUIBuilder".
     * 
     * @param driver -
     *            The Driver (the main executioner of the program) is passed as
     *            parameter because several calls are made from this method to
     *            modify the "state" of the Driver.
     */
    public FileSelector(final Driver driver) {

        this.driver = driver;

        // This is used to make sure the user doesn't select the same file
        // twice. no file duplicates are allowed.
        fileSet = new THashSet();

        // dimMap = new THashMap();

        // sets the look and feel of the GUI.
        this.setLookAndFeel();

        this.setTitle("Create File List");

        FileSelectorLayout customLayout = new FileSelectorLayout();

        getContentPane().setFont(new Font("Arial", Font.PLAIN, 12));
        getContentPane().setLayout(customLayout);

        // the rest of this method simply creates the GUI components and adds
        // them to the GUI.

        if(this.driver.getBackupFileListModel() == null)
        {
        	fileListModel = new DefaultListModel();
        	this.driver.setBackupFileListModel(fileListModel);
        }
        else
        {
        	fileListModel = this.driver.getBackupFileListModel();
        }
        
        fileList = new JList(fileListModel);
        fileList.addMouseListener(this);
        fileList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
        JScrollPane fileListScrollPane = new JScrollPane(fileList);
        getContentPane().add(fileListScrollPane);

        fileList.addListSelectionListener(new ListSelectionListener() {

            public void valueChanged(ListSelectionEvent e) {
                optionsButton.setEnabled(false);
                for (int i = fileListModel.size(); i >= 0; i--) {
                    if (fileList.isSelectedIndex(i)) {
                        optionsButton.setEnabled(true);
                        break;
                    }
                }
            }
        });

                
        goButton = new JButton("GO!");
        goButton.addActionListener(this);
        getContentPane().add(goButton);
//        goButton.setEnabled(false);
        if (this.fileListModel.size() > 0) {
            goButton.setEnabled(true);
        } else {
            goButton.setEnabled(false);
        }
        

        selectFileButton = new JButton("Add a File");
        selectFileButton.addActionListener(this);
        getContentPane().add(selectFileButton);

        deleteFileButton = new JButton("Remove a File");
        deleteFileButton.addActionListener(this);
        getContentPane().add(deleteFileButton);

        optionsButton = new JButton(">> options");
        optionsButton.addActionListener(this);
        getContentPane().add(optionsButton);
        optionsButton.setEnabled(false);



        
        
        
        // should we provide the functionality to provide an output filename
        // prefix???
        outputPrefixLabel = new JLabel("Output file prefix:");
        getContentPane().add(outputPrefixLabel);
        
        outputPrefix = new JTextField(this.driver.getOutputPrefix());      
        getContentPane().add(outputPrefix);
        
        //TODO apply the prefix in the generation of exporting CSV and ARFF files
        
//yichia
        uploadCheckBox = new JCheckBox("upload");
        uploadCheckBox.addActionListener(this);
        uploadCheckBox.setEnabled(true);
        getContentPane().add(uploadCheckBox);
        
        
        this.pack();

        this.setResizable(false);

        Dimension screen = Toolkit.getDefaultToolkit().getScreenSize();
        int width = this.getWidth();
        int height = this.getHeight();
        int left = (screen.width - width) / 2;
        int top = (screen.height - height) / 2;
        setBounds(left, top, width, height);

        this.setVisible(true);

        addWindowListener(new WindowAdapter() {
            public void windowClosing(WindowEvent e) {
                //System.exit(0);
            	e.getWindow().dispose();
            }
        });
    } //End of constructor

    /*
     * Catches all events in the FileSelector. (non-Javadoc)
     * 
     * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
     */
    public void actionPerformed(ActionEvent e) {
        if (e.getSource().equals(this.selectFileButton)) {
            
            File files [] = this.getFile();
            
            if (files == null) {
                return;
            }
            
            for (int i = 0; i < files.length; i++) {
                File file = files[i];
                
                if (file == null) {
                    return;
                }
                
                if (fileSet.contains(file.getName())) {
                    JOptionPane.showMessageDialog(this,
                            "There is already a file added with name: "
                                    + file.getName(), "Invalid Action",
                            JOptionPane.ERROR_MESSAGE);
                    return;
                } 
                else {
                    fileSet.add(file.getName());
                    OptionsModel optionsModel = new OptionsModel();
                    optionsModel.setArffFileName(file.getName() + ".arff");
                    this.driver.addFileOptionModelMap(file.getAbsolutePath(),
                            						  optionsModel);
                    fileListModel.addElement(file.getAbsolutePath());
                    fileList.repaint();
                }
            }                
            
            if (this.fileListModel.size() > 0) {
                goButton.setEnabled(true);
            } else {
                goButton.setEnabled(false);
            }

        } else if (e.getSource().equals(this.deleteFileButton)) {
            for (int i = this.fileListModel.size(); i >= 0; i--) {
                if (fileList.isSelectedIndex(i)) {
                    File file = new File(fileListModel.getElementAt(i).toString());
                    String filename = file.getName();
                    fileSet.remove(filename);
                    fileListModel.remove(i);
                    this.driver.removeFileOptionModelMap(file.getAbsolutePath());
                    fileList.repaint();

//                    Iterator keyIter = this.dimMap.keySet().iterator();
//                    while (keyIter.hasNext()) {
//                        String key = keyIter.next().toString();
//                        String value = this.dimMap.get(key).toString();
//                        if (value.equals(file.getAbsolutePath())) {
//                            keyIter.remove();
//                        }
//                    }
                }
            }
            if (this.fileListModel.size() > 0) {
                goButton.setEnabled(true);
            } else {
                goButton.setEnabled(false);
            }
        } else if (e.getSource().equals(this.goButton)) {
            this.dispose();
            this.driver.processAllFiles(this.outputPrefix.getText());
        } else if (e.getSource().equals(this.optionsButton)) {
            this.showOptionsFrame();
        } else if (e.getSource().equals(uploadCheckBox)) {
        	boolean enable = uploadCheckBox.isSelected();
        		this.driver.setUploadMode(enable);
        	System.out.println("upload? "+enable);
        }
        
    }

    /**
     * Opens file chooser to help the user add a file to the inventory of files
     * to be processed. The FileChooser is configuered (via ExampleFileFilter)
     * to display only files that end with an .xls extesion).
     * 
     * @return selected file
     */
    public File [] getFile() {

        File inputFile = null;
        File [] inputFiles;
        ExampleFileFilter filter = new ExampleFileFilter();
        filter.addExtension(FileSelector.FILE_EXT);
        filter.addExtension("csv");
        JFileChooser fc = new JFileChooser();
        fc.setFileFilter(filter);
        fc.setFileSelectionMode(JFileChooser.FILES_ONLY);
        fc.setDialogTitle("Choose " + FileSelector.FILE_EXT + " file to process");
        fc.setCurrentDirectory(new File(System.getProperty("user.dir")));
         //TODO figure out what's going on here        
        fc.setMultiSelectionEnabled(true);
        
        int returnVal = fc.showOpenDialog(null);
        if (returnVal != JFileChooser.APPROVE_OPTION) {
            return null;
        } else {
            inputFiles = fc.getSelectedFiles();
            inputFile = fc.getSelectedFile();
            if (!inputFile.isFile()) {
                return null;
            }
        }

        // if (validateFile(inputFile)) {
            return inputFiles;
        // } else {
        //    return null;
        // }
    }

    public void showOptionsFrame() {
        for (int i = fileListModel.size(); i >= 0; i--) {
            if (fileList.isSelectedIndex(i)) {
                String path = fileListModel.getElementAt(i).toString();
                File file = new File(path);
                OptionsModel optionsModel = driver.getOptionsModelForFile(path);
                String title = "Options: " + file.getName();
                OptionsFrame optionsFrame = new OptionsFrame(title, optionsModel, file);
                optionsFrame.display(true);
                break;
            }
        }
    }

    /**
     * Validates the file: makes sure that it was not added previously and makes
     * sure that the dimensions in it do not clash with any of the dimensions
     * within the files previously added. In other words, all dimension names
     * must be different among all files added.
     * 
     * @param inputFile
     * @return - true if the file is valid, false otherwise.
     */
    public boolean validateFile(File inputFile) {
        try {
            Sheet sheet = Workbook.getWorkbook(inputFile).getSheet(0);
            for (int i = 0; i < sheet.getColumns(); i++) {
                Cell cell = sheet.getCell(i, 0);
                String content = cell.getContents();
                content = content.trim().toLowerCase();
                if (!content.equals(InputFileReader.TEXT)) {
                    if (!this.dimMap.contains(content)) {
                        dimMap.put(content, inputFile.getAbsolutePath());
                        return true;
                    } else {
                        JOptionPane.showMessageDialog(this, "Duplicate Dimension Name: " + content + ". All dimensions (even in separate files) must have different names!", "INVALID ACTION", JOptionPane.INFORMATION_MESSAGE);
                        return false;
                    }
                } else {
                    return true;
                }
            }
            return false;
        } catch (Exception e) {
            JOptionPane.showMessageDialog(this, "Invalid Input File: " + inputFile.getName() + "." + e.getMessage(), "TERMINAL ERROR", JOptionPane.ERROR_MESSAGE);
            return false;
        }
    }

    public void mouseClicked(MouseEvent e) {
        if (e.getClickCount() >= 2) {
            this.showOptionsFrame();
        }
    }

    public void mousePressed(MouseEvent e) {
        // TODO Auto-generated method stub

    }

    public void mouseReleased(MouseEvent e) {
        // TODO Auto-generated method stub

    }

    public void mouseEntered(MouseEvent e) {
        // TODO Auto-generated method stub

    }

    public void mouseExited(MouseEvent e) {
        // TODO Auto-generated method stub

    }
}
