/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


package edu.cmu.taghelpertools.middle_layer;

import edu.cmu.taghelpertools.middle_layer.extractor.FeatureExtractionHandler2;
import edu.cmu.taghelpertools.middle_layer.extractor.FeatureSpaceMap;
import edu.cmu.taghelpertools.middle_layer.extractor.ThtParser;
import gnu.trove.THashMap;
import gnu.trove.THashSet;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;

import javax.swing.JOptionPane;

/**
 * An abstraction of a Data Point, either test set or training set. It contains
 * an array of the dimensions (dims) associated with this data point. Each
 * dimension is a separate classification problem. Also contains an array of
 * booleans (isReference) that specify if this data point is a training data
 * point or test data point with respect to the dimension in that array index.
 * Likewise, the array of scores indicates the prediction confidence of this
 * data point with respect to a dimension. Reference data points have a
 * confidence of 1, because they are user-specified. Data points that are
 * initially unknown (marked with a '?') have a confidence of 0. The text is
 * common to all dimensions. Alternatively, one can view a Data Point as one row
 * in the input .xls file.
 * 
 * @author Jaime Arguello, Mahesh Joshi, Hao-Chuan Wang
 * 
 */
public class DataPoint implements FeatureSpaceMapProvider{

    public static int dimNum = Integer.MIN_VALUE;

    // The set of nominal features should be the same for all
    // the datapoints (should hold a missing value if not
    // present)
    public static ArrayList nominalFeatureNames;
    public static THashMap nominalFeatureValueSet;
    public static int textColumn = Integer.MIN_VALUE;
    public static int auxTextColMin = Integer.MIN_VALUE;
    public static int auxTextColMax = Integer.MIN_VALUE;
    public static int nominalStartColumn = Integer.MIN_VALUE;
    public static int subPopColumn = Integer.MIN_VALUE;
    
    public static final String MISSING_NOMINAL_VALUE = "MISSING_NOMINAL_VALUE";
    
    private String[] dims;
    private String[] predictedDims;
    
    private String[] auxTexts;
    
    private double[] scores;
    private boolean[] isReference;
    private String text;
    // private THashSet featureSet;
    
    // stores the features of this datapoint (both
    // when in simple or in advanced feature mode)
    // note: currently not sure if this is required in the
    // advanced feature mode, but keeping it as is
    private THashMap featureSetMap;
    private FeatureSelector featureSelector;
    private int numTokens;
    private String subPopId;
    
    // for the developer mode feature with train/test pairs
    private boolean isATestPoint;
    
    private boolean usesAdvancedFeatures;
    
    //this is only for error analysis (visualization)
    private  HashSet isMatchedUserDefined = new HashSet();
    
    // set of additional features for each datapoint, that are not necessarily
    // lexical.
    // these can be features such as the agent of the sentence (tutor or student), the exchange
    // identifier etc. in a tutorial dialog
    private ArrayList nominalFeatureValues;

    //this is the inverted file for each data point. by Hao-Chuan Wang
    // stores the features specific to the datapoint, when using
    // the advanced features mode
    private FeatureSpaceMap featureSpaceMap = null;
    
    //yichia
    private String taggedResult = null;
    
    /**
     * Instatiates a data point with the dimensions and text specified.
     * 
     * @param dims
     * @param text
     */
    public DataPoint(String[] dims, String text, 
			 FeatureSelector featureSelector, 
			 ArrayList nominalFeatures, 
			 boolean testPoint) 
    {
    	this(dims, text, featureSelector, nominalFeatures, testPoint, false);
    }
    
    public DataPoint(String[] dims, String text, 
    				 FeatureSelector featureSelector, 
    				 ArrayList nominalFeatures, 
    				 boolean testPoint,
    				 boolean forFeatureExtraction) 
    {
        // feature selector
        this.featureSelector = featureSelector;
        this.isATestPoint = testPoint;
        this.usesAdvancedFeatures = forFeatureExtraction;
        
        if (dims.length != dimNum) {
            System.out.println("Invalid data. Data point needs " + dimNum + " dimensions.");
            System.exit(0);
        } else {

            this.text = text;
            this.dims = dims;
            this.predictedDims = new String[dims.length];

            // marking if the datapoint is training data (this is
            // dimension-specific)
            isReference = new boolean[dims.length];
            for (int i = 0; i < dims.length; i++) {
                if (dims[i].equals(InputFileReader.UNKNOWN_STRING)) {
                    isReference[i] = false;
                } else {
                    isReference[i] = true;
                }
            }

            // initiating scores to all zeroes
            this.scores = new double[dims.length];
            for (int i = 0; i < dims.length; i++) {
                if (isReference[i]) {
                    scores[i] = 1.0;
                } else {
                    scores[i] = 0.0;
                }
            }

            // extracting the features from text;
            if (testPoint) {
        			if (forFeatureExtraction) {
        				// the advanced feature extraction strategy has to be different
        				// in this case since we only want to use the existing
        				// features in the advanced features model.
        				
        				// therefore, simply calling the advanced feature set related
        				// .getFeatureSet(...) function won't work as in the case below
        				
        				// need to use the same strategy as SavedModelEvaluator class.
        				
        				FeatureSpaceMap tmpfsm = new FeatureSpaceMap();
        				// featureSpaceMap used by advanced feature extractor
        				this.featureSpaceMap = new FeatureSpaceMap();
        				// featureSetMap used when creating WEKA/MinorThird instances
        				this.featureSetMap = this.featureSelector.getFeatureSet(text,
        						this.featureSpaceMap, tmpfsm);
        				// doing advanced feature extraction
        				Iterator itr = this.featureSelector
        						.getGlobalUserDefinedFeatureSetIterator();
        				while (itr.hasNext()) {
        					String advAtt = itr.next().toString();
        					String expression = this.featureSelector
        							.getlobalUserDefinedFeatureExpression(advAtt);

        					FeatureExtractionHandler2.setParameters(this, this.featureSelector);
        					try {
        						new ThtParser(
        								new edu.cmu.taghelpertools.middle_layer.extractor.ThtYylex(
        										new java.io.StringReader(expression))).parse();
        					} catch (Exception ex) {
        						ex.printStackTrace();
        						JOptionPane.showMessageDialog(null,
        								"Error on processing data point. Please check.",
        								"SYNTAX ERROR", JOptionPane.ERROR_MESSAGE);
        					}
        				}
        			} else {            	
                this.featureSetMap = this.featureSelector.getTestPointFeatureSet(text);
        			}
            } 
            else 
            {
            	if(forFeatureExtraction)
            	{
            		FeatureSpaceMap tmpfsm = new FeatureSpaceMap();
            		this.featureSpaceMap = new FeatureSpaceMap();
            		this.featureSetMap = this.featureSelector.getFeatureSet(text, this.featureSpaceMap, tmpfsm);
            		this.featureSelector.addFeatures(tmpfsm);//(this.featureSpaceMap); 
            		//for visualization
            		this.taggedResult = this.featureSelector.getTaggedText();
            		//System.out.println("DP "+ taggedResult);
            	}
            	else
            	{
            		this.featureSetMap = this.featureSelector.getFeatureSet(text);
                // add these features to the collection of features for the whole
                // dataSet
                this.featureSelector.addFeatures(new THashSet(this.featureSetMap.keySet()));
            	}
            }
            
            // copy the nominal features 
            nominalFeatureValues = nominalFeatures;
        }
    }
    
    
    public boolean isMatchedUserDefined(String variableName)
    {
    	
    	//return this.featureSpaceMap.syntheticFeatureExpressions.containsKey(variableName);
    	return this.isMatchedUserDefined.contains(variableName);
    	
    }
    
    public void setMatchedUserDefined(String variableName)
    {
    	
    	this.isMatchedUserDefined.add(variableName);
    }
    
    public void unsetMatchedUserDefined(String variableName)
    {
    	
    	this.isMatchedUserDefined.remove(variableName);
    }
    
    public DataPoint(String text, 
    				FeatureSelector featureSelector) 
    {
    		this.featureSelector = featureSelector;

    		this.text = text;
    		this.dims = null;
    		this.predictedDims = null;

    		this.featureSetMap = this.featureSelector.getFeatureSet(text);
    	  this.featureSelector.addFeatures(new THashSet(this.featureSetMap.keySet()));            		              
    }

    public FeatureSpaceMap getFeatureSpaceMap()
    {
    	return this.featureSpaceMap;
    }

    public THashMap getFeatures() {
        return featureSetMap;
    }
    public void setFeatures(THashMap ths)
    {
    	this.featureSetMap = ths;
    }
    
	public void setFeatureSpaceMap(FeatureSpaceMap fsm) {
		// TODO Auto-generated method stub
		this.featureSpaceMap = fsm;
		
	}

    public ArrayList getNominalFeatures() {
      return nominalFeatureValues;
    }
    
    public double getScoreForDim(int index) {
        return scores[index];
    }

    public void setScoreForDim(int index, double score) {
        if (!isReference[index]) {
            scores[index] = score;
        } else {
            System.out.println("can not set score of a datapoint that is a reference.");
            System.exit(0);
        }
    }

    public String getValueAtDim(int index) {
        return dims[index];
    }

    public boolean getIsReferenceAtDim(int index) {
        return isReference[index];
    }

    public void setValueAtDim(int index, String value) {
      if (!isReference[index]) {
          dims[index] = value;
      } else {
          System.out.println("can not set value of a data point that is a reference");
          System.exit(0);
      }
    }
    
    // following 2 methods might be useful if we eventually
    // decide to report actual vs. predicted value of class for
    // cross-validation experiment
    public void setPredictedValueAtDim(int index, String value) {
    	predictedDims[index] = value;
    }

    public String getPredictedValueAtDim (int index) {
    	return predictedDims[index];
    }
    
    public void setIsReferenceAtDim(int index) {
        if (!isReference[index]) {
            isReference[index] = true;
        } else {
            System.out.println("trying to a dim as reference when it was already a reference");
            System.exit(0);
        }
    }

    public String getText() {
        // return text;
    	String retText = this.text;
    	
    	if (auxTexts != null)
      	for (int i = 0; i < auxTexts.length; i++)
      		retText = retText + " " + auxTexts[i];
    	
    	return retText;
    }

    public String getQuotedText(String separator) {
      // return text;
    	String retText = "\"" + this.text + "\"";
    	
    	if (auxTexts != null)
      	for (int i = 0; i < auxTexts.length; i++)
      		retText = retText + separator + "\"" + auxTexts[i] + "\"";
    	
    	return retText;
    }
    
    public void setText(String text) {
        this.text = text;
    }
    
    public String getTaggedResult()
    {
    	return this.taggedResult;
    }
    
    private void setNumTokens(int num)
    {
    	this.numTokens = num;
    }
    
    private int getNumTokens()
    {
    	return this.numTokens;
    }
    

    public String toString() {
        String result = "";
        for (int i = 0; i < this.dims.length; i++) {
            result += (dims[i].toString() + ",");
        }
        result += this.text;
        return result;
    }

		public String getSubPopId() {
			return subPopId;
		}

		public void setSubPopId(String subPopId) {
			this.subPopId = subPopId;
		}
		
		public void setAuxTexts(String [] auxTexts) {
			this.auxTexts = auxTexts;
			String auxTextPrefix = "auxtext";
			
			for (int i = 0; i < auxTexts.length; i++) {
				String customPrefix = auxTextPrefix + (i+2) + "_";
        // extracting the features from text;
        if (this.isATestPoint) {
    			if (this.usesAdvancedFeatures) {
    				FeatureSpaceMap tmpfsm = new FeatureSpaceMap();
    				// add auxiliary text features into existing feature space
    				this.featureSetMap.putAll(this.featureSelector.getFeatureSet(auxTexts[i],
    						this.featureSpaceMap, tmpfsm, customPrefix));
    				// doing advanced feature extraction
    				Iterator itr = this.featureSelector
    						.getGlobalUserDefinedFeatureSetIterator();
    				while (itr.hasNext()) {
    					String advAtt = itr.next().toString();
    					String expression = this.featureSelector
    							.getlobalUserDefinedFeatureExpression(advAtt);

    					FeatureExtractionHandler2.setParameters(this, this.featureSelector);
    					try {
    						new ThtParser(
    								new edu.cmu.taghelpertools.middle_layer.extractor.ThtYylex(
    										new java.io.StringReader(expression))).parse();
    					} catch (Exception ex) {
    						ex.printStackTrace();
    						JOptionPane.showMessageDialog(null,
    								"Error on processing data point. Please check.",
    								"SYNTAX ERROR", JOptionPane.ERROR_MESSAGE);
    					}
    				}
    			} else {
    				// add auxiliary text features into existing feature space
            this.featureSetMap.putAll(this.featureSelector.getTestPointFeatureSet(auxTexts[i], customPrefix));
    			}
        } 
        else 
        {
        	if(this.usesAdvancedFeatures)
        	{
        		FeatureSpaceMap tmpfsm = new FeatureSpaceMap();
        		this.featureSpaceMap = new FeatureSpaceMap();
        		this.featureSetMap.putAll(this.featureSelector.getFeatureSet(auxTexts[i], 
        				this.featureSpaceMap, tmpfsm, customPrefix));
        		this.featureSelector.addFeatures(tmpfsm);//(this.featureSpaceMap);
        		//for visualization
        		this.taggedResult = "  " + customPrefix + ": " + this.featureSelector.getTaggedText();
        	}
        	else
        	{
        		this.featureSetMap.putAll(this.featureSelector.getFeatureSet(auxTexts[i], customPrefix));
            // add these features to the collection of features for the whole
            // dataSet
            this.featureSelector.addFeatures(new THashSet(this.featureSetMap.keySet()));
        	}
        }
  		}
		}
}
