/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package edu.cmu.taghelpertools.middle_layer;

import edu.cmu.taghelpertools.io.Logger;
import edu.cmu.taghelpertools.io.Reader;
import edu.cmu.taghelpertools.middle_layer.extractor.FeatureSpaceMap;
import edu.cmu.taghelpertools.tokenizer.AbstractTokenizer;
import edu.cmu.taghelpertools.tokenizer.StopwordManager;
import edu.cmu.taghelpertools.tokenizer.Token;
import edu.cmu.taghelpertools.tokenizer.Tokenizer;
import edu.cmu.taghelpertools.tokenizer.Tokens;
import edu.cmu.taghelpertools.util.HelperFunctions;

import gnu.trove.THashMap;
import gnu.trove.THashSet;

import java.io.*;
import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.swing.JOptionPane;

import weka.core.Attribute;
import weka.core.FastVector;
import weka.core.Instance;
import weka.core.Instances;

/**
 * FeatureSelector does all the feature selection. In other words, it is told
 * which features to extract from the text (e.g., unigrams, bigrams, POS
 * bigrams, sentence length, punctuation, etc), and extracts them. It is also
 * responsible for creating the "DICT" file, which is a flat-file that contains
 * all the options used and all the features for a classifier.
 * 
 * @author Jaime Arguello, Hao-Chuan Wang
 * 
 */
public class FeatureSelector{

		public static boolean DEBUG = false;
	
    // used to read in input file
    public static String SETTINGS_FLAG = "###SETTINGS###";
    public static String DICTIONARY_FLAG = "###DICT###";
    public static String LABEL_FLAG = "###LABELS###";

    // flags used to read in input file
    public static String UNIGRAM_FLAG = "unigrams";
    public static String BIGRAM_FLAG = "bigrams";
    public static String POS_BIGRAM_FLAG = "pos_bigrams";
    public static String PUNCT_FLAG = "punct";
    public static String LINE_LENGTH_FLAG = "line_length";
    public static String CONTAINS_NON_STOPWORD_FLAG = "contains_stopword";
    public static String STEM_FLAG = "stem";
    public static String REMOVE_STOPWORDS_FLAG = "remove_stopwords";
    public static String REMOVE_RARE = "remove_rare";
    public static String REMOVE_RARE_THRESHOLD = "remove_rare_threshold";
    public static String MAX_CNT = "max_count";
    public static String MIN_CNT = "min_count";
    

    public static String CONTAINS_NON_STOPWORD_TRUE = "CONTAINS_NON_STOP";
    public static String CONTAINS_NON_STOPWORD_FALSE = "DOES_NOT_CONTAIN_NON_STOP";
    
    public static String CONTAINS_QUESTION_WORD_TRUE = "CONTAINS_QUESTION_WORD";
    public static String CONTAINS_QUESTION_WORD_FALSE = "DOES_NOT_CONTAIN_QUESTION_WORD";    

    // used to map punctuation to String
    //private static TIntObjectHashMap punctStringMap;
    private static THashMap punctStringMap;

//    private static THashMap questionWords;
    
    private static String COMMA = "COMMA";
    private static String PERIOD = "PERIOD";
    private static String SEMI_COLON = "SEMI_COLON";
    private static String COLON = "COLON";
    private static String DASH = "DASH";
    private static String QUESTION_MARK = "QUESTION_MARK";
    private static String EXCLAMATION_MARK = "EXCLAMATION_MARK";
    private static String ASTERISK = "ASTERISK";
    private static String AMPERSAND = "AMPERSAND";
    private static String AT_SIGN = "AT_SIGN";
    private static String NUMBER_SIGN = "NUMBER_SIGN";
    private static String PERCENT_SIGN = "PERCENT_SIGN";
    private static String OPEN_PAREN = "OPEN_PAREN";
    private static String CLOSE_PAREN = "CLOSE_PAREN";
    private static String OPEN_CURLY = "OPEN_CURLY";
    private static String CLOSE_CURLY = "CLOSE_CURLY";
    private static String LESS_THAN = "LESS_THAN";
    private static String GREATER_THAN = "GREATER_THAN";
    private static String DOUBLE_QUOTE = "DOUBLE_QUOTE";
    private static String DOLLAR ="DOLLAR";
    private static String BACKSLASH = "BACKSLASH";
    private static String FORWARDSLASH = "FORWARDSLASH";
    private static String PLUS_SIGN = "PLUSSIGN";
    private static String TILDE = "TILDE";
    private static String CARAT = "CARAT";
    private static String LEFT_SQUARE_PAREN = "LEFTSQUAREPAREN";
    private static String RIGHT_SQUARE_PAREN = "RIGHTSQUAREPAREN";

    static {
//        punctStringMap = new TIntObjectHashMap();
//        punctStringMap.put(new Character(',').hashCode(), FeatureSelector.COMMA);
//        punctStringMap.put(new Character('.').hashCode(), FeatureSelector.PERIOD);
//        punctStringMap.put(new Character(';').hashCode(), FeatureSelector.SEMI_COLON);
//        punctStringMap.put(new Character(':').hashCode(), FeatureSelector.COLON);
//        punctStringMap.put(new Character('-').hashCode(), FeatureSelector.DASH);
//        punctStringMap.put(new Character('?').hashCode(), FeatureSelector.QUESTION_MARK);
//        punctStringMap.put(new Character('!').hashCode(), FeatureSelector.EXCLAMATION_MARK);
//        punctStringMap.put(new Character('*').hashCode(), FeatureSelector.ASTERISK);
//        punctStringMap.put(new Character('&').hashCode(), FeatureSelector.AMPERSAND);
//        punctStringMap.put(new Character('@').hashCode(), FeatureSelector.AT_SIGN);
//        punctStringMap.put(new Character('#').hashCode(), FeatureSelector.NUMBER_SIGN);
//        punctStringMap.put(new Character('%').hashCode(), FeatureSelector.PERCENT_SIGN);
//        punctStringMap.put(new Character('(').hashCode(), FeatureSelector.OPEN_PAREN);
//        punctStringMap.put(new Character(')').hashCode(), FeatureSelector.CLOSE_PAREN);
//        punctStringMap.put(new Character('{').hashCode(), FeatureSelector.OPEN_CURLY);
//        punctStringMap.put(new Character('}').hashCode(), FeatureSelector.CLOSE_CURLY);
//        punctStringMap.put(new Character('<').hashCode(), FeatureSelector.LESS_THAN);
//        punctStringMap.put(new Character('>').hashCode(), FeatureSelector.GREATER_THAN);
//        punctStringMap.put(new Character('"').hashCode(), FeatureSelector.DOUBLE_QUOTE);
//        punctStringMap.put(new Character('$').hashCode(), FeatureSelector.DOLLAR); 

      punctStringMap = new THashMap();
      punctStringMap.put("<COMMA>", FeatureSelector.COMMA);
      punctStringMap.put("<PERIOD>", FeatureSelector.PERIOD);
      punctStringMap.put("<SEMICOLON>", FeatureSelector.SEMI_COLON);
      punctStringMap.put("<COLON>", FeatureSelector.COLON);
      punctStringMap.put("<DASH>", FeatureSelector.DASH);
      punctStringMap.put("<QUESTIONMARK>", FeatureSelector.QUESTION_MARK);
      punctStringMap.put("<EXCLAMATIONMARK>", FeatureSelector.EXCLAMATION_MARK);
      punctStringMap.put("<ASTERISK>", FeatureSelector.ASTERISK);
      punctStringMap.put("<AMPERSAND>", FeatureSelector.AMPERSAND);
      punctStringMap.put("<ATSIGN>", FeatureSelector.AT_SIGN);
      punctStringMap.put("<NUMBERSIGN>", FeatureSelector.NUMBER_SIGN);
      punctStringMap.put("<PERCENTSIGN>", FeatureSelector.PERCENT_SIGN);
      punctStringMap.put("<CLOSEPAREN>", FeatureSelector.CLOSE_PAREN);
      punctStringMap.put("<CLOSEPAREN>", FeatureSelector.CLOSE_PAREN);
      punctStringMap.put("<OPENCURLY>", FeatureSelector.OPEN_CURLY);
      punctStringMap.put("<CLOSECURLY>", FeatureSelector.CLOSE_CURLY);
      punctStringMap.put("<LESSTHAN>", FeatureSelector.LESS_THAN);
      punctStringMap.put("<GREATERTHAN>", FeatureSelector.GREATER_THAN);
      punctStringMap.put("<DOUBLEQUOTE>", FeatureSelector.DOUBLE_QUOTE);
      punctStringMap.put("<DOLLAR>", FeatureSelector.DOLLAR);
      punctStringMap.put("<BACKSLASH>", FeatureSelector.BACKSLASH);
      punctStringMap.put("<FORWARDSLASH>", FeatureSelector.FORWARDSLASH);
      punctStringMap.put("<PLUSSIGN>", FeatureSelector.PLUS_SIGN);
      punctStringMap.put("<TILDE>", FeatureSelector.TILDE);
      punctStringMap.put("<CARAT>", FeatureSelector.CARAT);
      punctStringMap.put("<LEFTSQUAREPAREN>", FeatureSelector.LEFT_SQUARE_PAREN);
      punctStringMap.put("<RIGHTSQUAREPAREN>", FeatureSelector.RIGHT_SQUARE_PAREN);
      
      String debug = System.getProperty("tht.debug");
      if (debug != null && debug.equals("1"))
      	FeatureSelector.DEBUG = true;
      
//		try{
//			String currentLine = ""; 
//			BufferedReader rd = new BufferedReader(new FileReader("chinese.txt"));             
//			for(int i=0; (currentLine = rd.readLine())!= null; i++)
//			{
//					String[] values= currentLine.split(" ");
//					if(values.length == 2)
//					{
//						punctStringMap.put(values[0].trim().hashCode(), values[1].trim());
//					}
//			}						
//		}           
//		catch(Exception ex)
//		{
//			ex.printStackTrace();
//		}        

        //TIntObjectIterator keyIter = punctStringMap.iterator();
      
//        while (keyIter.hasNext()) {
//            keyIter.advance();
//            int hashCode = keyIter.key();
//            System.out.println("key=" + hashCode + "," + "value=" + punctStringMap.get(hashCode));
//        }
        
//      Iterator puncIter = punctStringMap.keySet().iterator();
//      while(puncIter.hasNext())
//      {
//    	  System.out.println("Punc: "+(String)puncIter.next());
//      }
      
//        questionWords = new THashMap();
//        
//        questionWords.put("what", null);
//        questionWords.put("who", null);
//        questionWords.put("why", null);
//        questionWords.put("where", null);
//        questionWords.put("when", null);
//        questionWords.put("which", null);
//        questionWords.put("whose", null);
//        questionWords.put("whom", null);
//        questionWords.put("how", null);
//        questionWords.put("do", null);
//        questionWords.put("did", null);
    }

    /**
     * Contains the feature set.
     */
    private THashMap featureMap;
    
    /**
     * Contains user defined feature set;
     */
    private TreeMap<String, FeatureGlobalProperty> globalSynFeatureMap;

    /**
     * Contrains all the possible output labels of this particular classifier.
     */
    private FastVector labels;

    /**
     * Contains the minimum sentence length. This is used because a sentence's
     * length (an optional feature) is represented as: (X - minCount)/(maxCount -
     * minCount), where X is the length of *this* sentence.
     */
    private int minCount = Integer.MAX_VALUE;

    /**
     * Contains the maximum sentence length. This is used because a sentence's
     * length (an optional feature) is represented as: (X - minCount)/(maxCount -
     * minCount), where X is the length of *this* sentence.
     */
    private int maxCount = Integer.MIN_VALUE;


    /**
     * Contains the feature-selection settings
     */
    private OptionsModel optionsModel;

    public THashMap getFeatureMap()
    {
    	return this.featureMap;
    }
    public void setFeatureMap(THashMap thm)
    {
    	this.featureMap = thm;
    }
    /**
     * @return - an Iterator with all the features collected by this
     *         FeatureSelector
     */
    public Iterator getFeatureSetIterator() {
       	return getFeatureSetIterator(false);
    }
    
    public Iterator getFeatureSetIterator(boolean sorting) {
    	if(sorting == false)
    	{
    		return featureMap.keySet().iterator();
    	}
    	else
    	{
    		List l = new ArrayList();
    		Iterator itr = featureMap.keySet().iterator();
    		while(itr.hasNext())
    		{
    			l.add(itr.next());
    		}
    		    		
    		Collections.sort(l);
    		return l.iterator();
    	}
    }

    /**
     * @return - a FastVector containing all the string values of the labels the
     *         classifier is classifying into.
     */
    public FastVector getLabels() {
        return labels;
    }

    public void setMinCount(int minCount) {
        this.minCount = minCount;
    }

    public int getMinCount() {
        return this.minCount;
    }

    public void setMaxCount(int maxCount) {
        this.maxCount = maxCount;
    }

    public int getMaxCount() {
        return this.maxCount;
    }

    
    public String getUserDefinedText()
    {
    	StringBuffer sb = new StringBuffer();
    	String sp = System.getProperty("line.separator");       	

    	Iterator itr = this.getGlobalUserDefinedFeatureSetIterator();
    	while(itr.hasNext())
    	{
    		String id = itr.next().toString();
    		//sb.append(this.getlobalUserDefinedFeatureExpression(id).replaceAll("=", ":")).append(sp);
    		sb.append(this.getlobalUserDefinedFeatureExpression(id)).append(sp);
    	}

    	return sb.toString();
    }
    
    public int getNumGlobalUserDefiendFeature()
    {
    	if(this.globalSynFeatureMap != null)
    		return this.globalSynFeatureMap.size();
    	else
    		return 0;
    }
    
    public Iterator getGlobalUserDefinedFeatureSetIterator() {

    	// we should return the iterator such that composite features
    	// that refer to other user defined features are returned in
    	// the correct order
    	ArrayList<String> compositeFeatures = new ArrayList<String>();
    	ArrayList<String> allFeatures = new ArrayList<String>();
    	THashSet allFeat = new THashSet();
    	
    	Iterator itr = this.globalSynFeatureMap.keySet().iterator();
    	while (itr.hasNext()) {
    		String feat = itr.next().toString();
    		String expression = this.getlobalUserDefinedFeatureExpression(feat);
    		if (expression.indexOf("U:") != -1) {
    			compositeFeatures.add(feat);
    		} else {
    			allFeatures.add(feat);
    			allFeat.add(feat);
    		}
    	}
    	
    	while (compositeFeatures.size() != 0) {
    		ArrayList<String> tmpFeat = new ArrayList<String>();
    		int size = compositeFeatures.size();
    		for (int i = size - 1; i >= 0; i--) {
    			String feat = compositeFeatures.remove(i);
    			String expression = this.getlobalUserDefinedFeatureExpression(feat);
    			// get the feature definition
    			String definition = expression.split("=", 2)[1];
    			definition = definition.trim();
    			definition = definition.replaceAll(" ", "");
    			// next statement assumes that the first left parenthesis begins the operands
    			// in the feature definition, and that the last character is the ending
    			// right parenthesis
    			definition = definition.substring(definition.indexOf("(")+1, definition.length()-1);
    			String [] operands = definition.split(",");
    			boolean addedToTmp = false;
    			for (int j = 0; j < operands.length; j++) {
    				if (operands[j].startsWith("U:")) {
    					if (!allFeat.contains(operands[j])) {
    						tmpFeat.add(feat);
    						addedToTmp = true;
    						break;
    					}
    				}
    			}
    			if (!addedToTmp) {
    				allFeatures.add(feat);
    				allFeat.add(feat);
    			}
    		}
    		if (tmpFeat.size() > 0) {
    			compositeFeatures.addAll(tmpFeat);
    		}
    	}
    	
    	return allFeatures.iterator();
    }
    
    public  TreeMap<String, FeatureGlobalProperty> getGlobalUserDefinedFeature()
    {
    	return this.globalSynFeatureMap;
    }
    
    public String getlobalUserDefinedFeatureExpression(String featureName)
    {
    	FeatureGlobalProperty tmp = (FeatureGlobalProperty)this.globalSynFeatureMap.get(featureName);
    	if(tmp != null)
    		return tmp.getExpression();
    	else
    		return "";
    }
    
    public void removeGlobalUserDefiendFeature(String featureName)
    {
    	if(this.globalSynFeatureMap.containsKey(featureName)) 
    	{
    		this.globalSynFeatureMap.remove(featureName);
    		//System.out.println("FS - global user feature size: "+ this.globalSynFeatureMap.size());
    		this.optionsModel.getOptionsFrame().updateAdvFeatureLabel();    		
    	}
    	else
    	{
			
    	}    	
    }
    
    public void addGlobalUserDefiendFeature(String featureName, 
    								  		String expression)
    {
    	//System.out.println("FS - addGlobalUserDefiendFeature");
    	if(this.globalSynFeatureMap.containsKey(featureName)) 
    	{
    		//System.out.println("FS - addGlobalUserDefiendFeature 1");
    		FeatureGlobalProperty tmp = (FeatureGlobalProperty)this.globalSynFeatureMap.get(featureName);
    		
    		if(tmp.getExpression().equals(expression)) // check whether the feature is re-defined
    		{
    			tmp.addCount();
    		}
    		else
    		{
    			this.globalSynFeatureMap.put(featureName, 
    										 new FeatureGlobalProperty(1, expression)); // redefine the feeature with the particular ID
    			this.optionsModel.getOptionsFrame().updateAdvFeatureLabel();
    		}
    	}
    	else
    	{
    		//System.out.println("FS - addGlobalUserDefiendFeature 2");
			this.globalSynFeatureMap.put(featureName, 
										 new FeatureGlobalProperty(1, expression));
			this.optionsModel.getOptionsFrame().updateAdvFeatureLabel();			
    	}
    }
    
    public void addFeatures(THashSet featureSet) {
        Iterator iter = featureSet.iterator();
        while (iter.hasNext()) {
            String feature = iter.next().toString();
            if (this.featureMap.contains(feature)) {
                int count = ((Integer) this.featureMap.get(feature)).intValue();
                count++;
                featureMap.put(feature, new Integer(count));
            } else {
                this.featureMap.put(feature, new Integer(1));
            }
        }
    }
    
    public void addFeatures(FeatureSpaceMap fsm) {
        Iterator iter = fsm.getOrdKeyIterator();
        while (iter.hasNext()) {
            String featureName = iter.next().toString();
            if (this.featureMap.contains(featureName)) {
            	FeatureGlobalProperty tmp = (FeatureGlobalProperty)this.featureMap.get(featureName);
            	tmp.addCount();
//            	tmp.update(fsm.isPOS(featureName), fsm.isBigram(featureName));
            } 
            else 
            {
            	FeatureGlobalProperty tmp = new FeatureGlobalProperty(fsm.isPOS(featureName),
            														  fsm.isBigram(featureName), 
            														  fsm.isNonCountable(featureName),
            														  1);            
                this.featureMap.put(featureName, tmp);
            }
        }
    }    
    

    /**
     * Sets up the FeatureSelector
     * 
     * @param optionsModel --
     *            contains feature selection options
     */
    /**
     * @param optionsModel
     */
    public FeatureSelector(OptionsModel optionsModel)
    {
    	this(optionsModel, false);
    }
    
    public FeatureSelector(OptionsModel optionsModel, boolean isReload) {
        this.optionsModel = optionsModel;
        this.featureMap = new THashMap();    
        this.globalSynFeatureMap = new TreeMap<String, FeatureGlobalProperty>();
        if(!isReload)
        	this.optionsModel.setFeatureMap(featureMap);
    }
    
    public FeatureSelector(OptionsModel optionsModel, TreeMap<String, FeatureGlobalProperty> globalSynFeatureMap, boolean isReload) {
        this.optionsModel = optionsModel;
        this.featureMap = new THashMap();    
        this.globalSynFeatureMap = globalSynFeatureMap;
        if(!isReload)
        	this.optionsModel.setFeatureMap(featureMap);
    }    
   
    
    /**
     * Load during test phase. The input file must be .dic file
     */
    public FeatureSelector(File input) {
        this.labels = new FastVector();
        this.featureMap = new THashMap();
        this.readDicFile(input);
        
        //TODO: 
//        this.setMinCount(1);
//        this.setMaxCount(30);
    }

    
    /**
     * Feature counts are collected because the user can optionally remove rare
     * features (features that occur less than THRESH number of times). Those
     * features that occur less than THRESH number of times are removed from the
     * featureSet
     */
    public void removeRareFeatures() {
        Iterator featureIter = this.featureMap.keySet().iterator();
     // get the threshold that was set from the user iterface
        int thresh = this.optionsModel.getFrequencyThreshold();
        while (featureIter.hasNext()) {
            String feature = featureIter.next().toString();
            Object tmp =  featureMap.get(feature);
            int count;
            if(tmp.getClass()==Integer.class)
            	count = ((Integer)tmp).intValue();
            else
            	count = ((FeatureGlobalProperty)tmp).getCount();
            if (count < thresh) {
                featureIter.remove();
            }
        }
    }
    
    public void removePreprocessedOptionFeatures1(String matchedPrefix) {
        Iterator featureIter = this.featureMap.keySet().iterator();
        while (featureIter.hasNext()) 
        {
            String feature = featureIter.next().toString();
           //System.out.println("yichia "+ matchedPrefix + "@@" +feature);
            if (feature.startsWith("THF_") && feature.startsWith(matchedPrefix))
            {
            	//System.out.println("yichia "+ matchedPrefix + "@@" +feature);
            	featureIter.remove();
            }
            	//feature.substring(matchedPrefix.length());  
        }
        //System.out.println("removePreprocessedOptionFeatures1:: featureMapSize "+ featureMap.size());
        
    }
    
    public void removePreprocessedOptionFeatures2(String prefix1, String prefix2) {
        Iterator featureIter = this.featureMap.keySet().iterator();
        while (featureIter.hasNext()) 
        {
            String feature = featureIter.next().toString();
           //System.out.println("yichia "+ matchedPrefix + "@@" +feature);
            if (feature.startsWith(prefix1) && !feature.startsWith(prefix2))
            {
            	//System.out.println("yichia "+ prefix2 + "@@" +feature);
            	featureIter.remove();
            }
            	//feature.substring(matchedPrefix.length());  
        }
        // System.out.println("removePreprocessedOptionFeatures2:: featureMapSize "+ featureMap.size());
    }
    
    public void removeFeatureMapPrefix()
    {
    	Pattern p = Pattern.compile("(THF_)([a-zA-Z]+_)(.+)");
    	THashMap tmp = new THashMap();
    	Iterator featureIter = this.featureMap.keySet().iterator();
    	//System.out.println("featureMap  " + featureMap.size());
        while (featureIter.hasNext()) 
        {
            String feature = featureIter.next().toString();
            Object value = this.featureMap.get(feature);
            
    		Matcher m = p.matcher(feature);
    		if (m.find())
    		{//strip off prefix
    			String thirdToken = m.group(3).trim();
    			feature = thirdToken;
    		}
    		//System.out.println("feature  "+ feature);
    		tmp.put(feature, value);
        }
        //this.featureMap = null;
        this.featureMap = tmp;
        
        // need to update the feature map in optionsModel as
        // it holds a pointer to the old feature map
        this.optionsModel.setFeatureMap(this.featureMap);
        
        //System.out.println("featureMap2 "+ featureMap.size());
    }
    
    /**
     * Method to remove the features that contain punctuation.
     * Is called from MultiDimClassifier methods setUp() and
     * setPreprocessed().
     */
  	public void removePuncFeature() {
  
  		THashMap tmp = new THashMap();
  		Iterator featureIter = this.featureMap.keySet().iterator();
  		while (featureIter.hasNext()) {
  			String feature = featureIter.next().toString();
  			Object value = this.featureMap.get(feature);
  			// remove any features that contain punctuation (which is
  			// replaced by special tokens within <> tag
  			// however, we have to carefully avoid a POS-bigram feature
  			// like PRP<dollar>_NN or VBD_PRP<dollar>, so just check for a feature that
  			// begins with '<' (like <dash> or <dash>_<dash>) or
  			// that contains '_<' (like NN_<period>) and skip it
  			if (!feature.startsWith("<") && !feature.contains("_<")) {
  				tmp.put(feature, value);
  			}
  		}
  		// update the feature map with the reduced set of features
  		this.featureMap = tmp;
  		// need to update the feature map in optionsModel as
  		// it holds a pointer to the old feature map
  		this.optionsModel.setFeatureMap(this.featureMap);
  	}
	
	public void removeGlobalSynFeatureMapPrefix() {
		// TODO Auto-generated method stub

		TreeMap<String, FeatureGlobalProperty> tmpgsfm = new TreeMap();        	
    	Iterator iter = this.globalSynFeatureMap.keySet().iterator();
    	while (iter.hasNext())
    	{
    		String feature = (String)iter.next();
    		FeatureGlobalProperty value = (FeatureGlobalProperty)this.globalSynFeatureMap.get(feature);
	        String exptmp = value.getExpression();
	        
	        exptmp = exptmp.replaceAll("THF_UStemRmstopw_", "");
	        exptmp = exptmp.replaceAll("THF_UStemNormstopw_", "");
	        exptmp = exptmp.replaceAll("THF_UNostemNormstopw_", "");
	        exptmp = exptmp.replaceAll("THF_UNostemRmstopw_", "");
	        exptmp = exptmp.replaceAll("THF_BStemRmstopw_", "");
	        exptmp = exptmp.replaceAll("THF_BStemNormstopw_", "");
	        exptmp = exptmp.replaceAll("THF_BNostemNormstopw_", "");
	        exptmp = exptmp.replaceAll("THF_BNostemRmstopw_", "");
	        exptmp = exptmp.replaceAll("THF_POS_", "");
	        value.setExpression(exptmp);
	         
	        System.out.println("exp: "+ exptmp);
	       
    		
    	}
    	
		
	}
    /**
     * Given, a data point, output a Weka instance that the classifier can be
     * trained on
     * 
     * @param DataPoint
     *            datapoint
     * @param Weka
     *            Instances instances describes the dimensionality of the
     *            instance
     * @return Weka Instance
     */
    public Instance getInstanceFromDataPoint(DataPoint dataPoint, Instances instances) {
        Instance instance = new Instance(instances.numAttributes());
        THashMap dataPointFeatureMap = dataPoint.getFeatures();
        Iterator dataPointFeatIter = dataPointFeatureMap.keySet().iterator();
        boolean usingAdvFeatures = this.optionsModel.useAdvFeatures();
        
        if (this.optionsModel.isSubPop()) {
        	Attribute att = instances.attribute(InputFileReader.SUBPOP);
        	// on purpose, don't check att for null
        	// we want an exception if the ARFF header does not
        	// contain 'subpop'
        	instance.setValue(att, dataPoint.getSubPopId());
        }
        
        // Instead of iterating over the entire feature set, just iterate
        // over the set of features in the current data point
        while (dataPointFeatIter.hasNext()) {
        	String feature = dataPointFeatIter.next().toString();
        	if (usingAdvFeatures)
        		feature = feature.replaceFirst(".*?_", "").replaceFirst(".*?_", "");
        		//System.out.println("feature  "+ feature);
            Attribute att = instances.attribute(feature);
            if (att != null) {
                // if (dataPointFeatureSet.contains(feature)) {
                    // instance.setValue(att, 1);
            	double value = this.optionsModel.isOutputFreqCounts() ? (Integer) dataPointFeatureMap.get(feature) : 1;
              instance.setValue(att, value);
            } // else {
                // WE SHOULD IGNORE ANY FEATURE NOT PRESENT IN THE
                // INSTANCE SPACE (SINCE THIS WAS REMOVED FROM THE FEATURE
                // SPACE DUE TO FREQUENCY THRESHOLD, BUT NOT FROM THE
                // DATAPOINTS LIST OF FEATURES)
                // JOptionPane.showMessageDialog(null, "The attribute was not found in the instances object passed to getInstanceFromText(). Contact: jaime@cs.cmu.edu", "TERMINAL ERROR", JOptionPane.ERROR_MESSAGE);
                // System.exit(0);
            // }
        }
        
        if (this.optionsModel.isLineLength()) {
//            int wordCount = this.countWords(dataPoint.getText());
//            double num = (double) (wordCount - this.minCount);
//            double denom = (double) (this.maxCount - this.minCount);
        	double linelength = this.getLineLength(dataPoint.getText());
            Attribute att = instances.attribute(FeatureSelector.LINE_LENGTH_FLAG);
            if (att != null) {
                instance.setValue(att, linelength/*((double) (num / denom))*/);
            } else {
                JOptionPane.showMessageDialog(null, "The attribute was not found in the instances object passed to getInstanceFromText(). Contact: jaime@cs.cmu.edu", "TERMINAL ERROR", JOptionPane.ERROR_MESSAGE);
                System.exit(0);
            }
        }
        // replace all the missing values in the instance with 0
        double [] zeroes = new double [instances.numAttributes()];
        instance.replaceMissingValues(zeroes);
        
        // fill in the values for the nominal features in the dataset
        for (int i = 0; i < DataPoint.nominalFeatureNames.size(); i++) {
            String nominalAtt = (String) DataPoint.nominalFeatureNames.get(i);
            Attribute att = instances.attribute(nominalAtt);
            if (att != null) {
                if (((String) dataPoint.getNominalFeatures().get(i)).equals(DataPoint.MISSING_NOMINAL_VALUE)) {
                    instance.setValue(att, Instance.missingValue());
                } else {
                		// find out if the value is numeric or nominal
                		// and update accordingly
                		String nomVal = (String) dataPoint.getNominalFeatures().get(i);
                		if (att.type() == Attribute.NUMERIC) {
                			instance.setValue(att, Double.parseDouble(nomVal));
                		}
                		else {
                			instance.setValue(att, nomVal);
                		}
                }
            }
        }
        
        // use of advanced features
        if (this.optionsModel.useAdvFeatures()){        
	        Iterator itr = this.getGlobalUserDefinedFeatureSetIterator();
	        while(itr.hasNext())
	        {
	        	String advAtt = itr.next().toString();
	            Attribute att = instances.attribute(advAtt);
	            if(att != null)
	            {
	            	if(dataPoint.getFeatureSpaceMap().getSynFeature(advAtt)!=null) {
	            		instance.setValue(att, 1);
	            	} else {
	            		instance.setValue(att, 0);
	            	}
	            }
	            else
	            {
	            	System.err.println("ERROR: Should NEVER reach here, this means that " +
	            			"the set of user defined features has changed between creation" +
	            			"of ARFF file header and generation of DataPoint instances.");
	            }
	        }
        }

        return instance;
    }

    public Instance getInstanceFromDataPoint(EvalDataPoint extDataPoint, Instances instances) {
        Instance instance = new Instance(instances.numAttributes());
        THashMap dataPointFeatureMap = extDataPoint.getFeatures();
        Iterator dataPointFeatIter = dataPointFeatureMap.keySet().iterator();
        boolean usingAdvFeatures = this.optionsModel.useAdvFeatures();
        
        if (this.optionsModel.isSubPop()) {
        	Attribute att = instances.attribute(InputFileReader.SUBPOP);
        	// on purpose, don't check att for null
        	// we want an exception if the ARFF header does not
        	// contain 'subpop'
        	instance.setValue(att, extDataPoint.getSubPopId());
        }
        

        if (this.optionsModel.isOutputFreqCounts()) {
          while (dataPointFeatIter.hasNext()) {
              String feature = dataPointFeatIter.next().toString();
              if (usingAdvFeatures)
              	feature = feature.replaceFirst(".*?_", "").replaceFirst(".*?_", "");
              Attribute att = instances.attribute(feature);
              if (att != null) {
              	double value = (Integer) dataPointFeatureMap.get(feature);
              	instance.setValue(att, value);
              }
          }
        } else {
          while (dataPointFeatIter.hasNext()) {
            String feature = dataPointFeatIter.next().toString();
            if (usingAdvFeatures)
            	feature = feature.replaceFirst(".*?_", "").replaceFirst(".*?_", "");
            Attribute att = instances.attribute(feature);
            if (att != null) {
            	instance.setValue(att, 1);
            }
          }
        }
        
        if (this.optionsModel.isLineLength()) {
//            int wordCount = this.countWords(dataPoint.getText());            
//            double num = (double) (wordCount - this.minCount);
//            double denom = (double) (this.maxCount - this.minCount);
        	double linelength = this.getLineLength(extDataPoint.getText());
//            System.out.println("TEXT---->"+dataPoint.getText()+", count:"+wordCount+", value"+(double)(num/denom));
            Attribute att = instances.attribute(FeatureSelector.LINE_LENGTH_FLAG);
            if (att != null) {
                instance.setValue(att, linelength/*((double) (num / denom))*/);
            } else {
                JOptionPane.showMessageDialog(null, "The attribute was not found in the instances object passed to getInstanceFromText(). Contact: jaime@cs.cmu.edu", "TERMINAL ERROR", JOptionPane.ERROR_MESSAGE);
                System.exit(0);
            }
        }
        // replace all the missing values in the instance with 0
        double [] zeroes = new double [instances.numAttributes()];
        instance.replaceMissingValues(zeroes);

        // fill in the values for the nominal features in the dataset
        if(EvalDataPoint.nominalFeatureNames != null)
        {
	        for (int i = 0; i < EvalDataPoint.nominalFeatureNames.size(); i++) {
	            String nominalAtt = (String) EvalDataPoint.nominalFeatureNames.get(i);
	            Attribute att = instances.attribute(nominalAtt);
	            if (att != null) {
	                if (((String) extDataPoint.getNominalFeatures().get(i)).equals(EvalDataPoint.MISSING_NOMINAL_VALUE)) {
	                    instance.setValue(att, Instance.missingValue());
	                } else {
                		// find out if the value is numeric or nominal
                		// and update accordingly
                		String nomVal = (String) extDataPoint.getNominalFeatures().get(i);
                		if (att.type() == Attribute.NUMERIC)
                			instance.setValue(att, Double.parseDouble(nomVal));
                		else {
                			try {
                				instance.setValue(att, nomVal);
                			} catch (IllegalArgumentException e) {
                				// ignore the novel nominal value in test data
                				// set attribute to missing
                				System.out.println("WARNING: Ignoring novel nominal value: " + nomVal);
                				instance.setValue(att, Instance.missingValue());
                			}
                		}
	                }
	            }
	        }
        }
        
        // use of advanced features
        if (this.optionsModel.useAdvFeatures()){        
	        Iterator itr = this.getGlobalUserDefinedFeatureSetIterator();
	        while(itr.hasNext())
	        {
	        	String advAtt = itr.next().toString();
	            Attribute att = instances.attribute(advAtt);
	            if(att != null)
	            {
	            	if(extDataPoint.getFeatureSpaceMap().getSynFeature(advAtt)!=null)
	            		instance.setValue(att, 1);
	            	else
	            		instance.setValue(att, 0);
	            }
	            else
	            {
	            	System.err.println("ERROR: Should NEVER reach here, this means that " +
	            			"the set of user defined features has changed between creation" +
	            			"of ARFF file header and generation of EvalDataPoint instances.");
	            }
	        }
        }        
        return instance;
    }    
    
    /**
     * Method to create one instance in MinorThird format
     * 
     * @param dataPoint
     * @param dummy The ARFF format header for the feature set
     * 				(required for checking presence of numeric attributes
     * 				in the nominal set of features)
     * @return
     */
    public String getMTInstanceFromDataPoint(DataPoint dataPoint, Instances dummy,
    	String dimValue, String subPop, String prefix) {
        THashMap dataPointFeatureMap = dataPoint.getFeatures();
        Iterator dataPointFeatIter = dataPointFeatureMap.keySet().iterator();
        String retString = prefix + " " + subPop + " " + dimValue + " ";
        boolean usingAdvFeatures = this.optionsModel.useAdvFeatures();
        // Instead of iterating over the entire feature set, just iterate
        // over the set of features in the current data point
        if (this.optionsModel.isOutputFreqCounts()) {
          while (dataPointFeatIter.hasNext()) {
              String feature = dataPointFeatIter.next().toString();
              if (usingAdvFeatures)
              	feature = feature.replaceFirst(".*?_", "").replaceFirst(".*?_", "");
              if (this.featureMap.containsKey(feature)) {
                retString += feature + "=" + (Integer) dataPointFeatureMap.get(feature) + " ";
              }
          }
        } else {
          while (dataPointFeatIter.hasNext()) {
            String feature = dataPointFeatIter.next().toString();
            if (usingAdvFeatures)
            	feature = feature.replaceFirst(".*?_", "").replaceFirst(".*?_", "");
            if (this.featureMap.containsKey(feature)) {
              retString += feature + "=" + 1.0 + " ";
            }
          }
        }
        
        if (this.optionsModel.isLineLength()) {
        	double linelength = this.getLineLength(dataPoint.getText());
            retString += "LINE_LENGTH=" + linelength;//((double) (num / denom));
        }

        // fill in the values for the nominal features in the dataset
        for (int i = 0; i < DataPoint.nominalFeatureNames.size(); i++) {
            String nominalAtt = (String) DataPoint.nominalFeatureNames.get(i);
            if (!((String) dataPoint.getNominalFeatures().get(i)).equals(DataPoint.MISSING_NOMINAL_VALUE)) {
            	// we might actually cause an exception here, if the nominal
            	// attribute is not in the dummy ARFF header
            	if (dummy.attribute(nominalAtt).type() == Attribute.NOMINAL)
                retString +=  " " + nominalAtt + "_" + (String) dataPoint.getNominalFeatures().get(i);
            	else
            		retString +=  " " + nominalAtt + "=" + (String) dataPoint.getNominalFeatures().get(i);
            } 
        }
        
        // use of advanced features
        if (this.optionsModel.useAdvFeatures()){        
	        Iterator itr = this.getGlobalUserDefinedFeatureSetIterator();
	        while(itr.hasNext())
	        {
	        	String advAtt = itr.next().toString();
	          if(dataPoint.getFeatureSpaceMap().getSynFeature(advAtt)!=null)
	          	retString += " " + advAtt + "=1";
	        }
        }        

        return retString;
    }

    /**
     * Method to create one instance in GibbsLDA++ format
     * (just spitting out the features)
     * 
     * @param dataPoint
     * @param dummy The ARFF format header for the feature set
     * 				(required for checking presence of numeric attributes
     * 				in the nominal set of features)
     * @return
     */
    public String getGibbsLDAPlusPlusInstanceFromDataPoint(DataPoint dataPoint, Instances dummy) {
        THashMap dataPointFeatureMap = dataPoint.getFeatures();
        Iterator dataPointFeatIter = dataPointFeatureMap.keySet().iterator();
        String gibbsLDAStr = "";
        boolean usingAdvFeatures = this.optionsModel.useAdvFeatures();
        // Instead of iterating over the entire feature set, just iterate
        // over the set of features in the current data point
        if (this.optionsModel.isOutputFreqCounts()) {
          while (dataPointFeatIter.hasNext()) {
              String feature = dataPointFeatIter.next().toString();
              if (usingAdvFeatures)
              	feature = feature.replaceFirst(".*?_", "").replaceFirst(".*?_", "");
              if (this.featureMap.containsKey(feature)) {
                gibbsLDAStr += feature + " ";
              }
          }
        } else {
          while (dataPointFeatIter.hasNext()) {
            String feature = dataPointFeatIter.next().toString();
            if (usingAdvFeatures)
            	feature = feature.replaceFirst(".*?_", "").replaceFirst(".*?_", "");
            if (this.featureMap.containsKey(feature)) {
              gibbsLDAStr += feature + " ";
            }
          }
        }
        
        // fill in the values for the nominal features in the dataset
        for (int i = 0; i < DataPoint.nominalFeatureNames.size(); i++) {
            String nominalAtt = (String) DataPoint.nominalFeatureNames.get(i);
            if (!((String) dataPoint.getNominalFeatures().get(i)).equals(DataPoint.MISSING_NOMINAL_VALUE)) {
            	// we might actually cause an exception here, if the nominal
            	// attribute is not in the dummy ARFF header
            	if (dummy.attribute(nominalAtt).type() == Attribute.NOMINAL)
                gibbsLDAStr +=  nominalAtt + "_" + (String) dataPoint.getNominalFeatures().get(i) + " ";
            } 
        }
        
        // use of advanced features
        if (this.optionsModel.useAdvFeatures()){        
	        Iterator itr = this.getGlobalUserDefinedFeatureSetIterator();
	        while(itr.hasNext())
	        {
	        	String advAtt = itr.next().toString();
	          if(dataPoint.getFeatureSpaceMap().getSynFeature(advAtt)!=null)
	          	gibbsLDAStr += advAtt + " ";
	        }
        }        

        return gibbsLDAStr;
    }

    
    /**
     * Method to create one instance in SVM Light / HMM SVM Light format
     * 
     * @param dataPoint
     * @param dummy The ARFF format header for the feature set
     * 				(required for checking presence of numeric attributes
     * 				in the nominal set of features)
     * @param includeSubpopInfo whether to output the sub-population identifier
     * 				(in the form of a qid:x.y field immediately after the class value)
     * @return
     */
    public String getSVMLightInstanceFromDataPoint(DataPoint dataPoint, Instances dummy, 
    	boolean includeSubpopInfo, THashMap subPopHash, String dimValue) {
        THashMap dataPointFeatureMap = dataPoint.getFeatures();
        Iterator dataPointFeatIter = dataPointFeatureMap.keySet().iterator();
        String svmLightRep = "";
        if (!dimValue.equals(""))
        	svmLightRep = dimValue + " ";
        
        if (includeSubpopInfo) {
        	Integer seqNum = new Integer(Integer.MIN_VALUE);
        	String subPop = dataPoint.getSubPopId();
          if (subPopHash.containsKey(subPop)) {
          	seqNum = (Integer) subPopHash.get(subPop) + 1;
          	subPopHash.put(subPop, seqNum);
          	svmLightRep += "qid:" + subPop + "." + seqNum + " ";
          } else {
          	seqNum = 1;
          	subPopHash.put(subPop, seqNum);
          	svmLightRep += "qid:" + subPop + "." + seqNum + " ";
          }
        }
        
        TreeMap<Integer, Double> featMap = new TreeMap<Integer, Double>();
        boolean usingAdvFeatures = this.optionsModel.useAdvFeatures();
        // Instead of iterating over the entire feature set, just iterate
        // over the set of features in the current data point
        if (this.optionsModel.isOutputFreqCounts()) {
          while (dataPointFeatIter.hasNext()) {
              String feature = dataPointFeatIter.next().toString();
              if (usingAdvFeatures)
              	feature = feature.replaceFirst(".*?_", "").replaceFirst(".*?_", "");
              if (this.featureMap.containsKey(feature)) {
            	Integer myint = new Integer(dummy.attribute(feature).index()+1);
            	Double mydoub = new Double((Integer) dataPointFeatureMap.get(feature));
              	featMap.put(myint, mydoub);
              }
          }
        } else {
          while (dataPointFeatIter.hasNext()) {
            String feature = dataPointFeatIter.next().toString();
            if (usingAdvFeatures)
            	feature = feature.replaceFirst(".*?_", "").replaceFirst(".*?_", "");
            if (this.featureMap.containsKey(feature)) {
            	featMap.put(new Integer(dummy.attribute(feature).index()+1), new Double(1.0));
            }
          }
        }
        
        if (this.optionsModel.isLineLength()) {
        	double linelength = this.getLineLength(dataPoint.getText());
        	featMap.put(new Integer(dummy.attribute(FeatureSelector.LINE_LENGTH_FLAG).index()+1), 
        			new Double(linelength));
        }

        // fill in the values for the nominal features in the dataset
        for (int i = 0; i < DataPoint.nominalFeatureNames.size(); i++) {
            String nominalAtt = (String) DataPoint.nominalFeatureNames.get(i);
            if (!((String) dataPoint.getNominalFeatures().get(i)).equals(DataPoint.MISSING_NOMINAL_VALUE)) {
            	// we might actually cause an exception here, if the nominal
            	// attribute is not in the dummy ARFF header
            	//
            	// also, we have a problem here with nominal attributes, for now
            	// we are simply skipping them
            	//
            	// in MinorThird format, we simply make them binary features with
            	// composite names that are of the form nominal-name_nominal-value
            	// but, in svm-light format, we would need to assign them indices
            	// in the feature vector and accordingly adjust indices of subsequent
            	// nominal or numeric attributes. this is kept as a TODO item.
            	if (dummy.attribute(nominalAtt).type() == Attribute.NUMERIC)
            		featMap.put(new Integer(dummy.attribute(nominalAtt).index()+1), 
            				new Double(Double.parseDouble((String) dataPoint.getNominalFeatures().get(i))));
            } 
        }
        
        // use of advanced features
        if (this.optionsModel.useAdvFeatures()){        
	        Iterator itr = this.getGlobalUserDefinedFeatureSetIterator();
	        while(itr.hasNext())
	        {
	        	String advAtt = itr.next().toString();
	          if(dataPoint.getFeatureSpaceMap().getSynFeature(advAtt)!=null)
	          	featMap.put(new Integer(dummy.attribute(advAtt).index()+1),
	          			new Double(1.0));
	        }
        }        

        Iterator itr = featMap.keySet().iterator();
        while (itr.hasNext()) {
        	Integer i = (Integer) itr.next();
        	Double d = (Double) featMap.get(i);
        	svmLightRep += "" + i + ":" + d + " ";
        }
        
        return svmLightRep;
    }
    
    public int getFeatureSetCount() {
        return this.featureMap.size();
    }


    private int countWords(String text) {
        return Tokenizer.getInstance(this.optionsModel).tokenizeIncStopword(text).getNumTokens();
    }
    
    public double getLineLength(String text)
    {
    	if(this.maxCount ==Integer.MIN_VALUE || this.minCount == Integer.MAX_VALUE )
    	{
    		return (double)this.countWords(text);
    	}
    	else
    	{
            int wordCount = this.countWords(text);
            double num = (double) (wordCount - this.minCount);
            double denom = (double) (this.maxCount - this.minCount);
            return (double) (num / denom);
    	}
    }

    /**
     * Note: 08/14/2007, Mahesh Joshi.
     * 
     * This method is used by <code>SimpleFeatureSpaceBuilder</code> class, which is 
     * not used by any other current TagHelperTools2 class as of now. Actually the
     * signature of this should be changed to return a THashMap instead of a THashSet
     * since we are moving to storing frequency counts of features for each instance.
     * 
     * @param text
     * @param featureSet
     * @return
     */
    public THashSet getTestPointFeatureSet(String text, THashSet featureSet) {

        text = text.trim();
        THashSet dataPointFeatureSet = new THashSet();

        // UNIGRAMS - Extract unigrams depending on (1) stemming and
        // (2) stopword removal.
        if (optionsModel.isUnigrams()) {
            Tokens tokens = null;
            if (optionsModel.isRemoveStopwords()) {
                if (optionsModel.isStem()) {
                    tokens = Tokenizer.getInstance(this.optionsModel).lemmatizeExcStopword(text.toLowerCase());
                } else {
                    tokens = Tokenizer.getInstance(this.optionsModel).tokenizeExcStopword(text.toLowerCase());
                }
            } else {
                if (optionsModel.isStem()) {
                    tokens = Tokenizer.getInstance(this.optionsModel).lemmatizeIncStopword(text.toLowerCase());
                } else {
                    tokens = Tokenizer.getInstance(this.optionsModel).tokenizeIncStopword(text.toLowerCase());
                }
            }
            Iterator tokenIter = tokens.getIterator();
            while (tokenIter.hasNext()) {
                Token token = (Token) tokenIter.next();
                if (!token.getIndexToken().equals(Token.EMPTY)) {
                    if (featureSet.contains(token.getIndexToken())) {
                        dataPointFeatureSet.add(token.getIndexToken());
                    }
                }
            }
        }

        // BIGRAMS - Extract bigrams depending on (1) stemming and (2) stopword
        // removal.
        if (optionsModel.isBigrams()) {
            Tokens tokens = null;
            if (optionsModel.isRemoveStopwords()) {
                if (optionsModel.isStem()) {
                    tokens = Tokenizer.getInstance(this.optionsModel).lemmatizeBigramsExcStopwords(text.toLowerCase());
                } else {
                    tokens = Tokenizer.getInstance(this.optionsModel).tokenizeBigramsExcStopwords(text.toLowerCase());
                }
            } else {
                if (optionsModel.isStem()) {
                    tokens = Tokenizer.getInstance(this.optionsModel).lemmatizeBigramsIncStopwords(text.toLowerCase());
                } else {
                    tokens = Tokenizer.getInstance(this.optionsModel).tokenizeBigramsIncStopwords(text.toLowerCase());
                }
            }
            Iterator tokenIter = tokens.getIterator();
            while (tokenIter.hasNext()) {
                Token token = (Token) tokenIter.next();
                if (!token.getIndexToken().equals(Token.EMPTY)) {
                    if (featureSet.contains(token.getIndexToken())) {
                        dataPointFeatureSet.add(token.getIndexToken());
                    }
                }
            }
        }

        // POS BIGRAMS
        if (optionsModel.isPosBigrams()) {
          String newText = HelperFunctions.replaceSpecialTokensWithPunc(text, optionsModel.getLanguage());
        	Tokens tokens = Tokenizer.getInstance(this.optionsModel).getPOSBigrams(newText);
            Iterator tokenIter = tokens.getIterator();
            while (tokenIter.hasNext()) {
                Token token = (Token) tokenIter.next();
                if (!token.getIndexToken().equals(Token.EMPTY)) {
                	String str = HelperFunctions.replacePuncWithSpecialTokens(token.getIndexToken(), 
                			this.optionsModel.getLanguage(), false, true);
                    if (featureSet.contains(str)) {
                        dataPointFeatureSet.add(str);
                    }
                }
            }
        }

        // PUNCTUATION - Extract Punctuation. The last character in the sentence
        // is checked against the inventory of known punctuation marks, in
        // punctStringMap
        if (optionsModel.isPuctuation()) {
            if (text.length() > 0) {
                char lastChar = text.charAt(text.length() - 1);
                if (!Character.isLetterOrDigit(lastChar)) {
                    if (FeatureSelector.punctStringMap.containsKey(new Character(lastChar).hashCode())) {
                        if (featureSet.contains(punctStringMap.get(new Character(lastChar).hashCode()))) {
                            dataPointFeatureSet.add(punctStringMap.get(new Character(lastChar).hashCode()));
                        }
                    }
                }
            }
        }

        // CONTAINS NON_STOPWORD - this is true iff any of the words in the text
        // qualifies as a non-stopwords, based on the list of stopwords in the
        // default.stp file in the ROOT/etc directory. See StopwordManager.
        if (optionsModel.isContainsNonStopword()) 
        {
            Tokens tokens = Tokenizer.getInstance(this.optionsModel)
            						 .tokenizeIncStopword(text.toLowerCase());
            Iterator tokenIter = tokens.getIterator();
            boolean containsNonStopword = false;
            while (tokenIter.hasNext()) {
                Token token = (Token) tokenIter.next();
                if (!StopwordManager.getInstance(this.optionsModel).isStopword(token.getIndexToken()) /*&& 
                								 featureSet.contains(token.getIndexToken())*/) {
                    containsNonStopword = true;
                    break;
                }
            }
            if (containsNonStopword) {
                dataPointFeatureSet.add(FeatureSelector.CONTAINS_NON_STOPWORD_TRUE);
            } else {
                dataPointFeatureSet.add(FeatureSelector.CONTAINS_NON_STOPWORD_FALSE);
            }
        }
        
        // CONTAINS QUESTION_WORD - this is true iff the first word in the text
        // qualifies as a question word, based on the list of question words
        // hard-coded at the top
//        Tokens tokens = Tokenizer.getInstance().tokenizeIncStopword(
//                text.toLowerCase());
//        Iterator tokenIter = tokens.getIterator();
//        boolean containsQuestionWord = false;
//        while (tokenIter.hasNext()) {
//            Token token = (Token) tokenIter.next();
//            if (questionWords.containsKey(token.getIndexToken())) {
//                containsQuestionWord = true;
//            }
//            break;
//        }
//        if (containsQuestionWord) {
//            dataPointFeatureSet.add(FeatureSelector.CONTAINS_QUESTION_WORD_TRUE);
//        } else {
//            dataPointFeatureSet
//                    .add(FeatureSelector.CONTAINS_QUESTION_WORD_FALSE);
//        }

        return dataPointFeatureSet;
    }
    
    /**
     * Note: 08/14/2007, Mahesh Joshi.
     * Actually the
     * signature of this should be changed to return a THashMap instead of a THashSet
     * since we are moving to storing frequency counts of features for each instance.
     * 
     * Not used by any of the TagHelperTools2 code as of now.
     * 
     * @deprecated 
     * 
     * @param text
     * @param featureSet
     * @param fsm
     * @return
     */
    public THashSet getTestPointFeatureSet(String text, TreeSet featureSet, FeatureSpaceMap fsm) {

        text = text.trim();
        THashSet dataPointFeatureSet = new THashSet();

        // UNIGRAMS - Extract unigrams depending on (1) stemming and
        // (2) stopword removal.
        if (optionsModel.isUnigrams()) {
            Tokens tokens = null;
            if (optionsModel.isRemoveStopwords()) {
                if (optionsModel.isStem()) {
                    tokens = Tokenizer.getInstance(this.optionsModel).lemmatizeExcStopword(text.toLowerCase(), fsm);
                } else {
                    tokens = Tokenizer.getInstance(this.optionsModel).tokenizeExcStopword(text.toLowerCase(), fsm);
                }
            } else {
                if (optionsModel.isStem()) {
                    tokens = Tokenizer.getInstance(this.optionsModel).lemmatizeIncStopword(text.toLowerCase(), fsm);
                } else {
                    tokens = Tokenizer.getInstance(this.optionsModel).tokenizeIncStopword(text.toLowerCase(), fsm);
                }
            }
            Iterator tokenIter = tokens.getIterator();
            while (tokenIter.hasNext()) {
                Token token = (Token) tokenIter.next();
                if (!token.getIndexToken().equals(Token.EMPTY)) {
                    if (featureSet.contains(token.getIndexToken())) {
                        dataPointFeatureSet.add(token.getIndexToken());
                    }
                }
            }
        }

        // BIGRAMS - Extract bigrams depending on (1) stemming and (2) stopword
        // removal.
        if (optionsModel.isBigrams()) {
            Tokens tokens = null;
            if (optionsModel.isRemoveStopwords()) {
                if (optionsModel.isStem()) {
                    tokens = Tokenizer.getInstance(this.optionsModel).lemmatizeBigramsExcStopwords(text.toLowerCase(), fsm);
                } else {
                    tokens = Tokenizer.getInstance(this.optionsModel).tokenizeBigramsExcStopwords(text.toLowerCase(), fsm);
                }
            } else {
                if (optionsModel.isStem()) {
                    tokens = Tokenizer.getInstance(this.optionsModel).lemmatizeBigramsIncStopwords(text.toLowerCase(), fsm);
                } else {
                    tokens = Tokenizer.getInstance(this.optionsModel).tokenizeBigramsIncStopwords(text.toLowerCase(), fsm);
                }
            }
            Iterator tokenIter = tokens.getIterator();
            while (tokenIter.hasNext()) {
                Token token = (Token) tokenIter.next();
                if (!token.getIndexToken().equals(Token.EMPTY)) {
                    if (featureSet.contains(token.getIndexToken())) {
                        dataPointFeatureSet.add(token.getIndexToken());
                    }
                }
            }
        }

        // POS BIGRAMS
        if (optionsModel.isPosBigrams()) {
        	String newText = HelperFunctions.replaceSpecialTokensWithPunc(text, optionsModel.getLanguage());
            Tokens tokens = Tokenizer.getInstance(this.optionsModel).getPOSBigrams(newText, fsm);
            Iterator tokenIter = tokens.getIterator();
            while (tokenIter.hasNext()) {
                Token token = (Token) tokenIter.next();
                if (!token.getIndexToken().equals(Token.EMPTY)) {
                	String str = token.getIndexToken();
                    if (featureSet.contains(str)) {
                        dataPointFeatureSet.add(str);
                    }
                }
            }
        }

        // PUNCTUATION - Extract Punctuation. The last character in the sentence
        // is checked against the inventory of known punctuation marks, in
        // punctStringMap
        if (optionsModel.isPuctuation()) {
            if (text.length() > 0) {
                char lastChar = text.charAt(text.length() - 1);
                if (!Character.isLetterOrDigit(lastChar)) {
                    if (FeatureSelector.punctStringMap.containsKey(new Character(lastChar).hashCode())) {
                    	String tmpFeature =(String)punctStringMap.get(new Character(lastChar).hashCode());
                        if (featureSet.contains(tmpFeature)) {
                            dataPointFeatureSet.add(/*punctStringMap.get(new Character(lastChar).hashCode())*/tmpFeature);
                            fsm.addOrdFeature(tmpFeature, FeatureSpaceMap.NOT_COUNTABLE, FeatureSpaceMap.NOT_COUNTABLE, false);                            
                        }
                    }
                }
            }
        }

        // CONTAINS NON_STOPWORD - this is true iff any of the words in the text
        // qualifies as a non-stopwords, based on the list of stopwords in the
        // default.stp file in the ROOT/etc directory. See StopwordManager.
        if (optionsModel.isContainsNonStopword()) {
            Tokens tokens = Tokenizer.getInstance(this.optionsModel)
            						 .tokenizeIncStopword(text.toLowerCase());
            Iterator tokenIter = tokens.getIterator();
            boolean containsNonStopword = false;
            while (tokenIter.hasNext()) {
                Token token = (Token) tokenIter.next();
                if (!StopwordManager.getInstance(this.optionsModel).isStopword(token.getIndexToken())/*&& 
                								 featureSet.contains(token.getIndexToken())*/) {
                    containsNonStopword = true;
                    break;
                }
            }
            if (containsNonStopword) {
                dataPointFeatureSet.add(FeatureSelector.CONTAINS_NON_STOPWORD_TRUE);
                fsm.addOrdFeature(FeatureSelector.CONTAINS_NON_STOPWORD_TRUE, FeatureSpaceMap.NOT_COUNTABLE, 
						  FeatureSpaceMap.NOT_COUNTABLE, false);
            } else {
                dataPointFeatureSet.add(FeatureSelector.CONTAINS_NON_STOPWORD_FALSE);
                fsm.addOrdFeature(FeatureSelector.CONTAINS_NON_STOPWORD_FALSE, FeatureSpaceMap.NOT_COUNTABLE, 
                															  FeatureSpaceMap.NOT_COUNTABLE, false);  
            }
        }
        
        // CONTAINS QUESTION_WORD - this is true iff the first word in the text
        // qualifies as a question word, based on the list of question words
        // hard-coded at the top
//        Tokens tokens = Tokenizer.getInstance().tokenizeIncStopword(
//                text.toLowerCase());
//        Iterator tokenIter = tokens.getIterator();
//        boolean containsQuestionWord = false;
//        while (tokenIter.hasNext()) {
//            Token token = (Token) tokenIter.next();
//            if (questionWords.containsKey(token.getIndexToken())) {
//                containsQuestionWord = true;
//            }
//            break;
//        }
//        if (containsQuestionWord) {
//            dataPointFeatureSet.add(FeatureSelector.CONTAINS_QUESTION_WORD_TRUE);
//        } else {
//            dataPointFeatureSet
//                    .add(FeatureSelector.CONTAINS_QUESTION_WORD_FALSE);
//        }

        return dataPointFeatureSet;
    }    
    
    /**
     * Private utility function that adds or increments frequency count of the given 
     * feature to the ordinary feature map.
     * 
     * @param dataPointFeatureMap
     * @param feature
     */
    private void addFeatToFeatureMap(THashMap dataPointFeatureMap, String feature) {
      if (!feature.equals(Token.EMPTY)) {
        if (dataPointFeatureMap.containsKey(feature)) {
        	Integer freq = (Integer) dataPointFeatureMap.get(feature);
        	freq++;
        	dataPointFeatureMap.put(feature, freq);
        } else {
        	dataPointFeatureMap.put(feature, new Integer(1));
        }
      }
    }
    
    /**
     * This method is used in a "hidden functionality mode". This is invoked when the
     * input training file is paired with a test file, the convention to detect this
     * being based on the file names. So if the train file name is xxxxx.train.csv, then
     * the test file should be xxxxx.test.csv (basically the string ".train." in the
     * training file should be replaced by ".test." in the test file name.
     * 
     * @param text
     * @return
     */
    public THashMap getTestPointFeatureSet(String text) {
    	return this.getTestPointFeatureSet(text, "");
    }
    	
    /**
     * This method is used in a "hidden functionality mode". This is invoked when the
     * input training file is paired with a test file, the convention to detect this
     * being based on the file names. So if the train file name is xxxxx.train.csv, then
     * the test file should be xxxxx.test.csv (basically the string ".train." in the
     * training file should be replaced by ".test." in the test file name.
     * 
     * @param text
     * @return
     */
    public THashMap getTestPointFeatureSet(String text, String customPrefix) {

        text = text.trim();
        THashMap dataPointFeatureMap = new THashMap();

        // UNIGRAMS - Extract unigrams depending on (1) stemming and
        // (2) stopword removal.
        if (optionsModel.isUnigrams()) {
            Tokens tokens = null;
            if (optionsModel.isRemoveStopwords()) {
                if (optionsModel.isStem()) {
                    tokens = Tokenizer.getInstance(this.optionsModel).lemmatizeExcStopword(text.toLowerCase());
                } else {
                    tokens = Tokenizer.getInstance(this.optionsModel).tokenizeExcStopword(text.toLowerCase());
                }
            } else {
                if (optionsModel.isStem()) {
                    tokens = Tokenizer.getInstance(this.optionsModel).lemmatizeIncStopword(text.toLowerCase());
                } else {
                    tokens = Tokenizer.getInstance(this.optionsModel).tokenizeIncStopword(text.toLowerCase());
                }
            }
            Iterator tokenIter = tokens.getIterator();
            while (tokenIter.hasNext()) {
              String str = ((Token) tokenIter.next()).getIndexToken();
              if (this.featureMap.contains(str)) {
              	addFeatToFeatureMap(dataPointFeatureMap, customPrefix+str);
              }
            }
        }

        // BIGRAMS - Extract bigrams depending on (1) stemming and (2) stopword
        // removal.
        if (optionsModel.isBigrams()) {
            Tokens tokens = null;
            if (optionsModel.isRemoveStopwords()) {
                if (optionsModel.isStem()) {
                    tokens = Tokenizer.getInstance(this.optionsModel).lemmatizeBigramsExcStopwords(text.toLowerCase());
                } else {
                    tokens = Tokenizer.getInstance(this.optionsModel).tokenizeBigramsExcStopwords(text.toLowerCase());
                }
            } else {
                if (optionsModel.isStem()) {
                    tokens = Tokenizer.getInstance(this.optionsModel).lemmatizeBigramsIncStopwords(text.toLowerCase());
                } else {
                    tokens = Tokenizer.getInstance(this.optionsModel).tokenizeBigramsIncStopwords(text.toLowerCase());
                }
            }
            Iterator tokenIter = tokens.getIterator();
            while (tokenIter.hasNext()) {
                String str = ((Token) tokenIter.next()).getIndexToken();
                if (this.featureMap.contains(str)) {
                	addFeatToFeatureMap(dataPointFeatureMap, customPrefix+str);
                }
            }
        }

        // POS BIGRAMS
        if (optionsModel.isPosBigrams()) {
        	String newText = HelperFunctions.replaceSpecialTokensWithPunc(text, optionsModel.getLanguage());
            Tokens tokens = Tokenizer.getInstance(this.optionsModel).getPOSBigrams(newText);
            Iterator tokenIter = tokens.getIterator();
            while (tokenIter.hasNext()) {
                Token token = (Token) tokenIter.next();
                if (!token.getIndexToken().equals(Token.EMPTY)) {
                	String str = HelperFunctions.replacePuncWithSpecialTokens(token.getIndexToken(), 
                			this.optionsModel.getLanguage(), false, true);
                    if (this.featureMap.contains(str)) {
                    	addFeatToFeatureMap(dataPointFeatureMap, customPrefix+str);
                    }
                }
            }
        }

        // PUNCTUATION - Extract Punctuation. The last character in the sentence
        // is checked against the inventory of known punctuation marks, in
        // punctStringMap
        if (optionsModel.isPuctuation()) {
            if (text.length() > 0) {
          		String tmp = text.trim();//.toLowerCase();
              String[] tmplist = tmp.split("[ \t\n\r\f]");
          
              //System.out.println("XXX: "+tmplist[tmplist.length-1]);
              if (FeatureSelector.punctStringMap.containsKey(tmplist[tmplist.length-1])) 
              {
              	if (this.featureMap.contains(tmplist[tmplist.length-1]))
              		addFeatToFeatureMap(dataPointFeatureMap, customPrefix+(String) punctStringMap.get(tmplist[tmplist.length-1]));
              }
          }
        }

        // CONTAINS NON_STOPWORD - this is true iff any of the words in the text
        // qualifies as a non-stopwords, based on the list of stopwords in the
        // default.stp file in the ROOT/etc directory. See StopwordManager.
        if (optionsModel.isContainsNonStopword()) {
            Tokens tokens = Tokenizer.getInstance(this.optionsModel)
            						 .tokenizeIncStopword(text.toLowerCase());
            Iterator tokenIter = tokens.getIterator();
            boolean containsNonStopword = false;
            while (tokenIter.hasNext()) {
                Token token = (Token) tokenIter.next();
                if (!StopwordManager.getInstance(this.optionsModel).isStopword(token.getIndexToken()) /*&& 
                								this.featureMap.contains(token.getIndexToken())*/) {
                    containsNonStopword = true;
                    break;
                }
            }
            if (containsNonStopword) {
            	addFeatToFeatureMap(dataPointFeatureMap, customPrefix+FeatureSelector.CONTAINS_NON_STOPWORD_TRUE);
            } else {
            	addFeatToFeatureMap(dataPointFeatureMap, customPrefix+FeatureSelector.CONTAINS_NON_STOPWORD_FALSE);
            }
        }
        
        return dataPointFeatureMap;
    }

    /**
     * This method is used when no advanced features are specified. It
     * checks the static variables of the FeatureSelector, which are, in turn,
     * set by the OptionsModel of the classifier being trained. The only feature
     * that is not extracted in this method is sentence length. This method
     * extracts all the features that have a string value, and sentence length
     * has a numeric value.
     * 
     * @param text -
     *            The text of the current DataPoint for which features are being
     *            extracted.
     * @return - The set of features extracted from this text.
     */
    public THashMap getFeatureSet(String text) {
    	return this.getFeatureSet(text, "");
    }
    
    /**
     * This method is used when no advanced features are specified. It
     * checks the static variables of the FeatureSelector, which are, in turn,
     * set by the OptionsModel of the classifier being trained. The only feature
     * that is not extracted in this method is sentence length. This method
     * extracts all the features that have a string value, and sentence length
     * has a numeric value.
     * 
     * @param text -
     *            The text of the current DataPoint for which features are being
     *            extracted.
     * @return - The set of features extracted from this text.
     */
    public THashMap getFeatureSet(String text, String customPrefix) {

        text = text.trim();
        THashMap dataPointFeatureMap = new THashMap();

        // UNIGRAMS - Extract unigrams depending on (1) stemming and
        // (2) stopword removal.
        if (optionsModel.isUnigrams()) {
            Tokens tokens = null;
            if (optionsModel.isRemoveStopwords()) {
                if (optionsModel.isStem()) {
                    tokens = Tokenizer.getInstance(this.optionsModel).lemmatizeExcStopword(text.toLowerCase());
                } else {
                    tokens = Tokenizer.getInstance(this.optionsModel).tokenizeExcStopword(text.toLowerCase());
                }
            } 
            else 
            {
                if (optionsModel.isStem()) {
                    tokens = Tokenizer.getInstance(this.optionsModel).lemmatizeIncStopword(text.toLowerCase());
                } else {
                    tokens = Tokenizer.getInstance(this.optionsModel).tokenizeIncStopword(text.toLowerCase());
                }
            }
            Iterator tokenIter = tokens.getIterator();
            while (tokenIter.hasNext()) {
               addFeatToFeatureMap(dataPointFeatureMap, customPrefix+((Token) tokenIter.next()).getIndexToken());
            }
        }

        // BIGRAMS - Extract bigrams depending on (1) stemming and (2) stopword
        // removal.
        if (optionsModel.isBigrams()) {
            Tokens tokens = null;
            if (optionsModel.isRemoveStopwords()) {
                if (optionsModel.isStem()) {
                    tokens = Tokenizer.getInstance(this.optionsModel).lemmatizeBigramsExcStopwords(text.toLowerCase());
                } else {
                    tokens = Tokenizer.getInstance(this.optionsModel).tokenizeBigramsExcStopwords(text.toLowerCase());
                }
            } else {
                if (optionsModel.isStem()) {
                    tokens = Tokenizer.getInstance(this.optionsModel).lemmatizeBigramsIncStopwords(text.toLowerCase());
                } else {
                    tokens = Tokenizer.getInstance(this.optionsModel).tokenizeBigramsIncStopwords(text.toLowerCase());
                }
            }
            Iterator tokenIter = tokens.getIterator();
            while (tokenIter.hasNext()) {
            	addFeatToFeatureMap(dataPointFeatureMap, customPrefix+((Token) tokenIter.next()).getIndexToken());
            }
        }

        // POS BIGRAMS
        if (optionsModel.isPosBigrams()) {
        	// re-introduce punctuation, which is required for better POS-tagging
        	// (getPOSBigrams internally does sentence segmentation of the input text)
        	String newText = HelperFunctions.replaceSpecialTokensWithPunc(text, optionsModel.getLanguage());
            Tokens tokens = Tokenizer.getInstance(this.optionsModel).getPOSBigrams(newText);
            Iterator tokenIter = tokens.getIterator();
            while (tokenIter.hasNext()) {
                Token token = (Token) tokenIter.next();
                if (!token.getIndexToken().equals(Token.EMPTY)) {
                	String str = HelperFunctions.replacePuncWithSpecialTokens(token.getIndexToken(), 
                			this.optionsModel.getLanguage(), false, true);
                	addFeatToFeatureMap(dataPointFeatureMap, customPrefix+str);
                }
            }
        }

        // PUNCTUATION - Extract Punctuation. The last character in the sentence
        // is checked against the inventory of known punctuation marks, in
        // punctStringMap
        if (optionsModel.isPuctuation()) {
        	if (text.length() > 0)
        	{
        		String tmp = text.trim();//.toLowerCase();
                String[] tmplist = tmp.split("[ \t\n\r\f]");
                //System.out.println("FeatureSelector::getFeatureSet: "+tmp);
                
                for(String token:tmplist)
                {
                    if (FeatureSelector.punctStringMap.containsKey(token)) 
                    {
                    	addFeatToFeatureMap(dataPointFeatureMap, customPrefix+(String) punctStringMap.get(token));
                    }

                }
               
//                if (FeatureSelector.punctStringMap.containsKey(tmplist[tmplist.length-1])) 
//                {
//                	addFeatToFeatureMap(dataPointFeatureMap, customPrefix+(String) punctStringMap.get(tmplist[tmplist.length-1]));
//                }
            }
        }
        // CONTAINS NON_STOPWORD - this is true iff any of the words in the text
        // qualifies as a non-stopwords, based on the list of stopwords in the
        // default.stp file in the ROOT directory/etc. See StopwordManager.
        if (optionsModel.isContainsNonStopword()) {
            Tokens tokens = Tokenizer.getInstance(this.optionsModel).tokenizeIncStopword(text.toLowerCase());
            Iterator tokenIter = tokens.getIterator();
            boolean containsNonStopword = false;
            while (tokenIter.hasNext()) {
                Token token = (Token) tokenIter.next();
                if (!StopwordManager.getInstance(this.optionsModel).isStopword(token.getIndexToken())) {
                    containsNonStopword = true;
                    break;
                }
            }
            if (containsNonStopword) {
            	addFeatToFeatureMap(dataPointFeatureMap, customPrefix+FeatureSelector.CONTAINS_NON_STOPWORD_TRUE);
            } else {
            	addFeatToFeatureMap(dataPointFeatureMap, customPrefix+FeatureSelector.CONTAINS_NON_STOPWORD_FALSE);
            }
        }

        return dataPointFeatureMap;   	
    }

    /**
     *  This method is used when advanced features are requested.
     *  
     * @param text
     * @param fsm
     * @param prefixfsm
     * @return
     */
    public THashMap getFeatureSet(String text, FeatureSpaceMap fsm, 
    		FeatureSpaceMap prefixfsm) {
    	return this.getFeatureSet(text, fsm, prefixfsm, "");
    }
    
    /**
     *  This method is used when advanced features are requested.
     *  
     * @param text
     * @param fsm
     * @param prefixfsm
     * @return
     */
    public THashMap getFeatureSet(String text, FeatureSpaceMap fsm, 
    		FeatureSpaceMap prefixfsm, String customPrefix) {

        text = text.trim();
        THashMap dataPointFeatureMap = new THashMap();
        // UNIGRAMS - Extract unigrams depending on (1) stemming and
        // (2) stopword removal.
        {
        	//TODO: add prefix before each token, for adv feature editing 
        	Tokens tokens = null;
        	Iterator tokenIter = null;
        
        tokens = Tokenizer.getInstance(this.optionsModel).lemmatizeExcStopword(text.toLowerCase(), fsm);
        tokenIter = tokens.getIterator();
        while (tokenIter.hasNext()) {
            String str = ((Token) tokenIter.next()).getIndexToken();
            if (!str.equals(Token.EMPTY)) {
            	//dataPointFeatureSet.add("THF_UStemRmstopw_"+token.getIndexToken());
            	addFeatToFeatureMap(dataPointFeatureMap, "THF_UStemRmstopw_"+customPrefix+str);
            	prefixfsm.addOrdFeature("THF_UStemRmstopw_"+customPrefix+str, 0, 1, false);
            }
        }
        
        
        tokens = null;
        tokenIter = null;
        tokens = Tokenizer.getInstance(this.optionsModel).tokenizeExcStopword(text.toLowerCase(), fsm);
        tokenIter = tokens.getIterator();
        while (tokenIter.hasNext()) {
            String str = ((Token) tokenIter.next()).getIndexToken();
            if (!str.equals(Token.EMPTY)) {
            	addFeatToFeatureMap(dataPointFeatureMap, "THF_UNostemRmstopw_"+customPrefix+str);
            	prefixfsm.addOrdFeature("THF_UNostemRmstopw_"+customPrefix+str, 0, 1, false);
            }
        }
        
        
        tokens = null;
        tokenIter = null;
        tokens = Tokenizer.getInstance(this.optionsModel).lemmatizeIncStopword(text.toLowerCase(), fsm);
        tokenIter = tokens.getIterator();
        while (tokenIter.hasNext()) {
            String str = ((Token) tokenIter.next()).getIndexToken();
            if (!str.equals(Token.EMPTY)) {
            	addFeatToFeatureMap(dataPointFeatureMap, "THF_UStemNormstopw_"+customPrefix+str);
            	prefixfsm.addOrdFeature("THF_UStemNormstopw_"+customPrefix+str, 0, 1, false);
            }
        }

        tokens = null;
        tokenIter = null;
        tokens = Tokenizer.getInstance(this.optionsModel).tokenizeIncStopword(text.toLowerCase(), fsm);
        tokenIter = tokens.getIterator();
        while (tokenIter.hasNext()) {
            String str = ((Token) tokenIter.next()).getIndexToken();
            if (!str.equals(Token.EMPTY)) {
            	addFeatToFeatureMap(dataPointFeatureMap, "THF_UNostemNormstopw_"+customPrefix+str);
            	prefixfsm.addOrdFeature("THF_UNostemNormstopw_"+customPrefix+str, 0, 1, false);
            }
        }
            
         

        // BIGRAMS - Extract bigrams depending on (1) stemming and (2) stopword
        // removal.
        tokens = null;
        tokenIter = null;
        tokens = Tokenizer.getInstance(this.optionsModel).lemmatizeBigramsExcStopwords(text.toLowerCase(), fsm);
        tokenIter = tokens.getIterator();
        while (tokenIter.hasNext()) {
            String str = ((Token) tokenIter.next()).getIndexToken();
            if (!str.equals(Token.EMPTY)) {
            	addFeatToFeatureMap(dataPointFeatureMap, "THF_BStemRmstopw_"+customPrefix+str);
            	prefixfsm.addOrdFeature("THF_BStemRmstopw_"+customPrefix+str, 0, 2, false);
            }
        }
        
        
        tokens = null;
        tokenIter = null;
        tokens = Tokenizer.getInstance(this.optionsModel).tokenizeBigramsExcStopwords(text.toLowerCase(),fsm);
        tokenIter = tokens.getIterator();
        while (tokenIter.hasNext()) {
            String str = ((Token) tokenIter.next()).getIndexToken();
            if (!str.equals(Token.EMPTY)) {
            	addFeatToFeatureMap(dataPointFeatureMap, "THF_BNostemRmstopw_"+customPrefix+str);
            	prefixfsm.addOrdFeature("THF_BNostemRmstopw_"+customPrefix+str, 0, 2, false);
            }
        }
        
        
        tokens = null;
        tokenIter = null;
        tokens = Tokenizer.getInstance(this.optionsModel).lemmatizeBigramsIncStopwords(text.toLowerCase(), fsm);
        tokenIter = tokens.getIterator();
        while (tokenIter.hasNext()) {
        	String str = ((Token) tokenIter.next()).getIndexToken();
            if (!str.equals(Token.EMPTY)) {
            	addFeatToFeatureMap(dataPointFeatureMap, "THF_BStemNormstopw_"+customPrefix+str);
            	prefixfsm.addOrdFeature("THF_BStemNormstopw_"+customPrefix+str, 0, 2, false);
            }
        }

        tokens = null;
        tokenIter = null;
        tokens = Tokenizer.getInstance(this.optionsModel).tokenizeBigramsIncStopwords(text.toLowerCase(), fsm);
        tokenIter = tokens.getIterator();
        while (tokenIter.hasNext()) {
        	String str = ((Token) tokenIter.next()).getIndexToken();
            if (!str.equals(Token.EMPTY)) {
            	addFeatToFeatureMap(dataPointFeatureMap, "THF_BNostemNormstopw_"+customPrefix+str);
            	prefixfsm.addOrdFeature("THF_BNostemNormstopw_"+customPrefix+str, 0, 2, false);
            }
        }
        
        
        
        // POS BIGRAMS
        tokens = null;
        tokenIter = null;
        setTaggedText(null);
        String newText = HelperFunctions.replaceSpecialTokensWithPunc(text, this.optionsModel.getLanguage());
        tokens = Tokenizer.getInstance(this.optionsModel).getPOSBigrams(newText,fsm);
        setTaggedText( AbstractTokenizer.taggedResult);
        //System.out.println("FS "+AbstractTokenizer.taggedResult);
        tokenIter = tokens.getIterator();
            while (tokenIter.hasNext()) {
            	String str = ((Token) tokenIter.next()).getIndexToken();
                if (!str.equals(Token.EMPTY)) {
                	str = HelperFunctions.replacePuncWithSpecialTokens(str, 
                			this.optionsModel.getLanguage(), false, true);
                    //dataPointFeatureSet.add("THF_POS_"+token.getIndexToken());
                	addFeatToFeatureMap(dataPointFeatureMap, "THF_POS_"+customPrefix+str);
                  prefixfsm.addOrdFeature("THF_POS_"+customPrefix+str, 0, 2, true);}
                	
            }
       
    	
        }
        // PUNCTUATION - Extract Punctuation. The last character in the sentence
        // is checked against the inventory of known punctuation marks, in
        // punctStringMap
//      //  if (optionsModel.isPuctuation()) {
            	if (text.length() > 0)
            	{
            		String tmp = text.trim();//.toLowerCase();
                    String[] tmplist = tmp.split("[ \t\n\r\f]");
                    //System.out.println("XXX: "+tmplist[tmplist.length-1]);
                    if (FeatureSelector.punctStringMap.containsKey(tmplist[tmplist.length-1])) 
                    {
                    	String tmpFeature = (String)FeatureSelector.punctStringMap.get(tmplist[tmplist.length-1]);
                    	addFeatToFeatureMap(dataPointFeatureMap, "THF_PUN_"+customPrefix+tmpFeature);
                    	fsm.addOrdFeature(customPrefix+tmpFeature, FeatureSpaceMap.NOT_COUNTABLE, FeatureSpaceMap.NOT_COUNTABLE, false);
                      prefixfsm.addOrdFeature("THF_PUN_"+customPrefix+tmpFeature, FeatureSpaceMap.NOT_COUNTABLE, FeatureSpaceMap.NOT_COUNTABLE, false);
                    }
                    
                }
//            }
        // CONTAINS NON_STOPWORD - this is true iff any of the words in the text
        // qualifies as a non-stopwords, based on the list of stopwords in the
        // default.stp file in the ROOT directory/etc. See StopwordManager.
        //if (optionsModel.isContainsNonStopword()) {
            Tokens tokens = Tokenizer.getInstance(this.optionsModel).tokenizeIncStopword(text.toLowerCase());
            Iterator tokenIter = tokens.getIterator();
            boolean containsNonStopword = false;
            while (tokenIter.hasNext()) {
                Token token = (Token) tokenIter.next();
                if (!StopwordManager.getInstance(this.optionsModel).isStopword(token.getIndexToken())) {
                    containsNonStopword = true;
                    break;
                }
            }
            if (containsNonStopword) {
            	addFeatToFeatureMap(dataPointFeatureMap, "THF_ConStop_"+customPrefix+FeatureSelector.CONTAINS_NON_STOPWORD_TRUE);
              fsm.addOrdFeature(customPrefix+FeatureSelector.CONTAINS_NON_STOPWORD_TRUE, FeatureSpaceMap.NOT_COUNTABLE, 
                															  FeatureSpaceMap.NOT_COUNTABLE, false);                
              prefixfsm.addOrdFeature("THF_ConStop_"+customPrefix+FeatureSelector.CONTAINS_NON_STOPWORD_TRUE, FeatureSpaceMap.NOT_COUNTABLE, 
						  FeatureSpaceMap.NOT_COUNTABLE, false);
            } else {
            	addFeatToFeatureMap(dataPointFeatureMap, "THF_ConStop_"+customPrefix+FeatureSelector.CONTAINS_NON_STOPWORD_FALSE);
              fsm.addOrdFeature(customPrefix+FeatureSelector.CONTAINS_NON_STOPWORD_FALSE, FeatureSpaceMap.NOT_COUNTABLE, 
                															  FeatureSpaceMap.NOT_COUNTABLE, false);                 
              prefixfsm.addOrdFeature("THF_ConStop_"+customPrefix+FeatureSelector.CONTAINS_NON_STOPWORD_FALSE, FeatureSpaceMap.NOT_COUNTABLE, 
						  FeatureSpaceMap.NOT_COUNTABLE, false);
            }
        //}
        
        return dataPointFeatureMap;
    }

    public void readDicFile(File dicFile) {
        this.optionsModel = new OptionsModel();
        this.optionsModel.setFeatureMap(this.featureMap);
        boolean onSettings = false;
        boolean onDictionary = false;
        boolean onLabels = false;

        Reader reader = new Reader(dicFile.getAbsolutePath());

        String line = "";
        int lineNum = 1;
        while ((line = reader.readLine()) != null) {
            if (line.equals(FeatureSelector.SETTINGS_FLAG)) {
                onSettings = true;
                continue;
            } else if (line.equals(FeatureSelector.DICTIONARY_FLAG)) {
                onSettings = false;
                onDictionary = true;
                continue;
            } else if (line.equals(FeatureSelector.LABEL_FLAG)) {
                onSettings = false;
                onDictionary = false;
                onLabels = true;
                continue;
            }
            if (onSettings) {
                if (line.indexOf("=") > -1) {
                    String[] split = line.split("=");
                    String setting = split[0].trim().toLowerCase();
                    boolean flag = Boolean.parseBoolean(split[1].trim().toLowerCase());
                    if (setting.equals(FeatureSelector.UNIGRAM_FLAG)) {
                        this.optionsModel.setUnigrams(flag);
                    } else if (setting.equals(FeatureSelector.BIGRAM_FLAG)) {
                        optionsModel.setBigrams(flag);
                    } else if (setting.equals(FeatureSelector.POS_BIGRAM_FLAG)) {
                        optionsModel.setPosBigrams(flag);
                    } else if (setting.equals(FeatureSelector.PUNCT_FLAG)) {
                        optionsModel.setPunctuation(flag);
                    } else if (setting.equals(FeatureSelector.LINE_LENGTH_FLAG)) {
                        optionsModel.setLineLength(flag);
                    } else if (setting.equals(FeatureSelector.CONTAINS_NON_STOPWORD_FLAG)) {
                        optionsModel.setContainsNonStopword(flag);
                    } else if (setting.equals(FeatureSelector.STEM_FLAG)) {
                        optionsModel.setStem(flag);
                    } else if (setting.equals(FeatureSelector.REMOVE_STOPWORDS_FLAG)) {
                        optionsModel.setRemoveStopwords(flag);
                    } else if(setting.equals(FeatureSelector.REMOVE_RARE)){
                    	optionsModel.setRemoveRareFeatures(flag);                    	
                    } else if(setting.equals(FeatureSelector.REMOVE_RARE_THRESHOLD)){
                    	optionsModel.setFrequencyThreshold(Integer.parseInt(split[1].trim()));
                    } else if(setting.equals(FeatureSelector.MAX_CNT)){
                    	this.setMaxCount(Integer.parseInt(split[1].trim()));
                    } else if(setting.equals(FeatureSelector.MIN_CNT)){
                    	this.setMinCount(Integer.parseInt(split[1].trim()));
                    }                      	
                    else {
                        JOptionPane.showMessageDialog(null, "Unrecognized Setting in Dictionary File = " + dicFile.getName() + ". Line # =" + lineNum, "TERMINAL ERROR", JOptionPane.ERROR_MESSAGE);
                        System.exit(0);
                    }
                } else {
                    JOptionPane.showMessageDialog(null, "Invalid Setting in Dictionary File = " + dicFile.getName() + ". Line # =" + lineNum, "TERMINAL ERROR", JOptionPane.ERROR_MESSAGE);
                    System.exit(0);
                }
            } else if (onDictionary) {
                String term = line.trim();
                if (term.length() > 0) {
                    featureMap.put(term, new Integer(1));
                } else {
                    JOptionPane.showMessageDialog(null, "Empty Feature in Dictionary file = " + dicFile.getName() + ". Line # =" + lineNum, "TERMINAL ERROR", JOptionPane.ERROR_MESSAGE);
                    System.exit(0);
                }
            } else if (onLabels) {
                this.labels.addElement(line.trim());
            }
            lineNum++;
        }
    }

    public OptionsModel getOptionsModel() {   
        return this.optionsModel;
    }
    
    public void setOptionsModel(OptionsModel m)
    {
    	this.optionsModel = m;
    	this.optionsModel.setFeatureMap(this.featureMap);
    }

    public static void main(String args[]) {

    }

    public void writeDicFile(String filename, FastVector labels) {
        Logger logger = new Logger(filename);
        logger.writeLine(FeatureSelector.SETTINGS_FLAG);
        logger.writeLine(FeatureSelector.UNIGRAM_FLAG + "=" + optionsModel.isUnigrams());
        logger.writeLine(FeatureSelector.BIGRAM_FLAG + "=" + optionsModel.isBigrams());
        logger.writeLine(FeatureSelector.POS_BIGRAM_FLAG + "=" + optionsModel.isPosBigrams());
        logger.writeLine(FeatureSelector.PUNCT_FLAG + "=" + optionsModel.isPuctuation());
        logger.writeLine(FeatureSelector.LINE_LENGTH_FLAG + "=" + optionsModel.isLineLength());
        logger.writeLine(FeatureSelector.STEM_FLAG + "=" + optionsModel.isStem());
        logger.writeLine(FeatureSelector.REMOVE_STOPWORDS_FLAG + "=" + optionsModel.isRemoveStopwords());
        logger.writeLine(FeatureSelector.CONTAINS_NON_STOPWORD_FLAG + "=" + optionsModel.isContainsNonStopword());
        logger.writeLine(FeatureSelector.REMOVE_RARE + "=" + optionsModel.isRemoveRareFeatures());
        logger.writeLine(FeatureSelector.REMOVE_RARE_THRESHOLD + "=" + optionsModel.getFrequencyThreshold());
        logger.writeLine(FeatureSelector.MAX_CNT + "=" + this.getMaxCount());
        logger.writeLine(FeatureSelector.MIN_CNT + "=" + this.getMinCount());
        logger.writeLine(FeatureSelector.DICTIONARY_FLAG);
        Iterator featureIter = this.getFeatureSetIterator();
        while (featureIter.hasNext()) {
            String feature = featureIter.next().toString();
            logger.writeLine(feature);
        }
        if(labels != null)
        {
	        logger.writeLine(FeatureSelector.LABEL_FLAG);
	        for (int i = 0; i < labels.size(); i++) {
	            logger.writeLine(labels.elementAt(i).toString());
	        }
        }
        logger.close();
    }
    
    //for visualization
    private String TaggedText=null;
	public String getTaggedText() {
		// TODO Auto-generated method stub
		return TaggedText;
	}
	public void setTaggedText(String TaggedStr) {
		TaggedText = TaggedStr;
	}


}