package edu.cmu.taghelpertools.middle_layer;

import java.util.*;
import java.io.*;

import jxl.Cell;
import jxl.Sheet;
import jxl.Workbook;
import jxl.WorkbookSettings;
import jxl.read.biff.BiffException;
import jxl.write.Label;
import jxl.write.WritableCell;
import jxl.write.WritableCellFeatures;
import jxl.write.WritableSheet;
import jxl.write.WritableWorkbook;
import jxl.write.WriteException;
import jxl.write.biff.RowsExceededException;
/**
 * The class provides static methods aiming at randomly sampling input dataset 
 * based on the provided proportion parameter.
 * 
 * @author haochuan
 *
 */
public class InputDataSubsetter {
/**
 * randomly sample input data based on the proportion parameter.
 * return a Sheet.
 * @param inputData
 * @param proportion
 * @return
 */
	public static Sheet subsetData(Sheet inputData, double proportion)
	{
		return subsetData(proportion, inputData).getSheet(0);
	}
	
	/**
	 * Choose a random subset of rows from the original data matrix
	 * 
	 * @param dataMatrix
	 * @param proportion
	 * @return
	 */
	public static ArrayList <ArrayList <String>> subsetData(ArrayList <ArrayList <String>> dataMatrix, 
			double proportion) {
		
		ArrayList <ArrayList <String>> subsetMatrix = new ArrayList <ArrayList <String>>();
		
		int numOfRows = dataMatrix.size() - 1;
		int reqNumOfRows = (int)Math.floor(numOfRows*proportion);
		
		// copy data header
		ArrayList <String> headerRow = new ArrayList <String>();
		for(int i=0; i<dataMatrix.get(0).size(); i++)
		{
			headerRow.add(dataMatrix.get(0).get(i));
		}
		subsetMatrix.add(headerRow);
	
		// pick which rows to use as subset
		Set indices = pickNumbers(numOfRows, reqNumOfRows);
		for(int i=1; i<numOfRows; i++) // don't process the headline
		{
			if(indices.contains(new Integer(i-1)))
			{
				ArrayList <String> currRow = new ArrayList <String>();
				for(int j=0; j<dataMatrix.get(i).size(); j++)
				{
					currRow.add(dataMatrix.get(i).get(j));
				}
				subsetMatrix.add(currRow);
			}
		}
		
		return subsetMatrix;
	}
/**
 * randomly sample input data based on the proportion parameter.
 * return a WritableWorkbook which can be outputted as a file.
 * @param proportion
 * @param inputData
 * @return
 */
	public static WritableWorkbook subsetData(double proportion, Sheet inputData)
	{
		if(proportion > 1)
		{
			System.err.println("Error parameter for data subsetting. Proportion must be smaller than 1 (inclusive).");
			return null;
		}
		
		WritableWorkbook tmpBook = null;
		WritableSheet processed = null;
		try{
			tmpBook = Workbook.createWorkbook(new File("etc/tmpWorkbook.xls"));
			processed = tmpBook.createSheet("sheet0", 0);
		}
		catch(Exception ex)
		{
			System.err.println("Error on subsetting the input data. Contact the developer for details.");
			return null;
		}
				
		if(processed !=null)
		{
			try{
				int numOfRows = inputData.getRows() - 1;
				Cell[] headLine = inputData.getRow(0);
				int numOfColumns = headLine.length;
				int reqNumOfRows = (int)Math.floor(numOfRows*proportion);
				
				for(int i=0; i<numOfColumns; i++)
				{
					Label label = new Label (i, 0, headLine[i].getContents());
					processed.addCell(label);
				}
			
				Set indices = pickNumbers(numOfRows, reqNumOfRows);
				int count = 1;
				for(int i=1; i<numOfRows; i++) // don't process the headline
				{
					if(indices.contains(new Integer(i-1)))
					{
						Cell[] content = inputData.getRow(i);
						for(int j=0; j<numOfColumns; j++)
						{
							Label label = new Label (j, count, content[j].getContents());
							processed.addCell(label);
						}						
						count++;						
					}
				}
			}
			catch(Exception ex)
			{
				ex.printStackTrace();
				return null;
			}
		}
						
		return tmpBook;
	}
	
	  private static Set pickNumbers(int aUpperLimit, int size) {
		    if (aUpperLimit <=0) {
		      throw new IllegalArgumentException("UpperLimit must be positive: " + aUpperLimit);
		    }
		    if (size <=0) {
		      throw new IllegalArgumentException("Size of returned List must be greater than 0.");
		    }
		    if(size > aUpperLimit)
		    {
			      throw new IllegalArgumentException("Size must be smaller than or equal to the upperLimit");		    	
		    }
		    //reuse one and the same Random object to generate a series
		    //of random Integers
		    // use random seed if provided (pseudo-random for replicating data subsets)
		    Random generator = null;
		    if (Driver.getRandomSeed() != -1) {
		    	generator = new Random(Driver.getRandomSeed());
		    } else {
		    	generator = new Random();
		    }
		    
		    Set result = new TreeSet();
		    
		    while(true)
		    {
		    	if(result.size() == size)
		    		break;
		    	else
		    		result.add(new Integer(generator.nextInt(aUpperLimit)));
		    }
		    return result;
	  }
	  
	  public static void main(String[] args)
	  {
//		  Set s = pickNumbers(100, 100);
//		  System.out.println(s);
		  try
		  {
			  Workbook workbook = Workbook.getWorkbook(new File("sample_data/German_best_demo.xls"), new WorkbookSettings());
			  WritableWorkbook w = InputDataSubsetter.subsetData(1, workbook.getSheet(0));
			  w.write();
			  w.close();
			  System.out.println("done");
		  }
		  catch(Exception ex)
		  {
			  ex.printStackTrace();
		  }
	  }
		
}
