/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package edu.cmu.taghelpertools.middle_layer;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.TreeMap;
import java.util.TreeSet;

import javax.swing.JOptionPane;

import org.upsam.grade.util.DLLChineseSegmenter;

import com.Ostermiller.util.CSVParser;

import edu.cmu.taghelpertools.gui.ProgressDialog;
import edu.cmu.taghelpertools.io.Logger;
import edu.cmu.taghelpertools.io.Reader;
import edu.cmu.taghelpertools.tokenizer.ChineseTokenizer;
import edu.cmu.taghelpertools.tokenizer.Tokenizer;
import edu.cmu.taghelpertools.util.HelperFunctions;

import weka.core.FastVector;

import jxl.Cell;
import jxl.Sheet;
import jxl.Workbook;
import jxl.WorkbookSettings;
import jxl.read.biff.BiffException;
import jxl.write.Label;
import jxl.write.WritableCell;
import jxl.write.WritableCellFeatures;
import jxl.write.WritableSheet;
import jxl.write.WritableWorkbook;
import jxl.write.WriteException;
import jxl.write.biff.RowsExceededException;

import gnu.trove.THashMap;
import gnu.trove.THashSet;

/**
 * Reads in contents from input .xls file. Collects list of dimensions and all
 * distinct classes for each dimension. Each row in the input .xls file is one
 * data point (see DataPoint).
 * 
 * @author Jaime Arguello, Hao-Chuan Wang
 * 
 */

public class InputFileReader {

	public static final char UNKNOWN = '?';
	public static final String UNKNOWN_STRING = "UNKNOWN";
	public static final String TEXT = "text";
	public static final String SUBPOP = "subpop";

	// public static final String SPECIFIC_CSV_SEP ="###";

	private TreeSet[] dimSets;
	private String[] dimNames;

	public void initiateDimClasses(int numDimMaps) {
		dimSets = new TreeSet[numDimMaps];
		for (int i = 0; i < dimSets.length; i++) {
			dimSets[i] = new TreeSet();
		}
	}

	public Iterator getIteratorForDim(int index) {
		return dimSets[index].iterator();
	}

	public String getNameForDim(int index) {
		return dimNames[index];
	}

	public int getNumOfDims() {
		return dimNames.length;
	}

	/**
	 * Reads in the input file and creates a DataPoint for each row. Each
	 * DataPoint contains its values for all dimensions in the inptu file and the
	 * text, which is the same for all dimensions. Each dimension ultimately
	 * becomes a separate classification problem. Assumes that the first row in
	 * the input file contains the names of each dimension and the flag "text" to
	 * indicate the text column.
	 * 
	 * @param in
	 * @param pDialog
	 * @param featureSelector
	 * @return
	 * @throws Exception
	 */
	public Iterator readFile(Reader in, ProgressDialog pDialog,
			FeatureSelector featureSelector) throws Exception {
		try {
			return readFile(in, pDialog, featureSelector, false, null, false);
		} catch (Exception ex) {
			throw ex;
		}
	}

	/**
	 * Reads in the input file and creates a DataPoint/EvalDataPoint for each row.
	 * Each DataPoint contains its values for all dimensions in the inptu file and
	 * the text, which is the same for all dimensions. Each dimension ultimately
	 * becomes a separate classification problem. Assumes that the first row in
	 * the input file contains the names of each dimension and the flag "text" to
	 * indicate the text column. Note that this polymorphic method requires the
	 * boolean argeument, forFeatureExtraction. The value of this parameter will
	 * be passed to the constructor of DataPoint for a decision of whether to
	 * build an inverted file data structure for feature extraction.
	 * 
	 * @param in
	 * @param pDialog
	 * @param featureSelector
	 * @param forFeatureExtraction
	 * @return Iterator of DataPoints
	 * @throws Exception
	 */
	public Iterator readFile(Reader in, ProgressDialog pDialog,
			FeatureSelector featureSelector, boolean forFeatureExtraction)
			throws Exception {
		return readFile(in, pDialog, featureSelector, forFeatureExtraction, null,
				false);
	}

	/**
	 * Reads in the input file and creates a DataPoint/EvalDataPoint for each row.
	 * Each DataPoint contains its values for all dimensions in the inptu file and
	 * the text, which is the same for all dimensions. Each dimension ultimately
	 * becomes a separate classification problem. Assumes that the first row in
	 * the input file contains the names of each dimension and the flag "text" to
	 * indicate the text column. Note that this polymorphic method requires the
	 * boolean argeument, forFeatureExtraction. The value of this parameter will
	 * be passed to the constructor of DataPoint for a decision of whether to
	 * build an inverted file data structure for feature extraction.
	 * 
	 * @param in
	 * @param pDialog
	 * @param featureSelector
	 * @param forFeatureExtraction
	 * @return Iterator of DataPoints
	 * @throws Exception
	 */
	public Iterator readFile(Reader in, ProgressDialog pDialog,
			FeatureSelector featureSelector, boolean forFeatureExtraction,
			SavedModelEvaluator savedModelEvaluator) throws Exception {
		return readFile(in, pDialog, featureSelector, forFeatureExtraction,
				savedModelEvaluator, false);
	}

	/**
	 * Reads in the input file and creates a DataPoint/EvalDataPoint for each row.
	 * Each DataPoint contains its values for all dimensions in the inptu file and
	 * the text, which is the same for all dimensions. Each dimension ultimately
	 * becomes a separate classification problem. Assumes that the first row in
	 * the input file contains the names of each dimension and the flag "text" to
	 * indicate the text column. Note that this polymorphic method requires the
	 * boolean argeument, forFeatureExtraction. The value of this parameter will
	 * be passed to the constructor of DataPoint for a decision of whether to
	 * build an inverted file data structure for feature extraction.
	 * 
	 * @param in
	 * @param pDialog
	 * @param featureSelector
	 * @param forFeatureExtraction
	 * @return Iterator of DataPoints
	 * @throws Exception
	 */
	public Iterator readFile(Reader in, ProgressDialog pDialog,
			FeatureSelector featureSelector, boolean forFeatureExtraction,
			SavedModelEvaluator savedModelEvaluator, boolean isTestFile)
			throws Exception {

		ArrayList<ArrayList<String>> dataMatrix;

		if (pDialog != null) {
			pDialog.setLabel("Loading: " + in.getFilename());
			pDialog.setValue(0);
		}

		String filename = in.getFilename();

		if (in.getFilename().endsWith(".csv") || Driver.isForceCSV()) {
			// populate the dataMatrix from CSV file
			// dataMatrix = readCSV(in);
			dataMatrix = new ArrayList<ArrayList<String>>();
			CSVParser csvParser = new CSVParser(in.getReader());
			csvParser.setCommentStart("#");
			String[] lineValues = null;
			while ((lineValues = csvParser.getLine()) != null) {
				ArrayList<String> row = new ArrayList<String>();
				for (int i = 0; i < lineValues.length; i++) {
					row.add(lineValues[i]);
				}
				dataMatrix.add(row);
			}
		} else {
			// by default assume .xls file
			// open Excel workbook
			Workbook workbook = null;
			try {
				workbook = Workbook.getWorkbook(new File(filename),
						new WorkbookSettings());
			} catch (BiffException e) {
				JOptionPane.showMessageDialog(null, "Could not read .xls file: "
						+ in.getFilename(), "TERMINAL ERROR", JOptionPane.ERROR_MESSAGE);
				throw e;
				// System.exit(0);
			} catch (IOException e) {
				JOptionPane.showMessageDialog(null, "Could not read .xls file: "
						+ in.getFilename(), "TERMINAL ERROR", JOptionPane.ERROR_MESSAGE);
				throw e;
				// System.exit(0);
			}

			Sheet sheet = workbook.getSheet(0);

			dataMatrix = new ArrayList<ArrayList<String>>();
			for (int i = 0; i < sheet.getRows(); i++) {
				ArrayList<String> currRowData = new ArrayList<String>();
				Cell[] currRow = sheet.getRow(i);
				for (int j = 0; j < sheet.getColumns(); j++) {
					currRowData.add(currRow[j].getContents().trim());
				}
				dataMatrix.add(currRowData);
			}
		}

		// select a subset of datapoints if requested (only when training)
		if (featureSelector.getOptionsModel().isSubsetInput()
				&& savedModelEvaluator == null) {
			dataMatrix = InputDataSubsetter.subsetData(dataMatrix, featureSelector
					.getOptionsModel().getSubsetPortion());
		}

		int dataPointIndex = 0;
		// initializes list of DataPoints
		ArrayList dataPoints = new ArrayList();
		try {
			if (pDialog != null)
				pDialog.setMax(dataMatrix.size() - 1);
			ArrayList dimList = new ArrayList();
			// keeps track of which is the index of the 'text' column
			// (we initialize this to the value from DataPoint class,
			// to get the value set during processing training file --
			// only required in the "hidden functionality mode" of
			// having a train/test file pair)
			int textColumn = DataPoint.textColumn;
			// keeps track of when nominal features start, which is after
			// the 'text' column
			int nominalStartColumn = DataPoint.nominalStartColumn;
			boolean beginNominal = false;

			boolean subPopPresent = false; // set to true if there is a column named
																			// 'subpop'
			// immediately after the 'text' column
			int subPopCol = Integer.MIN_VALUE;

			int evalDim = -1; // used only when evaluating a model (reading the test
												// file)

			ArrayList<String> header = dataMatrix.get(0);
			int numColumns = header.size();
			int numDim = DataPoint.dimNum;
			if (savedModelEvaluator == null && !isTestFile) {
				DataPoint.nominalFeatureNames = new ArrayList();
				// we also dynamically build the list of possible values for
				// each nominal feature. this is required at the time of writing
				// the ARFF file
				DataPoint.nominalFeatureValueSet = new THashMap();
				for (int i = 0; i < numColumns; i++) {
					String heading = header.get(i);
					heading = heading.replaceAll("\\W+", "");
					if (beginNominal) {
						DataPoint.nominalFeatureNames.add(heading);
					} else {
						if (!heading.equalsIgnoreCase(InputFileReader.TEXT)) {
							dimList.add(heading);
						} else {
							textColumn = i;
							DataPoint.textColumn = i;
							if (Driver.getNumAuxTextCols() > 0) {
								// we are just trusting the command line and the person
								// who wrote it
								DataPoint.auxTextColMin = i + 1;
								DataPoint.auxTextColMax = i + Driver.getNumAuxTextCols();
								i = i + Driver.getNumAuxTextCols();
							}
							// check 'subpop' after 'text' column
							if ((i + 1) != numColumns
									&& header.get(i + 1).trim().equalsIgnoreCase(
											InputFileReader.SUBPOP)) {
								subPopPresent = true;
								subPopCol = i + 1;
								DataPoint.subPopColumn = subPopCol;
								i++;
							}
							beginNominal = true;
							nominalStartColumn = i + 1;
							DataPoint.nominalStartColumn = i + 1;
						}
					}
				}
				numDim = dimList.size();
				// sets the static variable dimNum. This is to make sure that all
				// rows have the same number of dimensions.
				DataPoint.dimNum = numDim;
			} else if (savedModelEvaluator != null && !isTestFile) {
				EvalDataPoint.nominalFeatureNames = featureSelector.getOptionsModel()
						.getNominalFeatureNames();
				int numNominalinTrain = EvalDataPoint.nominalFeatureNames != null ? EvalDataPoint.nominalFeatureNames
						.size()
						: 0;
				EvalDataPoint.nominalFeatureValueSet = featureSelector
						.getOptionsModel().getNominalFeatureValueSet();
				for (int i = 0; i < numColumns; i++) {
					String heading = header.get(i).trim();
					heading = heading.replaceAll("\\W+", "");
					if (beginNominal) {
						if (!((String) EvalDataPoint.nominalFeatureNames.get(i
								- nominalStartColumn)).equals(heading)) {
							JOptionPane
									.showMessageDialog(
											null,
											"ERROR: Mismatch in additional feature headers. Expecting "
													+ "feature \""
													+ (String) EvalDataPoint.nominalFeatureNames.get(i
															- nominalStartColumn)
													+ "\" in column "
													+ i
													+ ", found \""
													+ heading
													+ "\" instead! Note that feature names are case sensitive, "
													+ "\"FEATURE\", \"Feature\" and \"feature\" are all different.",
											"TERMINAL ERROR", JOptionPane.ERROR_MESSAGE);
							System.exit(-1);
						}
					} else {
						if (!heading.equalsIgnoreCase(InputFileReader.TEXT)) {
							dimList.add(heading);
							if (heading.equals(savedModelEvaluator.getDimName()))
								evalDim = i;
						} else {
							textColumn = i;
							if (Driver.getNumAuxTextCols() > 0) {
								// we are just trusting the command line and the person
								// who wrote it
								DataPoint.auxTextColMin = i + 1;
								DataPoint.auxTextColMax = i + Driver.getNumAuxTextCols();
								i = i + Driver.getNumAuxTextCols();
							}
							// check 'subpop' after 'text' column
							if ((i + 1) != numColumns
									&& header.get(i + 1).trim().equalsIgnoreCase(
											InputFileReader.SUBPOP)) {
								subPopPresent = true;
								subPopCol = i + 1;
								i++;
							}
							beginNominal = true;
							nominalStartColumn = i + 1;
							// here, also verify that we have the exact number of nominal
							// features
							if ((numColumns - nominalStartColumn) > numNominalinTrain) {
								JOptionPane.showMessageDialog(null,
										"ERROR: More number of additional feature headers than "
												+ "seen during training the model. Expecting "
												+ numNominalinTrain + ", found "
												+ (numColumns - nominalStartColumn) + " instead!",
										"TERMINAL ERROR", JOptionPane.ERROR_MESSAGE);
								System.exit(-1);
							} else if ((numColumns - nominalStartColumn) < numNominalinTrain) {
								JOptionPane.showMessageDialog(null,
										"ERROR: Less number of additional feature headers than "
												+ "seen during training the model. Expecting "
												+ numNominalinTrain + ", found only "
												+ (numColumns - nominalStartColumn) + " instead!",
										"TERMINAL ERROR", JOptionPane.ERROR_MESSAGE);
								System.exit(-1);
							}
						}
					}
					numDim = dimList.size();
				}

				// verify dimension name
				if (evalDim == -1) {
					JOptionPane
							.showMessageDialog(
									null,
									"ERROR: Dimension \""
											+ savedModelEvaluator.getDimName()
											+ "\" not encountered in evaluation file. Note that dimension names are case sensitive, "
											+ "\"DIMENSION\", \"Dimension\" and \"dimension\" are all different.",
									"TERMINAL ERROR", JOptionPane.ERROR_MESSAGE);
					System.exit(-1);
				}
			} else if (isTestFile) {
				if (DataPoint.subPopColumn != Integer.MIN_VALUE) {
					subPopPresent = true;
					subPopCol = DataPoint.subPopColumn;
				}
			}

			// set up the data structures to store different
			// values of each dimension
			if (savedModelEvaluator == null && !isTestFile) {
				// extract dimension names
				this.dimNames = new String[dimList.size()];
				int index = 0;
				Iterator dimNameIter = dimList.iterator();
				while (dimNameIter.hasNext()) {
					String dimName = dimNameIter.next().toString();
					dimNames[index] = dimName;
					System.out.println("DIM: " + dimName);
					index++;
				}
				this.initiateDimClasses(numDim);
			}

			int numInst = dataMatrix.size();
			for (dataPointIndex = 1; dataPointIndex < numInst; dataPointIndex++) {
				if (pDialog != null) {
					pDialog.setValue(dataPointIndex);
					pDialog.updateLabelWithSuffix("[instance " + dataPointIndex + "/"
							+ (numInst - 1) + "]");
				}
				String[] dims = new String[numDim];
				for (int j = 0; j < numDim; j++) {
					String content = dataMatrix.get(dataPointIndex).get(j);
					// DO NOT LOWERCASE THE CLASS VALUES!!!!
					content = content.trim();
					// if the dim value is UNKNOWN (?), then replace the '?'
					// with String UNKNOWN_STRING
					if (content.length() == 1) {
						if (content.charAt(0) == InputFileReader.UNKNOWN) {
							dims[j] = InputFileReader.UNKNOWN_STRING;
							continue;
						}
					}
					// if it's not unknown than set collect the class name.
					if (content.trim().length() == 0) {
						if (pDialog != null)
							pDialog.setLabel("found: " + dataPoints.size() + " data points.");
						return dataPoints.iterator();
					}

					if (savedModelEvaluator == null && !isTestFile)
						dimSets[j].add(content);
					else if (savedModelEvaluator == null) {
						if (!dimSets[j].contains(content)) {
							// processing a train/test pair
							System.out
									.println("WARNING: Found value "
											+ content
											+ " for dimension \""
											+ dimNames[j]
											+ "\", which was not seen during training. Can safely ignore if this is a numeric "
											+ "dimension. Else error will be thrown later!!");
						}
					} else if (savedModelEvaluator != null && !isTestFile) {
						// warn if a file
						// specified for evaluation contains a value never
						// seen before (Weka will throw exception in such
						// cases)
						if (!savedModelEvaluator.getDimLabels().contains(content)) {
							System.out
									.println("WARNING: Found value "
											+ content
											+ " for dimension \""
											+ savedModelEvaluator.getDimName()
											+ "\", which was not seen during training. Can safely ignore if this is a numeric "
											+ "dimension. Else error will be thrown later!!");
						}
					}

					dims[j] = content;
				}

				// get the text from the text column
				String text = null;
				try {
					text = dataMatrix.get(dataPointIndex).get(textColumn);
				} catch (ArrayIndexOutOfBoundsException e) {
					// e.printStackTrace();
					JOptionPane
							.showMessageDialog(null,
									"Could not find 'text' column header for feature extraction "
											+ e.getMessage(), "TERMINAL ERROR",
									JOptionPane.ERROR_MESSAGE);
					throw new TextNotFoundException();
				}

				// get auxiliary text columns if any
				String[] auxTexts = null;
				if (Driver.getNumAuxTextCols() > 0) {
					auxTexts = new String[Driver.getNumAuxTextCols()];
					for (int k = DataPoint.auxTextColMin, l = 0; k <= DataPoint.auxTextColMax; k++, l++)
						auxTexts[l] = dataMatrix.get(dataPointIndex).get(k);
				}

				String subPopId = "";
				if (subPopPresent) {
					subPopId = dataMatrix.get(dataPointIndex).get(subPopCol).trim();
				}

				// get any additional nominal features present after the text
				// column
				if (isTestFile) {
					nominalStartColumn = DataPoint.nominalStartColumn;
				}

				ArrayList nominalFeatures = new ArrayList();
				for (int j = nominalStartColumn; j < numColumns; j++) {
					String content = dataMatrix.get(dataPointIndex).get(j);
					content = content.trim();
					if (content.length() == 0) {
						nominalFeatures.add(DataPoint.MISSING_NOMINAL_VALUE);
					} else {
						nominalFeatures.add(content);
						if (savedModelEvaluator == null && !isTestFile) {
							// here we are training, so update values for nominal attributes
							// if not present
							// already

							// get the name of the nominal feature corresponding to the
							// current
							// column in the excel file
							String nominalFeatName = (String) DataPoint.nominalFeatureNames
									.get(j - nominalStartColumn);

							// if the feature is already present in the static values map,
							// update its
							// set of possible values
							if (DataPoint.nominalFeatureValueSet.containsKey(nominalFeatName)) {
								FastVector tmpVec = (FastVector) DataPoint.nominalFeatureValueSet
										.get(nominalFeatName);
								if (!tmpVec.contains(content)) {
									tmpVec.addElement(content);
									DataPoint.nominalFeatureValueSet.put(nominalFeatName, tmpVec);
								}
							} else {
								// need to create a new FastVector for the possible values of
								// the nominal feature.
								// we use the FastVector datatype since it is from weka API and
								// can be directly
								// used later to create nominal attributes
								FastVector tmpVec = new FastVector();
								tmpVec.addElement(content);
								// add this to the static values map
								DataPoint.nominalFeatureValueSet.put(nominalFeatName, tmpVec);
							}
						} else if (savedModelEvaluator != null && !isTestFile) {
							// here we are testing, so complain if we encounter a value for a
							// nominal attribute
							// that we never saw in the training set

							// get the name of the nominal feature corresponding to the
							// current
							// column in the excel file
							String nominalFeatName = (String) EvalDataPoint.nominalFeatureNames
									.get(j - nominalStartColumn);

							// actually, don't complain too hard, don't quit. this might be a
							// real valued
							// numeric attribute, when user can ignore this warning
							if (!((FastVector) EvalDataPoint.nominalFeatureValueSet
									.get(nominalFeatName)).contains(content)) {
								System.out
										.println("WARNING: Found value "
												+ content
												+ " for feature \""
												+ nominalFeatName
												+ "\", which was not seen during training. Can safely ignore if this is a numeric "
												+ "feature. Else error will be thrown later!!");
							}
						}
					}
				}

				// if the data is in Chinese, then do some basic language normalization
				// first
				text = HelperFunctions.replacePuncWithSpecialTokens(text,
						featureSelector.getOptionsModel().getLanguage());

				int numTokens = 0;
				if (savedModelEvaluator == null) {
					DataPoint dp = new DataPoint(dims, text, featureSelector,
							nominalFeatures, isTestFile, forFeatureExtraction);
					dp.setSubPopId(subPopId);
					if (auxTexts != null)
						dp.setAuxTexts(auxTexts);
					numTokens = Tokenizer.getInstance(featureSelector.getOptionsModel())
							.tokenizeIncStopword(dp.getText()).getNumTokens();
					dataPoints.add(dp);
				} else {
					EvalDataPoint edp = new EvalDataPoint(dims[evalDim], text,
							featureSelector, nominalFeatures, savedModelEvaluator
									.getFeatureSet(), true, forFeatureExtraction);
					edp.setSubPopId(subPopId);
					if (auxTexts != null)
						edp.setAuxTexts(auxTexts);
					numTokens = Tokenizer.getInstance(featureSelector.getOptionsModel())
							.tokenizeIncStopword(edp.getText()).getNumTokens();
					dataPoints.add(edp);
				}

				// updating the min/max token count for the feature selector.
				// min/max are used if the user specifies that they want to use
				// sentence length as a feature: sent_length = (X - min)/(max -
				// min)

				if (numTokens > featureSelector.getMaxCount()) {
					featureSelector.setMaxCount(numTokens);
				}
				if (numTokens < featureSelector.getMinCount()) {
					featureSelector.setMinCount(numTokens);
				}

			}
			if (pDialog != null)
				pDialog.setLabel("found: " + dataPoints.size() + " data points.");
			return dataPoints.iterator();
		} catch (TextNotFoundException e) {
			throw e;
		} catch (Exception e) {
			e.printStackTrace();
			JOptionPane.showMessageDialog(null, "Failed to read/process input file: "
					+ in.getFilename() + " at line " + dataPointIndex + ". ERROR: "
					+ e.getMessage(), "TERMINAL ERROR", JOptionPane.ERROR_MESSAGE);
			throw e;
		}
	}

	/**
	 * A static method for converting a CSV-like file into an XLS file by provided
	 * a specific separator (not limited to comma).
	 * 
	 * @param csvFile
	 * @param specific_sep
	 * @return
	 */
	public static String convertCSVToXLS(Reader csvFile, String specific_sep) {

		if (csvFile == null) {
			return null;
		}

		// convert a csv file to an excel file
		String xlsFileName = csvFile.getFilename().replaceAll(".csv", ".xls");
		try {

			WritableWorkbook xls = Workbook.createWorkbook(new File(xlsFileName),
					new WorkbookSettings());
			WritableSheet sheet = xls.createSheet("Sheet 1", 0);
			String line;
			int currRow = 0;
			while ((line = csvFile.readLine()) != null) {
				// sheet.insertRow(currRow);
				int currCol = 0;
				boolean inQuote = false;

				String[] values = line.split(specific_sep);
				for (currCol = 0; currCol < values.length; currCol++) {
					String content = values[currCol];
					Label l = new Label(currCol, currRow, content.trim());
					sheet.addCell(l);
				}

				currRow++;
			}
			xls.write();
			xls.close();

		} catch (IOException e) {
			JOptionPane.showMessageDialog(null,
					"IOException: Could not create intermediate .xls file: "
							+ xlsFileName, "TERMINAL ERROR", JOptionPane.ERROR_MESSAGE);
			System.exit(0);
		} catch (RowsExceededException e) {
			JOptionPane.showMessageDialog(null,
					"RowsExceededException: Could not create intermediate .xls file: "
							+ xlsFileName, "TERMINAL ERROR", JOptionPane.ERROR_MESSAGE);
			System.exit(0);
		} catch (WriteException e) {
			JOptionPane.showMessageDialog(null,
					"WriteException: Could not create intermediate .xls file: "
							+ xlsFileName, "TERMINAL ERROR", JOptionPane.ERROR_MESSAGE);
			System.exit(0);
		}

		return xlsFileName;
	}

	/**
	 * NOT USED ANYMORE
	 * 
	 * A static method for converting a CSV-like file into an XLS file.
	 * 
	 * @param csvFile
	 * @return
	 * 
	 * @deprecated
	 */
	public static String convertCSVToXLS(Reader csvFile) {

		if (csvFile == null) {
			return null;
		}

		// convert a csv file to an excel file
		String xlsFileName = csvFile.getFilename().replaceAll(".csv", ".xls"); // csvFile.getFilename()
																																						// +
																																						// ".xls"
		try {

			WritableWorkbook xls = Workbook.createWorkbook(new File(xlsFileName),
					new WorkbookSettings());
			WritableSheet sheet = xls.createSheet("Sheet 1", 0);
			String line;
			int currRow = 0;
			while ((line = csvFile.readLine()) != null) {
				// sheet.insertRow(currRow);
				int currCol = 0;
				String currContent = "";
				boolean inQuote = false;

				char chars[] = line.toCharArray();

				for (int i = 0; i < chars.length; i++) {
					if (inQuote) {
						if (String.valueOf(chars[i]).equals("\"")) {
							inQuote = false;
						} else {
							currContent += String.valueOf(chars[i]);
						}

						if (i == chars.length - 1) {
							Label l = new Label(currCol, currRow, currContent.trim());
							sheet.addCell(l);
							currCol++;
							currContent = "";
						}
					} else {
						if (String.valueOf(chars[i]).equals(",")) {
							Label l = new Label(currCol, currRow, currContent.trim());
							sheet.addCell(l);
							currCol++;
							currContent = "";
						} else {
							if (String.valueOf(chars[i]).equals("\"")) {
								inQuote = true;
							} else {
								currContent += String.valueOf(chars[i]);
							}
						}

						if (i == chars.length - 1) {
							Label l = new Label(currCol, currRow, currContent.trim());
							sheet.addCell(l);
							currCol++;
							currContent = "";
						}
					}
				}
				currRow++;
			}
			xls.write();
			xls.close();

		} catch (IOException e) {
			JOptionPane.showMessageDialog(null,
					"IOException: Could not create intermediate .xls file: "
							+ xlsFileName, "TERMINAL ERROR", JOptionPane.ERROR_MESSAGE);
			System.exit(0);
		} catch (RowsExceededException e) {
			JOptionPane.showMessageDialog(null,
					"RowsExceededException: Could not create intermediate .xls file: "
							+ xlsFileName, "TERMINAL ERROR", JOptionPane.ERROR_MESSAGE);
			System.exit(0);
		} catch (WriteException e) {
			JOptionPane.showMessageDialog(null,
					"WriteException: Could not create intermediate .xls file: "
							+ xlsFileName, "TERMINAL ERROR", JOptionPane.ERROR_MESSAGE);
			System.exit(0);
		}

		return xlsFileName;
	}

	/**
	 * A private & static method for removing quotation marks from the provided
	 * String.
	 * 
	 * @param input
	 * @return
	 */
	private static String removeQuotation(String input) {
		char[] chars = input.trim().toCharArray();
		String content = "";
		for (int i = 0; i < chars.length; i++) {
			if (String.valueOf(chars[i]).equals("\"")
					&& (i == 0 || i == chars.length - 1)) {
				continue;
			} else {
				content = content + String.valueOf(chars[i]);
			}
		}

		return content;
	}

	/**
	 * Writes all distinct classes for each of the dimensions in the input file.
	 * This is done as a sanity check.
	 */
	public void writeDims() {
		System.out.print("DIMENSIONS:" + "\n");
		for (int i = 0; i < dimSets.length; i++) {
			Iterator iter = dimSets[i].iterator();
			while (iter.hasNext()) {
				if (iter.hasNext()) {
					System.out.print(iter.next() + ",");
				} else {
					System.out.print(iter.next());
				}
			}
			System.out.println();
		}
	}

}
