/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package edu.cmu.taghelpertools.middle_layer;

import java.io.File;
import java.io.FileWriter;
import java.io.FileOutputStream;
import java.io.PrintWriter;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Random;
import java.util.TreeMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.print.attribute.standard.PDLOverrideSupported;
import javax.swing.JOptionPane;


import edu.cmu.taghelpertools.gui.Classifiers;
import edu.cmu.taghelpertools.gui.ProgressDialog;
import edu.cmu.taghelpertools.io.Logger;
import edu.cmu.taghelpertools.io.Reader;
import edu.cmu.taghelpertools.middle_layer.extractor.FeatureSpaceMap;
import edu.cmu.taghelpertools.util.GenericNode;
import edu.cmu.taghelpertools.util.GenericNodeSorter;
import gnu.trove.THashMap;
import gnu.trove.THashSet;

import weka.classifiers.Classifier;
import weka.classifiers.CostMatrix;
import weka.classifiers.Evaluation;
import weka.classifiers.evaluation.NominalPrediction;
import weka.core.Attribute;
import weka.core.FastVector;
import weka.core.Instance;
import weka.core.Instances;
import weka.core.Utils;
import weka.core.converters.ArffLoader;
import weka.estimators.Estimator;
import weka.estimators.KernelEstimator;

/**
 * MultiDimClassifier controls the processing steps for a single file. This
 * includes the feature selection, training, testing, and self-training
 * iterations.
 * 
 * @author Jaime Arguello, Mahesh Joshi, Hao-Chuan Wang
 * 
 */
public class MultiDimClassifier {

    // output directories. the CLASSIFIERS directory holds subdirectories (one
    // for each classification problem).
    // Each of those subdirectories will contain the .model file (the weka
    // model) and the .dic file (the directionary file
    // that holds all the features used to train the classifier.

    public static final String CLASSIFIERS = "CLASSIFIERS";
    public static final String OUTPUT = "OUTPUT";
    public static final String SCRIPTS = "SCRIPTS";
    public static final String ARFF = "ARFF";

    public static final int MAX_ITERATIONS = 10;
    public static final String TRAIN_SET = "training set";
    public static final String TEST_SET = "test set";
    public static final String CLASS = "CLASS";
    
    public static final int FORMAT_MINORTHIRD = 1;
    public static final int FORMAT_SVMLIGHT = 2;
    public static final int FORMAT_GIBBSLDA = 3;

    public static final String FORMAT_MINORTHIRD_EXT = ".MT";
    public static final String FORMAT_SVMLIGHT_EXT = ".dat";
    public static final String FORMAT_GIBBSLDA_EXT = ".gibbsdat";
    public static final String FEATURES_EXT = ".features";

    public static String MTPREFIX = "k";
    
    // iteration modes:
    // (1) add fixed number of predictions to train set
    // (2) add a percentage of predictions to train set
    // (3) add all predictions w/ confidence > thresh to train set.

    public static final int ADD_FIXED_NUM = 0;
    public static final int ADD_THRESH = 1;
    public static final int ADD_PERCENT = 2;
    
    private static final boolean isTenCV = true;

    private Reader input;
    private Reader test;
    private ArrayList dataPoints;
    private ArrayList testDataPoints;
    private FeatureSelector featureSelector;
    private OptionsModel optionsModel;
    private InputFileReader inputFileReader;
    private Classifier classifier;
    private int iterationMode;
    private int fixedNum;
    private double confThresh;
    private double percent;
    
    private Instances trainSet;
    private Instances testSet;
    private Instances backupSet;
    
    /*
     * schaudhu@ 08/30/2007
     * my variables
     */
    protected int m_NumClasses;
    protected int m_NumFolds;
    protected Estimator m_ErrorEstimator;
    protected double m_MissingClass;
    protected double m_SumPriorEntropy;
    protected double m_SumSchemeEntropy;
    protected Estimator m_PriorErrorEstimator;
    protected double m_TotalCost;
    protected double m_SumErr;
    protected double m_SumAbsErr;
    protected double m_SumSqrErr;
    protected double m_SumClass;
    protected double m_SumSqrClass;
    protected double m_SumPredicted;
    protected double m_SumSqrPredicted;
    protected double [] m_ClassPriors;
    protected double m_ClassPriorsSum;
    protected CostMatrix m_CostMatrix;
    protected double m_WithClass;
    protected double [][] m_ConfusionMatrix;
    protected String [] m_ClassNames;
    private FastVector m_Predictions;
    protected double m_MarginCounts [];
    protected int m_NumTrainClassVals;
    protected double [] m_TrainClassVals;
    protected double [] m_TrainClassWeights;
    protected double m_SumClassPredicted;
    protected double m_Unclassified;
    protected double m_Correct;
    protected double m_Incorrect;
    protected boolean m_NoPriors = false;
    protected boolean m_ClassIsNominal;
    protected double m_SumKBInfo;
    protected double m_SumPriorAbsErr;
    protected double m_SumPriorSqrErr;

    protected static int k_MarginResolution = 500;
    protected static final double MIN_SF_PROB = Double.MIN_VALUE;

    String dimn;

    //end of myVariables  
                  
    static {
    	if (System.getProperty("tht.mtprefix") != null)
    		MultiDimClassifier.MTPREFIX = System.getProperty("tht.mtprefix");
    }
    
    /**
     * Constructor of MultiDimClassifier
     * 
     * @param input -
     *            the input file (see Reader)
     * @param test -
     *            recommended to set to null by default            
     * @param iterationMode -
     *            the model of re-training. Re-training can be done by either
     *            (1) adding the most confident fixed number of predictions to
     *            the test set, (2) adding the most confident % of predicitions
     *            to the test set, or (3) adding all predictions above a
     *            certrain confidence threshold to the test set.
     * 
     * @param optionsModel -
     *            Contains all the feature selection options, such as unigrams,
     *            bigrams, punctuation, etc., language options, output options, 
     *            evaluation options and self-training options.
     * 
     */
    public MultiDimClassifier(Reader input, 
    						  Reader test, 
    						  int iterationMode, 
    						  OptionsModel optionsModel) {
        this.input = input;
        this.test = test;
        this.optionsModel = optionsModel;
        this.iterationMode = iterationMode;
        this.dataPoints = new ArrayList();
        this.testDataPoints = new ArrayList();
    }
        
    private void setFixedNum(int fixedNum) {
        this.fixedNum = fixedNum;
    }

    private void setConfThresh(double confThresh) {
        this.confThresh = confThresh;
    }

    private void setPercent(double percent) {
        this.percent = percent;
    }

    /**
     * Sets up all the needed parameters for starting the training, testing, and
     * re-training process.
     * 
     * @param pDialog -
     *            GUI components that shows how the process is progressing.
     * @throws throw an Exception object if any error occured
     */
    public void setUp(ProgressDialog pDialog) throws Exception{
        System.out.println("reading file " + input.getFilename() + "...");
        // holds all a list of texts of the data points.
//        ArrayList textList = new ArrayList();
//        ArrayList testTextList = new ArrayList();

        // gets the feature selector. From the constructor of FeatureSelector,
        // all the features from the text are chosen.
        // the "vocabulary" of features is set. Later, this vocabulary of
        // features is used to extract the vector of features from each data
        // point during training and testing.

        this.featureSelector = new FeatureSelector(optionsModel);

        // setting the iteration settings according to the OptionsModel.

        if (this.iterationMode == MultiDimClassifier.ADD_FIXED_NUM) 
        {
            this.setFixedNum(this.optionsModel.getFixedNum());
        } 
        else if (this.iterationMode == MultiDimClassifier.ADD_THRESH) 
        {
            this.setConfThresh(this.optionsModel.getThresh());
        } 
        else if (this.iterationMode == MultiDimClassifier.ADD_PERCENT) 
        {
            this.setPercent(this.optionsModel.getPercent());
        } 
        else 
        {
            JOptionPane.showMessageDialog(null, "Unrecognized Operation Mode (THIS IS A BUG). Contact: jaime@cs.cmu.edu", "TERMINAL ERROR", JOptionPane.ERROR_MESSAGE);
            System.exit(0);
        }

        // the inputFileReader is the object that reads in the excel file
        	inputFileReader = new InputFileReader();


        // passes the input file and gets back an iterator of DataPoints. Also,
        // when the data point is instantiated, it's featues are added to the
        // featureSelector featureSet. In other words, the feature selector
        // feature set contains the union of features in the dataPoints.
        Iterator iter;        	
        try{
        		 iter = inputFileReader.readFile(input, pDialog, featureSelector);
        }
        catch(Exception ex)
        {
        	throw ex;
        }        
        
        // adds each DataPoint's individual text (the utterance) to a text list.
        while (iter.hasNext()) {
            DataPoint dataPoint = (DataPoint) iter.next();
         // textList.add(dataPoint.getText());
            dataPoints.add(dataPoint);
        }

        // removing rare features from the featureSelector
        if (optionsModel.isRemoveRareFeatures()) {
            this.featureSelector.removeRareFeatures();
        }

      	// remove features that contain punctuations, since punctuation
      	// is a feature that we handle differently
        this.featureSelector.removePuncFeature();
        
        this.optionsModel.setNominalFeatureNames(DataPoint.nominalFeatureNames);
        this.optionsModel.setNominalFeatureValueSet(DataPoint.nominalFeatureValueSet);
        System.out.println("feature space size: " + this.featureSelector.getFeatureSetCount());

        // print out the dimensions/classes found in input file
        inputFileReader.writeDims();

        // print out number of datapoints found in the input file
        System.out.println("found " + dataPoints.size() + " datapoints");
                
        if (Driver.isSUPPLY_TESTING_DATASET() && (this.test != null)) {
            Iterator testIter;
            try{
            	// passing a test file here
            	testIter = inputFileReader.readFile(test, pDialog, featureSelector, false, null, true);
            }
            catch(Exception ex)
            {
            	throw ex;
            }
            // adds each DataPoint's individual text (the utterance) to a text list.
            while (testIter.hasNext()) {
                DataPoint dataPoint = (DataPoint) testIter.next();
               // testTextList.add(dataPoint.getText());
                testDataPoints.add(dataPoint);
            }            
        }
    }
    
    /**     
     * A protected method not to be publicly accessed.
     * this method is used by advanced feature extraction.
     * 
     */    
    protected void setPreprocessed(ArrayList<DataPoint> trainDataPoints, 
    							ArrayList<DataPoint> testDataPoints, 
    							InputFileReader ifr,
    							FeatureSelector fs)
    {
    	this.dataPoints = trainDataPoints;
    	this.testDataPoints = testDataPoints;
    	this.inputFileReader = ifr;
    	this.featureSelector = fs;
        
    	if (this.optionsModel.isRemoveRareFeatures()) {
            this.featureSelector.removeRareFeatures();
        }
    	
    	// remove the set of features that are not required for the
    	// final stemming/stoplisting setting
    	if(!this.optionsModel.isUnigrams())
    		fs.removePreprocessedOptionFeatures1("THF_U");
    	else
    	{
    		if     ( this.optionsModel.isStem() && this.optionsModel.isRemoveStopwords() )
    			fs.removePreprocessedOptionFeatures2("THF_U", "THF_UStemRmstopw_");
    		else if( this.optionsModel.isStem() && !this.optionsModel.isRemoveStopwords() )
    			fs.removePreprocessedOptionFeatures2("THF_U", "THF_UStemNormstopw_");
    		else if( !this.optionsModel.isStem() && !this.optionsModel.isRemoveStopwords() )
    			fs.removePreprocessedOptionFeatures2("THF_U", "THF_UNostemNormstopw_");
    		else if( !this.optionsModel.isStem() && this.optionsModel.isRemoveStopwords() )
    			fs.removePreprocessedOptionFeatures2("THF_U", "THF_UNostemRmstopw_");
    			
    	}
    	if(!this.optionsModel.isBigrams())
    		fs.removePreprocessedOptionFeatures1("THF_B");
    	else
    	{
    		if     ( this.optionsModel.isStem() && this.optionsModel.isRemoveStopwords() )
    			fs.removePreprocessedOptionFeatures2("THF_B", "THF_BStemRmstopw_");
    		else if( this.optionsModel.isStem() && !this.optionsModel.isRemoveStopwords() )
    			fs.removePreprocessedOptionFeatures2("THF_B", "THF_BStemNormstopw_");
    		else if( !this.optionsModel.isStem() && !this.optionsModel.isRemoveStopwords() )
    			fs.removePreprocessedOptionFeatures2("THF_B", "THF_BNostemNormstopw_");
    		else if( !this.optionsModel.isStem() && this.optionsModel.isRemoveStopwords() )
    			fs.removePreprocessedOptionFeatures2("THF_B", "THF_BNostemRmstopw_");
    			
    	}   
    	if(!this.optionsModel.isPosBigrams())
    		fs.removePreprocessedOptionFeatures1("THF_POS");
    	if(!this.optionsModel.isPuctuation())
    		fs.removePreprocessedOptionFeatures1("THF_PUN");
    	if(!this.optionsModel.isContainsNonStopword())
    		fs.removePreprocessedOptionFeatures1("THF_ConStop");
    	
    	// update feature map to remove unwanted prefixes
    	// and features
    	// note: order of calling the two methods below is important!
    	fs.removeFeatureMapPrefix();
    	// remove features that contain punctuations, since punctuation
    	// is a feature that we handle differently
    	fs.removePuncFeature();
    }


    /**      
     * A private method not to be publicly accessed.
     * this is a common method for training the classifier.
     * 
     */ 
    private void buildClassifier(Instances trainSet, ProgressDialog pDialog) throws Exception
    {
    	this.classifier = this.optionsModel.getClassifier();
		try {
			if (pDialog != null) {
				pDialog.setLabel("Training "+this.classifier.getClass().getSimpleName()+" ("+ trainSet.numAttributes() +" features / "+trainSet.numInstances()+" datapoints)");
				pDialog.setIsIndeterminate(true);
			}
			this.classifier.buildClassifier(trainSet);
		} 
		catch (Exception e) 
		{
			throw e;
		}
    }

    /**
     * trains, tests, and re-trains ONE classifier until convergence. This
     * method uses some Weka-specific data-structures used for training/testing.
     * This method is called from the Driver.
     * 
     * @param index -
     *            the index of the classifier that is currently being processed.
     *            A file can contain multiple classification problems. Each will
     *            be its own classifier. This index uniquely identifies this
     *            classifier based on its column in the input file
     * @param pDialog -
     *            Maintains a record of where the process is for feedback to the
     *            user (GUI component).
     * @return
     */
    public ClassifierModel run(int index, ProgressDialog pDialog) throws Exception 
    {
        System.out.println("Within classifier model *******************");

        // holds the confidence value of each iteraction. When the confidence of
        // the previous iteraction is greater than that of the current
        // iteraction, the process completes.
        double prevConfidence = (double) -1;
        String outputPostfix = "full_set.arff";

        // counts the number of iterations so that the system doesn't iterate
        // more than MAX_ITER.
        int iter = 1;
              
        while (true) {  
        	System.out.println("Within while(true) loop***********");
            DecimalFormat df = new DecimalFormat("#.####");
            if (pDialog != null) {
            	pDialog.setIsIndeterminate(false);
              pDialog.setTitle("DIM: " + inputFileReader.getNameForDim(index));
              pDialog.setLabel("ITER:" + iter + " -- CONFIDENCE:" + df.format(prevConfidence));
              pDialog.setValue(0);
              pDialog.setMax(this.dataPoints.size());
            }
            System.out.println("ITER:" + iter + " -- CONFIDENCE:" + df.format(prevConfidence));

            // Sets a FastVector of all the labels that this classifier
            // classifies into.
            FastVector atts = this.getAttributes(index);
            HashSet testDataPointSet = new HashSet();                        
            Instances[] datasets = this.getTrainInstances(index, testDataPointSet, pDialog, false);
            this.trainSet = datasets[0];
            this.testSet = datasets[1];  
//            this.testSet =  new Instances(MultiDimClassifier.TEST_SET, atts, 0);
            testSet.setClassIndex(testSet.numAttributes() - 1);            
            this.backupSet = new Instances(MultiDimClassifier.TEST_SET, atts, 0);
            backupSet.setClassIndex(backupSet.numAttributes() - 1);            

            System.out.println("set training/test set");              
            // /////////////////////////
            // if the test set is empty, then the processes ends (no need to
            // re-train).
            // /////////////////////////
            if (testDataPointSet.size() == 0) {
                ClassifierModel model = new ClassifierModel();

//                String trainSetFilename = this.writeARFFForDim(index, outputPostfix, this.optionsModel.getOutputPrefix(), true);
//                ArffLoader arffLoader = new ArffLoader();
//                arffLoader.setFile(new File(trainSetFilename));
//                System.out.println("arff file name: " + trainSetFilename);
                
        		if(this.classifier == null)
        		{
//TODO:         	
//        			this.trainSet = arffLoader.getDataSet();
//        			this.trainSet.setClassIndex(this.trainSet.numAttributes()-1);
        			try{
        				System.out.println("instances size: " + this.trainSet.numInstances());
        				this.buildClassifier(trainSet, pDialog);
        				
        			}
        			catch(Exception ex)
        			{
        				 JOptionPane.showMessageDialog(null, "Cannot build classifier " + 
        						 							 this.classifier.getClass().getName() + ": " + ex.getMessage(), 
        						 							 "TERMINAL ERROR", JOptionPane.ERROR_MESSAGE);
        				 if(ex.getMessage().startsWith("More attributes requested"))
            				 JOptionPane.showMessageDialog(null, "Max num. of attributes: "+ trainSet.numAttributes(), 
		 							 "TERMINAL ERROR", JOptionPane.ERROR_MESSAGE);
        					 
        				 throw ex;
        			}
        		}                
                                                   
                
                model.setName(this.inputFileReader.getNameForDim(index));
                model.setLabels(this.getLabels(index)); 
                model.setInstances(backupSet);

                if(this.optionsModel.isEvaluating())
                {     
                	System.out.println("self training*10 cross validation for the TRAINING set");
            		String tmpHead = "*10 cross validation for the TRAINING set";
            		try{
            			this.exportEvaluation(tmpHead,index, trainSet, null, pDialog);
            		}
            		catch(Exception ex)
            		{
            			JOptionPane.showMessageDialog(null, "Cannot evaluate " + this.classifier.getClass().getName() + ": " + 
           					 							  ex.getMessage(), "TERMINAL ERROR", JOptionPane.ERROR_MESSAGE);	
            			
                		throw ex;
            		}
                }
                this.writeOutputFiles(model, this.classifier, inputFileReader.getNameForDim(index), model.getLabels());
                this.writeARFFForDim(index, outputPostfix, this.optionsModel.getOutputPrefix(), true, pDialog);
                              
                return model;
            }
            
            
            // displays size of training/test set.
            System.out.println("TRAIN: " + trainSet.numInstances() + ", TEST: " + testDataPointSet.size());

            // /////////////////////////
            // building the classifier from the training set
            // /////////////////////////
			try{
				this.buildClassifier(trainSet, pDialog);
			}
			catch(Exception ex)
			{
				 JOptionPane.showMessageDialog(null, "Cannot build classifier " + 
						 							 this.classifier.getClass().getName() + ": " + ex.getMessage(), 
						 							 "TERMINAL ERROR", JOptionPane.ERROR_MESSAGE);     
				 if(ex.getMessage().startsWith("More attributes requested"))
    				 JOptionPane.showMessageDialog(null, "Max num. of attributes: "+ trainSet.numAttributes(), 
 							 "TERMINAL ERROR", JOptionPane.ERROR_MESSAGE);				 
				 throw ex;
			}            

            System.out.println("classifier built");
        	//output some info for debugging by Hao-Chuan 20070110
        	System.out.println("Classifier: "+this.classifier.getClass());
    	        System.out.print("Classifier parameters: \n");
    	        for(int i=0; i<this.classifier.getOptions().length; i++)
    	        {
    	            System.out.print(this.classifier.getOptions()[i]+" ");
    	        }
    	    System.out.print("\n");

            // /////////////////////////
            // applying the classifier to test set
            // /////////////////////////

            // stores the average confidence of the ALL predictions on the test
            // set.
            double confidence = 0.0;

            // maintains a sorted list of data points, sorted in descending
            // order of confidence value.
            FastVector labels = this.getLabels(index);
            GenericNodeSorter gs = GenericNodeSorter.getNewInstance();
            Iterator testDataPointIter = testDataPointSet.iterator();
            while (testDataPointIter.hasNext()) {
                DataPoint dataPoint = (DataPoint) testDataPointIter.next();
                Instance instance = this.featureSelector.getInstanceFromDataPoint(dataPoint, testSet);
                instance.setDataset(testSet);                
                instance.setClassMissing();
                
                // holds double[] of confidence values for all labels this
                // classifier is classifying into. The index with the largest
                // double will be the ultimate predicted class.
                double[] dist;
                try {
                    dist = this.classifier.distributionForInstance(instance);

                    // find the index with the largest double value.
                    double max = (double) Integer.MIN_VALUE;
                    int max_index = Integer.MIN_VALUE;
                    for (int j = 0; j < dist.length; j++) {
                        if (dist[j] > max) {
                            max = dist[j];
                            max_index = j;
                        }
                    }

                    // add the confidence of the most confident label to
                    // *confidence*
                    confidence += max;

                    String predictedLabel = labels.elementAt(max_index).toString();
                    dataPoint.setValueAtDim(index, predictedLabel);
                	System.out.println("------------------> Label: "+predictedLabel+", "+dataPoint.getValueAtDim(index));                    
                    dataPoint.setScoreForDim(index, max);
                    gs.addGenericNode(new GenericNode(dataPoint, max));
                } 
                catch (Exception e) {

                    JOptionPane.showMessageDialog(null, 
                    		"Error classifying test instance.", "TERMINAL ERROR", JOptionPane.ERROR_MESSAGE);
                    if(e.getMessage().startsWith("Can't normalize array"))
                        JOptionPane.showMessageDialog(null, 
                        		"Avoid using SMO when only one classification label is in use.", "TERMINAL ERROR", JOptionPane.ERROR_MESSAGE);                    	
                    throw e;
                }
            }

            // /////////////////////////
            // confidence = avg(data point confidence). This is the average
            // confidence of all predictions on the test set, during this
            // iteration.
            // /////////////////////////
            confidence /= ((double) testDataPointSet.size());

            // increment the number of iterations, so that we don't iterate more
            // than MAX_ITERATIONS times.
            iter++;

            if (this.optionsModel.isSelfTrain()) {
                // depending on the mode of iteraction, add the most confident
                // predictions to the test set prior to running another train-test
                // iteration.
                if (this.iterationMode == MultiDimClassifier.ADD_FIXED_NUM) {
                    this.addFixedNum(gs, index);
                } else if (this.iterationMode == MultiDimClassifier.ADD_THRESH) {
                    this.addThresh(gs, index);
                } else if (this.iterationMode == MultiDimClassifier.ADD_PERCENT) {
                    this.addPercent(gs, index);
                } else {
                    JOptionPane.showMessageDialog(null, "Unrecognized Operation Mode (This IS A BUG). " +
                    		"Contact: jaime@cs.cmu.edu", "TERMINAL ERROR", JOptionPane.ERROR_MESSAGE);
                    System.exit(0);
                }
                	
                // if the # of iterations is beyond MAX_ITERATIONS, then stop.
                if (iter > MultiDimClassifier.MAX_ITERATIONS) {
                    ClassifierModel model = new ClassifierModel();
                    model.setName(this.inputFileReader.getNameForDim(index));
                    model.setLabels(labels);
                    model.setInstances(backupSet);
                    this.writeOutputFiles(model, this.classifier, inputFileReader.getNameForDim(index), labels);
                    this.writeARFFForDim(index, outputPostfix, this.optionsModel.getOutputPrefix(), true, pDialog);//"full_set");
                    return model;
                } else {
                    // if the (avg) confidence over this test set is greater
                    // than the confidence obtained during the previous
                    // iteration, then keep going.
                    if (confidence == 1.0) {
                        ClassifierModel model = new ClassifierModel();
                        model.setName(this.inputFileReader.getNameForDim(index));
                        model.setLabels(labels);
                        model.setInstances(backupSet);
                        this.writeOutputFiles(model, this.classifier, inputFileReader.getNameForDim(index), labels);
                        this.writeARFFForDim(index, outputPostfix, this.optionsModel.getOutputPrefix(), true, pDialog);//"full_set");
                        return model;
                    } else {
                        if (confidence > prevConfidence) {
                            prevConfidence = confidence;
                            continue;
                        } else {
                            // Otherwise, stop because the confidence is
                            // dropping.
                            ClassifierModel model = new ClassifierModel();
                            model.setName(this.inputFileReader.getNameForDim(index));
                            model.setLabels(labels);
                            model.setInstances(backupSet);
                            this.writeOutputFiles(model, this.classifier, inputFileReader.getNameForDim(index), labels);
                            this.writeARFFForDim(index, outputPostfix, this.optionsModel.getOutputPrefix(), true, pDialog);//"full_set");
                            return model;
                        }
                    }
                }
            } 
            else // no self-training 
            {
                ClassifierModel model = new ClassifierModel();
                model.setName(this.inputFileReader.getNameForDim(index));
                model.setLabels(labels);
                model.setInstances(backupSet);
                
                if(this.optionsModel.isEvaluating())
                {
                	String tmpHead = "*10 cross validation for the TRAINING set";
                	try{
                		this.exportEvaluation(tmpHead, index, trainSet, testSet, pDialog);
                	}
                	catch(Exception ex)
                	{
            			JOptionPane.showMessageDialog(null, "Cannot evaluate " + this.classifier.getClass().getName() + ": " + 
		 							  ex.getMessage(), "TERMINAL ERROR", JOptionPane.ERROR_MESSAGE);				    
            			throw ex;                		
                	}
                } 
                
                this.writeOutputFiles(model, this.classifier, inputFileReader.getNameForDim(index), labels);
                this.writeARFFForDim(index, outputPostfix, this.optionsModel.getOutputPrefix(), true, pDialog);//"full_set");
                               
                return model;
            }
        }
    }

    /**
     * Add a fixed number of predictions to the test set.
     * 
     * @param gs -
     *            Sorted list of data points in test set, sorted in descending
     *            order of confidence.
     * @param index -
     *            The index of the classifier being training (0 corresponds to
     *            the first column in the input .xls file)
     */
    public void addFixedNum(GenericNodeSorter gs, int index) {
        GenericNode[] gNodes = gs.getTopGenericNodes(this.fixedNum);
        for (int i = 0; i < gNodes.length; i++) {
            DataPoint dPoint = (DataPoint) gNodes[i].getO();
            System.out.println("adding prediction: " + dPoint.getValueAtDim(index) + "," + dPoint.getScoreForDim(index));
            dPoint.setIsReferenceAtDim(index);
        }
    }

    /**
     * Add all predictions with confidence above a certain threshold
     * (*this.confThresh*)to the test set.
     * 
     * @param gs -
     *            Sorted list of data points in test set, sorted in descending
     *            order of confidence.
     * @param index -
     *            The index of the classifier being training (0 corresponds to
     *            the first column in the input .xls file)
     */
    public void addThresh(GenericNodeSorter gs, int index) {
        GenericNode[] gNodes = gs.getAllGenericNodes();
        for (int i = 0; i < gNodes.length; i++) {
            DataPoint dPoint = (DataPoint) gNodes[i].getO();
            double conf = dPoint.getScoreForDim(index);
            if (conf > this.confThresh) {
                dPoint.setIsReferenceAtDim(index);
                System.out.println("adding prediction: " + dPoint.getValueAtDim(index) + ": " + dPoint.getScoreForDim(index));
            }
        }
    }

    /**
     * Add a percentage of the test set to the training set (add the most
     * confident predictions).
     * 
     * @param gs -
     *            Sorted list of data points in test set, sorted in descending
     *            order of confidence.
     * @param index -
     *            The index of the classifier being training (0 corresponds to
     *            the first column in the input .xls file)
     */
    public void addPercent(GenericNodeSorter gs, int index) {
        int size = gs.getSize();
        int addSize = (int) (((double) size) * this.percent);
        GenericNode[] gNodes = gs.getTopGenericNodes(addSize);
        for (int i = 0; i < gNodes.length; i++) {
            DataPoint dPoint = (DataPoint) gNodes[i].getO();
            dPoint.setIsReferenceAtDim(index);
            System.out.println("adding prediction: " + dPoint.getValueAtDim(index) + ": " + dPoint.getScoreForDim(index));
        }
    }

    /**
     * Writes the .dic and .model files. A number of objects are serialized for later reload and reuse (by SavedModelEvaluator).
     * 
     * @param model -
     *            ClassifierModel that contains the classification settings.
     * @param classifier -
     *            Weka classifier to be serialized.
     * @param name -
     *            The name of the classifier (originated from the label in the
     *            corresponding column in the input .xls file)
     * @param labels -
     *            The values of the classes that this classifier classifies
     *            into.
     */
    public void writeOutputFiles(ClassifierModel model, Classifier classifier, String name, FastVector labels) {
        try {
        	//temporary use for avoiding unserializable issue, 2007 April by Hao-Chuan
        	this.optionsModel.clearPreprocessed();
        	
            String username = System.getProperty("user.name");

            String prefix = this.optionsModel.getOutputPrefix().equals("")?"":this.optionsModel.getOutputPrefix()+"_";
            File dir = new File(MultiDimClassifier.CLASSIFIERS + "/" + prefix + name);
            dir.mkdir();
            
            
            //creating directory for scripts
            File script_dir = new File(MultiDimClassifier.SCRIPTS);
            if(!script_dir.exists())
            	script_dir.mkdir();
            File script_file = new File(script_dir.getAbsolutePath() + "/" + name + "_" + username + ".script");
            PrintWriter script_out = new PrintWriter(new FileOutputStream(script_file));
            script_out.println("<classifier>");
            script_out.println("	@Classifier_Name " + name + "_" + username);
            
            

            // creating dictionary file
            File dicFile = new File(dir.getAbsolutePath() + "/" + name + "_" + username + ".dic");
            model.setDicFile(dicFile);
            this.featureSelector.writeDicFile(dicFile.getAbsolutePath(), labels);
            script_out.println("	@Dictionary " + dicFile.getName());
            
            // writting model
            File modFile = new File(dir.getAbsolutePath() + "/" + name + "_" + username + ".model");
            model.setModFile(modFile);
            FileOutputStream fos = new FileOutputStream(modFile);
            ObjectOutputStream oos = new ObjectOutputStream(fos);
            oos.writeObject(classifier);
            oos.flush();
            oos.close();
            script_out.println("	@Classifier_MOD " + modFile.getName());
            
            //writing optionsModel
            File omFile = new File(dir.getAbsolutePath() + "/" + name + "_" + username + ".om");
            FileOutputStream fos2 = new FileOutputStream(omFile);
            ObjectOutputStream oos2 = new ObjectOutputStream(fos2);
            oos2.writeObject(this.optionsModel);
            oos2.flush();
            oos2.close();     
            script_out.println("	@Classifier_OM " + omFile.getName());
            
            
            //writing Instances
            File insFile = new File(dir.getAbsolutePath() + "/" + name + "_" + username + ".ins");
            FileOutputStream fos3 = new FileOutputStream(insFile);
            ObjectOutputStream oos3 = new ObjectOutputStream(fos3);
            oos3.writeObject(model.getInstances());
            oos3.flush();
            oos3.close();               
            script_out.println("	@Classifier_INS " + insFile.getName());

            //writing featureSet
            File featureFile = new File(dir.getAbsolutePath() + "/" + name + "_" + username + ".feature");
            FileOutputStream fos4 = new FileOutputStream(featureFile);
            ObjectOutputStream oos4 = new ObjectOutputStream(fos4);
            oos4.writeObject(this.optionsModel.getFeatureSet());
            oos4.flush();
            oos4.close();    
            script_out.println("	@Classifier_FEATURE " + featureFile.getName());
            
            //writing nominal feature names
            File nomFeatureFile = new File(dir.getAbsolutePath() + "/" + name + "_" + username + ".nfn");
            FileOutputStream fos5 = new FileOutputStream(nomFeatureFile);
            ObjectOutputStream oos5 = new ObjectOutputStream(fos5);
            oos5.writeObject(this.optionsModel.getNominalFeatureNames());
            oos5.flush();
            oos5.close();            
            script_out.println("	@Classifier_NFN " + nomFeatureFile.getName());
            
            //writing nominal feature value set
            File nomFeatureVSFile = new File(dir.getAbsolutePath() + "/" + name + "_" + username + ".nfvs");
            FileOutputStream fos6 = new FileOutputStream(nomFeatureVSFile);
            ObjectOutputStream oos6 = new ObjectOutputStream(fos6);
            oos6.writeObject(this.optionsModel.getNominalFeatureValueSet());
            oos6.flush();
            oos6.close();    
            script_out.println("	@Classifier_NFVS " + nomFeatureVSFile.getName());
            
            //writing user defined features
            File userFeatFile = new File(dir.getAbsolutePath() + "/" + name + "_" + username + ".udf");
            FileOutputStream fos7 = new FileOutputStream(userFeatFile);
            ObjectOutputStream oos7 = new ObjectOutputStream(fos7);
            oos7.writeObject(this.featureSelector.getGlobalUserDefinedFeature());
            oos7.flush();
            oos7.close(); 
            script_out.println("	@Classifier_UDF " + userFeatFile.getName());
            
            //script file: writing class labels
            String label_str = "";
            if(labels != null)
            {
    	        for (int i = 0; i < labels.size(); i++) {
    	            label_str = label_str + labels.elementAt(i).toString() + " ";
    	        }
            }
            script_out.println("	@Class_Labels " + label_str);
            script_out.println("</classifier>");
            script_out.close();
            
            //Similarly, but writing user defined features to text files to the output folder
            if(!this.featureSelector.getUserDefinedText().equals(""))
            {
	    		String path= MultiDimClassifier.OUTPUT+"/";
	    		if(!this.optionsModel.getOutputPrefix().equals(""))
	    		{
	    			path = path+this.optionsModel.getOutputPrefix()+"_";
	    		}                		
	    		path = path+"UserDefinedFeatures_";
	    		path = path +this.optionsModel.getArffFileName().replaceAll("arff", "txt");
	            
	            File userFeatTextFile = new File(path);
	            FileWriter writer8 = new FileWriter(userFeatTextFile);
	            writer8.write(this.featureSelector.getUserDefinedText());
	            writer8.close();
            } 
                                               
        } 
        catch (IOException e) 
        {
            e.printStackTrace();
            System.exit(0);
        }
    }
   
    /**
     * 
     * this is a protected method not to be accessed publicly.
     * Writes the predictions and confidence values to a XLS file.
     * 
     * @param output -- specifying the output file
     * @param csvOutput -- should the output be in CSV format?
     */
    protected void writeOutput(File output, boolean csvOutput) {
    	String specific_sep = csvOutput ? "," : "&##&";
    	String escapeForDoubleQuote = csvOutput ? "\\" : "\"";
        Logger logger = new Logger(output.getAbsolutePath());
        String line = "";
        for (int i = 0; i < DataPoint.dimNum; i++) {
            line += (this.inputFileReader.getNameForDim(i) + specific_sep + 
            		"prediction confidence" + specific_sep);
        }
        line += "text";
        logger.writeLine(line);
        Iterator dataPointIter = dataPoints.iterator();
        while (dataPointIter.hasNext()) {
            DataPoint dataPoint = (DataPoint) dataPointIter.next();
            line = "";
            for (int i = 0; i < DataPoint.dimNum; i++) {
                line += (dataPoint.getValueAtDim(i) + specific_sep + 
                		dataPoint.getScoreForDim(i) + specific_sep);
            }
            
            if (csvOutput) {
              // text always goes in double quotes in the CSV
              // insert the correct escape character for the " (double quote)
              line += dataPoint.getQuotedText(specific_sep);
            } else {
            	// no enclosing in double quotes for XLS
            	// no escaping the double quotes
            	line += dataPoint.getText();
            }
            logger.writeLine(line);
        }
        logger.close();
        
        //convert CSV to XLS
        if (!csvOutput) {
          Reader tmpCSV = new Reader(output.getAbsolutePath());
          InputFileReader.convertCSVToXLS(tmpCSV, specific_sep);
          tmpCSV.close();
          output.delete();
        }
    }
    
    /*
     * schaudhu @ 08/30/2007
     * crossValidateModel function emulating weka to implement 
     * user aids during training process
     */
    
    /**
     * Update the cumulative record of classification margins
     *
     * @param predictedDistribution the probability distribution predicted for
     * the current instance
     * @param actualClass the index of the actual instance class
     * @param weight the weight assigned to the instance
     */
    protected void updateMargins(double [] predictedDistribution, 
    	      int actualClass, double weight) {

    	    double probActual = predictedDistribution[actualClass];
    	    double probNext = 0;

    	    for(int i = 0; i < m_NumClasses; i++)
    	      if ((i != actualClass) &&
    		  (predictedDistribution[i] > probNext))
    		probNext = predictedDistribution[i];

    	    double margin = probActual - probNext;
    	    int bin = (int)((margin + 1.0) / 2.0 * k_MarginResolution);
    	    m_MarginCounts[bin] += weight;
    	  }
    
    /**
     * Update the numeric accuracy measures. For numeric classes, the
     * accuracy is between the actual and predicted class values. For 
     * nominal classes, the accuracy is between the actual and 
     * predicted class probabilities.
     *
     * @param predicted the predicted values
     * @param actual the actual value
     * @param weight the weight associated with this prediction
     */
    protected void updateNumericScores(double [] predicted, 
    	      double [] actual, double weight) {

    	    double diff;
    	    double sumErr = 0, sumAbsErr = 0, sumSqrErr = 0;
    	    double sumPriorAbsErr = 0, sumPriorSqrErr = 0;
    	    for(int i = 0; i < m_NumClasses; i++) {
    	      diff = predicted[i] - actual[i];
    	      sumErr += diff;
    	      sumAbsErr += Math.abs(diff);
    	      sumSqrErr += diff * diff;
    	      diff = (m_ClassPriors[i] / m_ClassPriorsSum) - actual[i];
    	      sumPriorAbsErr += Math.abs(diff);
    	      sumPriorSqrErr += diff * diff;
    	    }
    	    m_SumErr += weight * sumErr / m_NumClasses;
    	    m_SumAbsErr += weight * sumAbsErr / m_NumClasses;
    	    m_SumSqrErr += weight * sumSqrErr / m_NumClasses;
    	    m_SumPriorAbsErr += weight * sumPriorAbsErr / m_NumClasses;
    	    m_SumPriorSqrErr += weight * sumPriorSqrErr / m_NumClasses;
    	  }
    
    /**
     * Adds a numeric (non-missing) training class value and weight to 
     * the buffer of stored values.
     *
     * @param classValue the class value
     * @param weight the instance weight
     */
    protected void addNumericTrainClass(double classValue, double weight) {

        if (m_TrainClassVals == null) {
          m_TrainClassVals = new double [100];
          m_TrainClassWeights = new double [100];
        }
        if (m_NumTrainClassVals == m_TrainClassVals.length) {
          double [] temp = new double [m_TrainClassVals.length * 2];
          System.arraycopy(m_TrainClassVals, 0, 
    	  temp, 0, m_TrainClassVals.length);
          m_TrainClassVals = temp;

          temp = new double [m_TrainClassWeights.length * 2];
          System.arraycopy(m_TrainClassWeights, 0, 
    	  temp, 0, m_TrainClassWeights.length);
          m_TrainClassWeights = temp;
        }
        m_TrainClassVals[m_NumTrainClassVals] = classValue;
        m_TrainClassWeights[m_NumTrainClassVals] = weight;
        m_NumTrainClassVals++;
      }

    /**
     * Sets up the priors for numeric class attributes from the 
     * training class values that have been seen so far.
     */
    protected void setNumericPriorsFromBuffer() {

        double numPrecision = 0.01; // Default value
        if (m_NumTrainClassVals > 1) {
          double [] temp = new double [m_NumTrainClassVals];
          System.arraycopy(m_TrainClassVals, 0, temp, 0, m_NumTrainClassVals);
          int [] index = Utils.sort(temp);
          double lastVal = temp[index[0]];
          double deltaSum = 0;
          int distinct = 0;
          for (int i = 1; i < temp.length; i++) {
    	double current = temp[index[i]];
    	if (current != lastVal) {
    	  deltaSum += current - lastVal;
    	  lastVal = current;
    	  distinct++;
    	}
          }
          if (distinct > 0) {
    	numPrecision = deltaSum / distinct;
          }
        }
        m_PriorErrorEstimator = new KernelEstimator(numPrecision);
        m_ErrorEstimator = new KernelEstimator(numPrecision);
        m_ClassPriors[0] = 0;
       	m_ClassPriorsSum = 0;
        for (int i = 0; i < m_NumTrainClassVals; i++) {
          m_ClassPriors[0] += m_TrainClassVals[i] * m_TrainClassWeights[i];
          m_ClassPriorsSum += m_TrainClassWeights[i];
          m_PriorErrorEstimator.addValue(m_TrainClassVals[i],
    	  m_TrainClassWeights[i]);
        }
      }

    /**
     * Sets the class prior probabilities
     *
     * @param train the training instances used to determine
     * the prior probabilities
     * @throws Exception if the class attribute of the instances is not
     * set
     */
    public void setPriors(Instances train) throws Exception {
        m_NoPriors = false;

        if (!m_ClassIsNominal) {

          m_NumTrainClassVals = 0;
          m_TrainClassVals = null;
          m_TrainClassWeights = null;
          m_PriorErrorEstimator = null;
          m_ErrorEstimator = null;

          for (int i = 0; i < train.numInstances(); i++) {
    	Instance currentInst = train.instance(i);
    	if (!currentInst.classIsMissing()) {
    	  addNumericTrainClass(currentInst.classValue(), 
    	      currentInst.weight());
    	}
          }

        } else {
          for (int i = 0; i < m_NumClasses; i++) {
    	m_ClassPriors[i] = 1;
          }
          m_ClassPriorsSum = m_NumClasses;
          for (int i = 0; i < train.numInstances(); i++) {
    	if (!train.instance(i).classIsMissing()) {
    	  m_ClassPriors[(int)train.instance(i).classValue()] += 
    	    train.instance(i).weight();
    	  m_ClassPriorsSum += train.instance(i).weight();
    	}
          }
        }
      }
    
    /**
     * Evaluates the classifier on a given set of instances. Note that
     * the data must have exactly the same format (e.g. order of
     * attributes) as the data used to train the classifier! Otherwise
     * the results will generally be meaningless.
     *
     * @param classifier machine learning classifier
     * @param data set of test instances for evaluation
     * @return the predictions
     * @throws Exception if model could not be evaluated 
     * successfully 
     */
    public double[] evaluateModel(Classifier classifier,
    	      Instances data) throws Exception {

    	    double predictions[] = new double[data.numInstances()];

    	    // Need to be able to collect predictions if appropriate (for AUC)

    	    for (int i = 0; i < data.numInstances(); i++) {
    	      predictions[i] = evaluateModelOnceAndRecordPrediction((Classifier)classifier, 
    		  data.instance(i));
    	    }

    	    return predictions;
    	  }
    
    /**
     * Updates all the statistics about a classifiers performance for 
     * the current test instance.
     *
     * @param predictedDistribution the probabilities assigned to 
     * each class
     * @param instance the instance to be classified
     * @throws Exception if the class of the instance is not
     * set
     */
    protected void updateStatsForClassifier(double [] predictedDistribution,
    	      Instance instance)
    	  throws Exception {

    	    int actualClass = (int)instance.classValue();

    	    if (!instance.classIsMissing()) {
    	      updateMargins(predictedDistribution, actualClass, instance.weight());

    	      // Determine the predicted class (doesn't detect multiple 
    	      // classifications)
    	      int predictedClass = -1;
    	      double bestProb = 0.0;
    	      for(int i = 0; i < m_NumClasses; i++) {
    		if (predictedDistribution[i] > bestProb) {
    		  predictedClass = i;
    		  bestProb = predictedDistribution[i];
    		}
    	      }

    	      m_WithClass += instance.weight();

    	      // Determine misclassification cost
    	      if (m_CostMatrix != null) {
    		if (predictedClass < 0) {
    		  // For missing predictions, we assume the worst possible cost.
    		  // This is pretty harsh.
    		  // Perhaps we could take the negative of the cost of a correct
    		  // prediction (-m_CostMatrix.getElement(actualClass,actualClass)),
    		  // although often this will be zero
    		  m_TotalCost += instance.weight()
    		  * m_CostMatrix.getMaxCost(actualClass, instance);
    		} else {
    		  m_TotalCost += instance.weight() 
    		  * m_CostMatrix.getElement(actualClass, predictedClass,
    		      instance);
    		}
    	      }

    	      // Update counts when no class was predicted
    	      if (predictedClass < 0) {
    		m_Unclassified += instance.weight();
    		return;
    	      }

    	      double predictedProb = Math.max(MIN_SF_PROB,
    		  predictedDistribution[actualClass]);
    	      double priorProb = Math.max(MIN_SF_PROB,
    		  m_ClassPriors[actualClass]
    		                / m_ClassPriorsSum);
    	      if (predictedProb >= priorProb) {
    		m_SumKBInfo += (Utils.log2(predictedProb) - 
    		    Utils.log2(priorProb))
    		    * instance.weight();
    	      } else {
    		m_SumKBInfo -= (Utils.log2(1.0-predictedProb) - 
    		    Utils.log2(1.0-priorProb))
    		    * instance.weight();
    	      }

    	      m_SumSchemeEntropy -= Utils.log2(predictedProb) * instance.weight();
    	      m_SumPriorEntropy -= Utils.log2(priorProb) * instance.weight();

    	      updateNumericScores(predictedDistribution, 
    		  makeDistribution(instance.classValue()), 
    		  instance.weight());

    	      // Update other stats
    	      m_ConfusionMatrix[actualClass][predictedClass] += instance.weight();
    	      if (predictedClass != actualClass) {
    		m_Incorrect += instance.weight();
    	      } else {
    		m_Correct += instance.weight();
    	      }
    	    } else {
    	      m_MissingClass += instance.weight();
    	    }
    	  }
    
    /**
     * Convert a single prediction into a probability distribution
     * with all zero probabilities except the predicted value which
     * has probability 1.0;
     *
     * @param predictedClass the index of the predicted class
     * @return the probability distribution
     */
    protected double [] makeDistribution(double predictedClass) {

        double [] result = new double [m_NumClasses];
        if (Instance.isMissingValue(predictedClass)) {
          return result;
        }
        if (m_ClassIsNominal) {
          result[(int)predictedClass] = 1.0;
        } else {
          result[0] = predictedClass;
        }
        return result;
      } 

    /**
     * Evaluates the classifier on a single instance and records the
     * prediction (if the class is nominal).
     *
     * @param classifier machine learning classifier
     * @param instance the test instance to be classified
     * @return the prediction made by the clasifier
     * @throws Exception if model could not be evaluated 
     * successfully or the data contains string attributes
     */
    public double evaluateModelOnceAndRecordPrediction(Classifier classifier,
    	      Instance instance) throws Exception {

    	    Instance classMissing = (Instance)instance.copy();
    	    double pred = 0;
    	    classMissing.setDataset(instance.dataset());
    	    classMissing.setClassMissing();
    	    if (m_ClassIsNominal) {
    	      if (m_Predictions == null) {
    		m_Predictions = new FastVector();
    	      }
    	      double [] dist = classifier.distributionForInstance(classMissing);
    	      pred = Utils.maxIndex(dist);
    	      if (dist[(int)pred] <= 0) {
    		pred = Instance.missingValue();
    	      }
    	      updateStatsForClassifier(dist, instance);
    	      m_Predictions.addElement(new NominalPrediction(instance.classValue(), dist, 
    		  instance.weight()));
    	    } else {
    	      pred = classifier.classifyInstance(classMissing);
    	      updateStatsForPredictor(pred, instance);
    	    }
    	    return pred;
    	  }

    /**
     * Updates all the statistics about a predictors performance for 
     * the current test instance.
     *
     * @param predictedValue the numeric value the classifier predicts
     * @param instance the instance to be classified
     * @throws Exception if the class of the instance is not
     * set
     */
    protected void updateStatsForPredictor(double predictedValue,
    	      Instance instance) 
    	  throws Exception {

    	    if (!instance.classIsMissing()){

    	      // Update stats
    	      m_WithClass += instance.weight();
    	      if (Instance.isMissingValue(predictedValue)) {
    		m_Unclassified += instance.weight();
    		return;
    	      }
    	      m_SumClass += instance.weight() * instance.classValue();
    	      m_SumSqrClass += instance.weight() * instance.classValue()
    	      *	instance.classValue();
    	      m_SumClassPredicted += instance.weight() 
    	      * instance.classValue() * predictedValue;
    	      m_SumPredicted += instance.weight() * predictedValue;
    	      m_SumSqrPredicted += instance.weight() * predictedValue * predictedValue;

    	      if (m_ErrorEstimator == null) {
    		setNumericPriorsFromBuffer();
    	      }
    	      double predictedProb = Math.max(m_ErrorEstimator.getProbability(
    		  predictedValue 
    		  - instance.classValue()),
    		  MIN_SF_PROB);
    	      double priorProb = Math.max(m_PriorErrorEstimator.getProbability(
    		  instance.classValue()),
    		  MIN_SF_PROB);

    	      m_SumSchemeEntropy -= Utils.log2(predictedProb) * instance.weight();
    	      m_SumPriorEntropy -= Utils.log2(priorProb) * instance.weight();
    	      m_ErrorEstimator.addValue(predictedValue - instance.classValue(), 
    		  instance.weight());

    	      updateNumericScores(makeDistribution(predictedValue),
    		  makeDistribution(instance.classValue()),
    		  instance.weight());

    	    } else
    	      m_MissingClass += instance.weight();
    	  }
    
    /*
     * Used to initialize the variables that are created and/or 
     * initialized within the constructor of Evaluation.java
     */
    void init_newVars(Instances data)
    {
    	m_NumClasses = data.numClasses();
        m_NumFolds = 1;
        m_ClassIsNominal = data.classAttribute().isNominal();

        if (m_ClassIsNominal) {
          m_ConfusionMatrix = new double [m_NumClasses][m_NumClasses];
          m_ClassNames = new String [m_NumClasses];
          for(int i = 0; i < m_NumClasses; i++) {
    	m_ClassNames[i] = data.classAttribute().value(i);
          }
        }
        /*m_CostMatrix = costMatrix;
        if (m_CostMatrix != null) {
          if (!m_ClassIsNominal) {
    	throw new Exception("Class has to be nominal if cost matrix " + 
    	"given!");
          }
          if (m_CostMatrix.size() != m_NumClasses) {
    	throw new Exception("Cost matrix not compatible with data!");
          }
        }*/
        m_ClassPriors = new double [m_NumClasses];
        try {
			setPriors(data);
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
        m_MarginCounts = new double [k_MarginResolution + 1];
    }
    
    /**
     * Returns the correlation coefficient if the class is numeric.
     *
     * @return the correlation coefficient
     * @throws Exception if class is not numeric
     */
    public final double correlationCoefficient() throws Exception {

      if (m_ClassIsNominal) {
        throw
        new Exception("Can't compute correlation coefficient: " + 
        "class is nominal!");
      }

      double correlation = 0;
      double varActual = 
        m_SumSqrClass - m_SumClass * m_SumClass / m_WithClass;
      double varPredicted = 
        m_SumSqrPredicted - m_SumPredicted * m_SumPredicted / 
        m_WithClass;
      double varProd = 
        m_SumClassPredicted - m_SumClass * m_SumPredicted / m_WithClass;

      if (varActual * varPredicted <= 0) {
        correlation = 0.0;
      } else {
        correlation = varProd / Math.sqrt(varActual * varPredicted);
      }

      return correlation;
    }

    /**
     * Returns the mean absolute error. Refers to the error of the
     * predicted values for numeric classes, and the error of the 
     * predicted probability distribution for nominal classes.
     *
     * @return the mean absolute error 
     */
    public final double meanAbsoluteError() {

      return m_SumAbsErr / m_WithClass;
    }

    /**
     * Returns the mean absolute error of the prior.
     *
     * @return the mean absolute error 
     */
    public final double meanPriorAbsoluteError() {

      if (m_NoPriors)
        return Double.NaN;

      return m_SumPriorAbsErr / m_WithClass;
    }

    /**
     * Returns the relative absolute error.
     *
     * @return the relative absolute error 
     * @throws Exception if it can't be computed
     */
    public final double relativeAbsoluteError() throws Exception {

      if (m_NoPriors)
        return Double.NaN;

      return 100 * meanAbsoluteError() / meanPriorAbsoluteError();
    }

    /**
     * Returns the root mean squared error.
     *
     * @return the root mean squared error 
     */
    public final double rootMeanSquaredError() {

      return Math.sqrt(m_SumSqrErr / m_WithClass);
    }

    /**
     * Returns the root mean prior squared error.
     *
     * @return the root mean prior squared error 
     */
    public final double rootMeanPriorSquaredError() {

      if (m_NoPriors)
        return Double.NaN;

      return Math.sqrt(m_SumPriorSqrErr / m_WithClass);
    }

    /**
     * Returns the root relative squared error if the class is numeric.
     *
     * @return the root relative squared error 
     */
    public final double rootRelativeSquaredError() {

      if (m_NoPriors)
        return Double.NaN;

      return 100.0 * rootMeanSquaredError() / 
      rootMeanPriorSquaredError();
    }

    /**
     * Calculate the entropy of the prior distribution
     *
     * @return the entropy of the prior distribution
     * @throws Exception if the class is not nominal
     */
    public final double priorEntropy() throws Exception {

      if (!m_ClassIsNominal) {
        throw
        new Exception("Can't compute entropy of class prior: " + 
        "class numeric!");
      }

      if (m_NoPriors)
        return Double.NaN;

      double entropy = 0;
      for(int i = 0; i < m_NumClasses; i++) {
        entropy -= m_ClassPriors[i] / m_ClassPriorsSum 
        * Utils.log2(m_ClassPriors[i] / m_ClassPriorsSum);
      }
      return entropy;
    }

    /**
     * Return the total Kononenko & Bratko Information score in bits
     *
     * @return the K&B information score
     * @throws Exception if the class is not nominal
     */
    public final double KBInformation() throws Exception {

      if (!m_ClassIsNominal) {
        throw
        new Exception("Can't compute K&B Info score: " + 
        "class numeric!");
      }

      if (m_NoPriors)
        return Double.NaN;

      return m_SumKBInfo;
    }

    /**
     * Return the Kononenko & Bratko Information score in bits per 
     * instance.
     *
     * @return the K&B information score
     * @throws Exception if the class is not nominal
     */
    public final double KBMeanInformation() throws Exception {

      if (!m_ClassIsNominal) {
        throw
        new Exception("Can't compute K&B Info score: "
  	  + "class numeric!");
      }

      if (m_NoPriors)
        return Double.NaN;

      return m_SumKBInfo / m_WithClass;
    }

    /**
     * Return the Kononenko & Bratko Relative Information score
     *
     * @return the K&B relative information score
     * @throws Exception if the class is not nominal
     */
    public final double KBRelativeInformation() throws Exception {

      if (!m_ClassIsNominal) {
        throw
        new Exception("Can't compute K&B Info score: " + 
        "class numeric!");
      }

      if (m_NoPriors)
        return Double.NaN;

      return 100.0 * KBInformation() / priorEntropy();
    }

    /**
     * Returns the total entropy for the null model
     * 
     * @return the total null model entropy
     */
    public final double SFPriorEntropy() {

      if (m_NoPriors)
        return Double.NaN;

      return m_SumPriorEntropy;
    }

    /**
     * Returns the entropy per instance for the null model
     * 
     * @return the null model entropy per instance
     */
    public final double SFMeanPriorEntropy() {

      if (m_NoPriors)
        return Double.NaN;

      return m_SumPriorEntropy / m_WithClass;
    }

    /**
     * Returns the total entropy for the scheme
     * 
     * @return the total scheme entropy
     */
    public final double SFSchemeEntropy() {

      if (m_NoPriors)
        return Double.NaN;

      return m_SumSchemeEntropy;
    }

    /**
     * Returns the entropy per instance for the scheme
     * 
     * @return the scheme entropy per instance
     */
    public final double SFMeanSchemeEntropy() {

      if (m_NoPriors)
        return Double.NaN;

      return m_SumSchemeEntropy / m_WithClass;
    }

    /**
     * Returns the total SF, which is the null model entropy minus
     * the scheme entropy.
     * 
     * @return the total SF
     */
    public final double SFEntropyGain() {

      if (m_NoPriors)
        return Double.NaN;

      return m_SumPriorEntropy - m_SumSchemeEntropy;
    }

    /**
     * Returns the SF per instance, which is the null model entropy
     * minus the scheme entropy, per instance.
     * 
     * @return the SF per instance
     */
    public final double SFMeanEntropyGain() {

      if (m_NoPriors)
        return Double.NaN;

      return (m_SumPriorEntropy - m_SumSchemeEntropy) / m_WithClass;
    }

    /**
     * Calls toMatrixString() with a default title.
     *
     * @return the confusion matrix as a string
     * @throws Exception if the class is numeric
     */
    public String toMatrixString() throws Exception {

      return toMatrixString("=== Confusion Matrix ===\n");
    }

    /**
     * Method for generating indices for the confusion matrix.
     *
     * @param num 	integer to format
     * @param IDChars	the characters to use
     * @param IDWidth	the width of the entry
     * @return 		the formatted integer as a string
     */
    protected String num2ShortID(int num, char[] IDChars, int IDWidth) {

      char ID [] = new char [IDWidth];
      int i;

      for(i = IDWidth - 1; i >=0; i--) {
        ID[i] = IDChars[num % IDChars.length];
        num = num / IDChars.length - 1;
        if (num < 0) {
  	break;
        }
      }
      for(i--; i >= 0; i--) {
        ID[i] = ' ';
      }

      return new String(ID);
    }
    
    /**
     * Outputs the performance statistics as a classification confusion
     * matrix. For each class value, shows the distribution of 
     * predicted class values.
     *
     * @param title the title for the confusion matrix
     * @return the confusion matrix as a String
     * @throws Exception if the class is numeric
     */
    public String toMatrixString(String title) throws Exception {

      StringBuffer text = new StringBuffer();
      char [] IDChars = {'a','b','c','d','e','f','g','h','i','j',
  	'k','l','m','n','o','p','q','r','s','t',
  	'u','v','w','x','y','z'};
      int IDWidth;
      boolean fractional = false;

      if (!m_ClassIsNominal) {
        throw new Exception("Evaluation: No confusion matrix possible!");
      }

      // Find the maximum value in the matrix
      // and check for fractional display requirement 
      double maxval = 0;
      for(int i = 0; i < m_NumClasses; i++) {
        for(int j = 0; j < m_NumClasses; j++) {
  	double current = m_ConfusionMatrix[i][j];
  	if (current < 0) {
  	  current *= -10;
  	}
  	if (current > maxval) {
  	  maxval = current;
  	}
  	double fract = current - Math.rint(current);
  	if (!fractional
  	    && ((Math.log(fract) / Math.log(10)) >= -2)) {
  	  fractional = true;
  	}
        }
      }

      IDWidth = 1 + Math.max((int)(Math.log(maxval) / Math.log(10) 
  	+ (fractional ? 3 : 0)),
  	(int)(Math.log(m_NumClasses) / 
  	    Math.log(IDChars.length)));
      text.append(title).append("\n");
      for(int i = 0; i < m_NumClasses; i++) {
        if (fractional) {
  	text.append(" ").append(num2ShortID(i,IDChars,IDWidth - 3))
  	.append("   ");
        } else {
  	text.append(" ").append(num2ShortID(i,IDChars,IDWidth));
        }
      }
      text.append("   <-- classified as\n");
      for(int i = 0; i< m_NumClasses; i++) { 
        for(int j = 0; j < m_NumClasses; j++) {
  	text.append(" ").append(
  	    Utils.doubleToString(m_ConfusionMatrix[i][j],
  		IDWidth,
  		(fractional ? 2 : 0)));
        }
        text.append(" | ").append(num2ShortID(i,IDChars,IDWidth))
        .append(" = ").append(m_ClassNames[i]).append("\n");
      }
      return text.toString();
    }

    /**
     * Output the cumulative margin distribution as a string suitable
     * for input for gnuplot or similar package.
     *
     * @return the cumulative margin distribution
     * @throws Exception if the class attribute is nominal
     */
    public String toCumulativeMarginDistributionString() throws Exception {

      if (!m_ClassIsNominal) {
        throw new Exception("Class must be nominal for margin distributions");
      }
      String result = "";
      double cumulativeCount = 0;
      double margin;
      for(int i = 0; i <= k_MarginResolution; i++) {
        if (m_MarginCounts[i] != 0) {
  	cumulativeCount += m_MarginCounts[i];
  	margin = (double)i * 2.0 / k_MarginResolution - 1.0;
  	result = result + Utils.doubleToString(margin, 7, 3) + ' ' 
  	+ Utils.doubleToString(cumulativeCount * 100 
  	    / m_WithClass, 7, 3) + '\n';
        } else if (i == 0) {
  	result = Utils.doubleToString(-1.0, 7, 3) + ' ' 
  	+ Utils.doubleToString(0, 7, 3) + '\n';
        }
      }
      return result;
    }

    /**
     * Performs a (stratified if class is nominal) cross-validation 
     * for a classifier on a set of instances. Now performs
     * a deep copy of the classifier before each call to 
     * buildClassifier() (just in case the classifier is not
     * initialized properly).
     *
     * @param classifier the classifier with any options set.
     * @param data the data on which the cross-validation is to be 
     * performed 
     * @param numFolds the number of folds for the cross-validation
     * @param random random number generator for randomization 
     * @throws Exception if a classifier could not be generated 
     * successfully or the class is not defined
     */
    public void crossValidateModel(Classifier classifier,
    	      Instances data, int numFolds, Random random, ProgressDialog pDialog) 
    	  throws Exception {

    	    // Make a copy of the data we can reorder
    	    data = new Instances(data);
    	    data.randomize(random);
    	    if (data.classAttribute().isNominal()) {
    	      data.stratify(numFolds);
    	    }
    	    // Do the folds
    	    int foldsDone;
    	    if (pDialog != null) pDialog.setLabel("Performing cross validation.");
    	    
    	    for (int i = 0; i < numFolds; i++) {
    	      Instances train = data.trainCV(numFolds, i, random);
    	      setPriors(train);
    	      Classifier copiedClassifier = Classifier.makeCopy(classifier);
    	      copiedClassifier.buildClassifier(train);
    	      Instances test = data.testCV(numFolds, i);
    	      evaluateModel(copiedClassifier, test);
    	      foldsDone = i+1;
    	      if (pDialog != null) pDialog.setLabel("Performing cross validation on fold " + foldsDone + " out of " + numFolds);
      	      int pctDone = (foldsDone/numFolds)*100;
      	      if (pDialog != null) pDialog.setValue(pctDone); 
      	      //calc kappa here
      	      String kappa = String.valueOf(kappa());
      	      System.out.println("Kappa for dimension = " + dimn + " after fold " + i + "= " + kappa + " with percent correct = " + pctCorrect() + "%");
      	      if (pDialog != null) pDialog.setLabel("Kappa for dimension = " + dimn + " after fold " + foldsDone + "= " + kappa + " with percent correct = " + pctCorrect() + "%");

    	    }
    	    m_NumFolds = numFolds;
    	  }
    
    /**
     * Gets the number of instances correctly classified (that is, for
     * which a correct prediction was made). (Actually the sum of the weights
     * of these instances)
     *
     * @return the number of correctly classified instances
     */
    public final double correct() {

      return m_Correct;
    }

    /**
     * Gets the percentage of instances correctly classified (that is, for
     * which a correct prediction was made).
     *
     * @return the percent of correctly classified instances (between 0 and 100)
     */
    public final double pctCorrect() {

      return 100 * m_Correct / m_WithClass;
    }
    
    /**
     * Gets the number of instances incorrectly classified (that is, for
     * which an incorrect prediction was made). (Actually the sum of the weights
     * of these instances)
     *
     * @return the number of incorrectly classified instances 
     */
    public final double incorrect() {

      return m_Incorrect;
    }

    /**
     * Gets the percentage of instances incorrectly classified (that is, for
     * which an incorrect prediction was made).
     *
     * @return the percent of incorrectly classified instances 
     * (between 0 and 100)
     */
    public final double pctIncorrect() {

      return 100 * m_Incorrect / m_WithClass;
    }
    
    /**
     * Returns value of kappa statistic if class is nominal.
     *
     * @return the value of the kappa statistic
     */
    public final double kappa() {


      double[] sumRows = new double[m_ConfusionMatrix.length];
      double[] sumColumns = new double[m_ConfusionMatrix.length];
      double sumOfWeights = 0;
      for (int i = 0; i < m_ConfusionMatrix.length; i++) {
        for (int j = 0; j < m_ConfusionMatrix.length; j++) {
  	sumRows[i] += m_ConfusionMatrix[i][j];
  	sumColumns[j] += m_ConfusionMatrix[i][j];
  	sumOfWeights += m_ConfusionMatrix[i][j];
        }
      }
      double correct = 0, chanceAgreement = 0;
      for (int i = 0; i < m_ConfusionMatrix.length; i++) {
        chanceAgreement += (sumRows[i] * sumColumns[i]);
        correct += m_ConfusionMatrix[i][i];
      }
      chanceAgreement /= (sumOfWeights * sumOfWeights);
      correct /= sumOfWeights;

      if (chanceAgreement < 1) {
        return (correct - chanceAgreement) / (1 - chanceAgreement);
      } else {
        return 1;
      }
    }

    /**
     * Gets the number of instances not classified (that is, for
     * which no prediction was made by the classifier). (Actually the sum
     * of the weights of these instances)
     *
     * @return the number of unclassified instances
     */
    public final double unclassified() {

      return m_Unclassified;
    }
    
    /**
     * Gets the percentage of instances not classified (that is, for
     * which no prediction was made by the classifier).
     *
     * @return the percent of unclassified instances (between 0 and 100)
     */
    public final double pctUnclassified() {

      return 100 * m_Unclassified / m_WithClass;
    }
    
    /**
     * Calls toSummaryString() with a default title.
     *
     * @param printComplexityStatistics if true, complexity statistics are
     * returned as well
     * @return the summary string
     */
    public String toSummaryString(boolean printComplexityStatistics) {

      return toSummaryString("=== Summary ===\n", printComplexityStatistics);
    }

    /**
     * Outputs the performance statistics in summary form. Lists 
     * number (and percentage) of instances classified correctly, 
     * incorrectly and unclassified. Outputs the total number of 
     * instances classified, and the number of instances (if any) 
     * that had no class value provided. 
     *
     * @param title the title for the statistics
     * @param printComplexityStatistics if true, complexity statistics are
     * returned as well
     * @return the summary as a String
     */
    public String toSummaryString(String title, 
        boolean printComplexityStatistics) { 

      StringBuffer text = new StringBuffer();

      if (printComplexityStatistics && m_NoPriors) {
        printComplexityStatistics = false;
        System.err.println("Priors disabled, cannot print complexity statistics!");
      }

      text.append(title + "\n");
      try {
        if (m_WithClass > 0) {
  	if (m_ClassIsNominal) {

  	  text.append("Correctly Classified Instances     ");
  	  text.append(Utils.doubleToString(correct(), 12, 4) + "     " +
  	      Utils.doubleToString(pctCorrect(),
  		  12, 4) + " %\n");
  	  text.append("Incorrectly Classified Instances   ");
  	  text.append(Utils.doubleToString(incorrect(), 12, 4) + "     " +
  	      Utils.doubleToString(pctIncorrect(),
  		  12, 4) + " %\n");
  	  text.append("Kappa statistic                    ");
  	  text.append(Utils.doubleToString(kappa(), 12, 4) + "\n");

  	  /*if (m_CostMatrix != null) {
  	    text.append("Total Cost                         ");
  	    text.append(Utils.doubleToString(totalCost(), 12, 4) + "\n");
  	    text.append("Average Cost                       ");
  	    text.append(Utils.doubleToString(avgCost(), 12, 4) + "\n");
  	  }*/
  	  if (printComplexityStatistics) {
  	    text.append("K&B Relative Info Score            ");
  	    text.append(Utils.doubleToString(KBRelativeInformation(), 12, 4) 
  		+ " %\n");
  	    text.append("K&B Information Score              ");
  	    text.append(Utils.doubleToString(KBInformation(), 12, 4) 
  		+ " bits");
  	    text.append(Utils.doubleToString(KBMeanInformation(), 12, 4) 
  		+ " bits/instance\n");
  	  }
  	} else {        
  	  text.append("Correlation coefficient            ");
  	  text.append(Utils.doubleToString(correlationCoefficient(), 12 , 4) +
  	  "\n");
  	}
  	if (printComplexityStatistics) {
  	  text.append("Class complexity | order 0         ");
  	  text.append(Utils.doubleToString(SFPriorEntropy(), 12, 4) 
  	      + " bits");
  	  text.append(Utils.doubleToString(SFMeanPriorEntropy(), 12, 4) 
  	      + " bits/instance\n");
  	  text.append("Class complexity | scheme          ");
  	  text.append(Utils.doubleToString(SFSchemeEntropy(), 12, 4) 
  	      + " bits");
  	  text.append(Utils.doubleToString(SFMeanSchemeEntropy(), 12, 4) 
  	      + " bits/instance\n");
  	  text.append("Complexity improvement     (Sf)    ");
  	  text.append(Utils.doubleToString(SFEntropyGain(), 12, 4) + " bits");
  	  text.append(Utils.doubleToString(SFMeanEntropyGain(), 12, 4) 
  	      + " bits/instance\n");
  	}

  	text.append("Mean absolute error                ");
  	text.append(Utils.doubleToString(meanAbsoluteError(), 12, 4) 
  	    + "\n");
  	text.append("Root mean squared error            ");
  	text.append(Utils.
  	    doubleToString(rootMeanSquaredError(), 12, 4) 
  	    + "\n");
  	if (!m_NoPriors) {
  	  text.append("Relative absolute error            ");
  	  text.append(Utils.doubleToString(relativeAbsoluteError(), 
  	      12, 4) + " %\n");
  	  text.append("Root relative squared error        ");
  	  text.append(Utils.doubleToString(rootRelativeSquaredError(), 
  	      12, 4) + " %\n");
  	}
        }
        if (Utils.gr(unclassified(), 0)) {
  	text.append("UnClassified Instances             ");
  	text.append(Utils.doubleToString(unclassified(), 12,4) +  "     " +
  	    Utils.doubleToString(pctUnclassified(),
  		12, 4) + " %\n");
        }
        text.append("Total Number of Instances          ");
        text.append(Utils.doubleToString(m_WithClass, 12, 4) + "\n");
        if (m_MissingClass > 0) {
  	text.append("Ignored Class Unknown Instances            ");
  	text.append(Utils.doubleToString(m_MissingClass, 12, 4) + "\n");
        }
      } catch (Exception ex) {
        // Should never occur since the class is known to be nominal 
        // here
        System.err.println("Arggh - Must be a bug in Evaluation class");
      }

      return text.toString(); 
    }

    
    /*
     * end of code added by schaudhu @ 08/30/07
     */
        
    /**     
     * A private method not to be accessed publicly.
     * Used for generating evaluation report. 
     *      
     * @param headline
     * @param index
     * @param trainSet
     * @param testSet
     * @param pDialog
     * @throws Exception
     */
    
    private void exportEvaluation(String headline, 
    		int index,
    		Instances trainSet, 
    		Instances testSet,
    		ProgressDialog pDialog) throws Exception  
    {	
	    	try{
	    		StringBuffer evalReport = new StringBuffer();
	    		String sp = System.getProperty("line.separator");    		
	    		//Evaluation eTest = new Evaluation(trainSet);
	    		if(MultiDimClassifier.isTenCV)
	    		{
	    			int numOfFolds = 10;    			
	    			if(trainSet.numInstances() < 10)
	    				numOfFolds = trainSet.numInstances();
	
	    			headline = headline.replaceAll("10", String.valueOf(numOfFolds));
	    			evalReport.append(headline).append(sp).append(sp);
	    			if (pDialog != null) {
  	    			pDialog.setLabel("Evaluating "+this.classifier.getClass().getSimpleName()+" ("+numOfFolds+" CV)");
  	    			pDialog.setIsIndeterminate(true);
	    			}
	    			/*
	    			 * initializing variables added from Evaluation.java
	    			 */
	    			init_newVars(trainSet);
	    			dimn = inputFileReader.getNameForDim(index);
	    			System.out.println("Dimension name being cross validated = " + dimn);
	    			/*
	    			 * following line was earlier eTest.crossValidateModel(...)
	    			 */
	    			crossValidateModel(
	    					this.classifier, trainSet, numOfFolds, trainSet.getRandomNumberGenerator(1), pDialog);	    		}
	    		else
	    		{
	    			evalReport.append(headline).append(sp).append(sp);    			
	    			if (pDialog != null) pDialog.setLabel("Evaluating the Model (specified TEST dataset)");    			
	    			// eTest.evaluateModel(this.classifier, testSet);
	    			evaluateModel(this.classifier, testSet);
	    		}
	    		evalReport.append("Dataset: ").append(this.optionsModel.getArffFileName().replaceAll(".arff", "")).append(sp);
	    		evalReport.append("Language of the dataset: ").append(this.optionsModel.getLanguage()).append(sp);
	    		evalReport.append("Dimension: ").append(inputFileReader.getNameForDim(index)).append(sp);
	    		evalReport.append("Num. of Instances: ").append(trainSet.numInstances()).append(sp);
	    		evalReport.append("Num. of Attributes: ").append(trainSet.numAttributes()).append(sp);
	    		evalReport.append("Classifier: "+this.classifier.getClass()).append(sp);
	    		evalReport.append("Classifier parameters: ").append(sp);
	    		for(int i=0; i<this.classifier.getOptions().length; i++)
	    		{
	    			evalReport.append(this.classifier.getOptions()[i]+" ");
	    		}
	    		evalReport.append(sp).append(sp);
	    		evalReport.append(this.classifier.toString().replaceAll("\n", sp)).append(sp);
	    		evalReport.append(this.optionsModel.toString()).append(sp);	        
	    		//evalReport.append(eTest.toSummaryString(true).replaceAll("\n", sp)).append(sp);
	    		evalReport.append(toSummaryString(true).replaceAll("\n", sp)).append(sp);
	    		if (!(trainSet.attribute(MultiDimClassifier.CLASS).type() == Attribute.NUMERIC)) {
	    			// only print confusion matrix is class is not numeric
	    			//evalReport.append(eTest.toMatrixString().replaceAll("\n", sp)).append(sp);
	    			evalReport.append(toMatrixString().replaceAll("\n", sp)).append(sp);
	    		}
	
	    		String path= MultiDimClassifier.OUTPUT+"/";
	    		if(!/*Driver.outputPrefix*/this.optionsModel.getOutputPrefix().equals(""))
	    		{
	    			path = path+/*Driver.outputPrefix*/this.optionsModel.getOutputPrefix()+"_";
	    		}                		
	    		path = path+"Eval_"+inputFileReader.getNameForDim(index)+"_";
	    		path = path +this.optionsModel.getArffFileName().replaceAll("arff", "txt"); //+".txt";
	
	    		FileWriter writer  = new FileWriter(path);
	    		writer.write(evalReport.toString());
	    		writer.flush();
	    		writer.close();
	    		
	    		if(this.optionsModel.isComparison() &&
	    				!(trainSet.attribute(MultiDimClassifier.CLASS).type() == Attribute.NUMERIC))
	    		{
	    			//String kappa = String.valueOf(eTest.kappa());
	    			String kappa = String.valueOf(kappa());
	    			String ratio = String.valueOf(optionsModel.getSubsetPortion());
	    			String content = ratio+", "+kappa+sp;
	    			this.optionsModel.appendStatToComparisonPath(content);
	    		}
	    	}
	    	catch(Exception ex)
	    	{
	    		throw ex;
	    	}         	

    }   
    
    private FastVector getLabels(int index)
    {
        FastVector labels = new FastVector();
        Iterator classIter = this.inputFileReader.getIteratorForDim(index);
        while (classIter.hasNext()) {
            String label = classIter.next().toString();
            if (!label.equals(InputFileReader.UNKNOWN_STRING)) {
                labels.addElement(label);
            }
        }
        return labels;
    }
    
    /**
     * This is (and should remain) the ONLY place where the structure
     * of the ARFF header is decided
      
     * @param index
     * @return
     */
    private FastVector getAttributes(int index)
    {
        FastVector atts = new FastVector();
        FastVector labels = this.getLabels(index);        
        System.out.println("set classes");        

        // before anything, add the sub-population string attribute if asked
        if (this.optionsModel.isSubPop()) {
        	Attribute att = new Attribute(InputFileReader.SUBPOP, (FastVector) null);
        	atts.addElement(att);
        }
        
        Iterator featureIter = this.featureSelector.getFeatureSetIterator(true);
        while (featureIter.hasNext()) {
            Attribute att = new Attribute(featureIter.next().toString());
            atts.addElement(att);
        }
        
        System.out.println("set regular features");        
        
        if (optionsModel.isLineLength()) {
            Attribute att = new Attribute(FeatureSelector.LINE_LENGTH_FLAG);
            atts.addElement(att);
            System.out.println("set line length feature");             
        }
                       
        for (int i = 0; i < DataPoint.nominalFeatureNames.size(); i++) {
          String nominalAtt = (String) DataPoint.nominalFeatureNames.get(i);
          FastVector nominalAttValues = (FastVector) DataPoint.nominalFeatureValueSet.get(nominalAtt);
    			if (nominalAttValues != null && nominalAttValues.size() > 0) {
    				// deduce the type of attribute to be created (nominal / numeric)
    				// automatically
    				// presence of all numeric values makes an attribute numeric
    				// else, it remains nominal (even if a single non-numeric value is encountered)
    				boolean isNumeric = true;
    				for (int j = 0; j < nominalAttValues.size(); j++) {
    					try {
    						double val =
    								Double.parseDouble(((String) nominalAttValues.elementAt(j)));
    					} catch (NumberFormatException e) {
    						isNumeric = false;
    						break;
    					}
    				}

    				Attribute att;
    				if (!isNumeric) {
    					System.out.println("Creating NOMINAL attribute: " + nominalAtt);
    					att = new Attribute(nominalAtt, nominalAttValues);
    				} else {
    					// we are creating a numeric attribute
    					System.out.println("Creating NUMERIC attribute: " + nominalAtt);
    					att = new Attribute(nominalAtt);
    				}
    				atts.addElement(att);
    			}
    		}
        
        System.out.println("set nominal features");
        
        if(optionsModel.useAdvFeatures() && this.featureSelector.getNumGlobalUserDefiendFeature() >0){
            Iterator advFeatureItr = this.featureSelector.getGlobalUserDefinedFeatureSetIterator();            
            while (advFeatureItr.hasNext()) {
                Attribute att = new Attribute(advFeatureItr.next().toString());
                atts.addElement(att);
            }        	
            System.out.println("set user defined features");        	
        }


    		// adds the class labels to the attribute set.
    		// deduce the type of class attribute to be created (nominal / numeric)
    		// automatically
				// presence of all numeric values makes an attribute numeric
				// else, it remains nominal (even if a single non-numeric value is encountered)
    		boolean isNumeric = true;
    		for (int j = 0; j < labels.size(); j++) {
    			try {
    				double val =
    						Double.parseDouble(((String) labels.elementAt(j)));
    			} catch (NumberFormatException e) {
    				isNumeric = false;
    				break;
    			}
    		}

    		Attribute att;
    		if (!isNumeric) {
    			att = new Attribute(MultiDimClassifier.CLASS, labels);
    		} else {
    			// we are creating a numeric CLASS attribute
    			att = new Attribute(MultiDimClassifier.CLASS);
    		}
    		atts.addElement(att);

        return atts;
    }
    
    private Instances[] getTrainInstances(int index, HashSet testDataPointSet, 
    									  ProgressDialog pDialog, boolean resetInstancesForFullset) //index 0- trainSet, index 1-testSet
    {
    	FastVector atts = this.getAttributes(index);
    	
        // training set
        Instances trainSet = new Instances(MultiDimClassifier.TRAIN_SET, atts, 0);
        trainSet.setClassIndex(trainSet.numAttributes() - 1);

        // test set (consisting of the instances whose class is not known)
        Instances testSet = new Instances(MultiDimClassifier.TEST_SET, atts, 0);
        testSet.setClassIndex(trainSet.numAttributes() - 1);
        
        Iterator dataPointIter = dataPoints.iterator();
        int totalSize = dataPoints.size();

        int dataPointNum = 1;
        // splits data points into training/test set.
        while (dataPointIter.hasNext()) {
        	if(pDialog !=null) {
        		pDialog.setValue(dataPointNum);
        		pDialog.updateLabelWithSuffix("[instance " + dataPointNum + "/" + totalSize + "]");
        	}
            DataPoint dataPoint = (DataPoint) dataPointIter.next();
            dataPointNum++;
            String dimValue = dataPoint.getValueAtDim(index);
            boolean isReference = dataPoint.getIsReferenceAtDim(index);
            if(!resetInstancesForFullset)
            {
	            if(isReference){			/*!dimValue.equals(InputFileReader.UNKNOWN_STRING)) {*/
	                Instance instance = this.featureSelector.getInstanceFromDataPoint(dataPoint, trainSet);
	                if (trainSet.attribute(MultiDimClassifier.CLASS).type() == Attribute.NUMERIC) {
	                	instance.setValue(trainSet.attribute(MultiDimClassifier.CLASS), Double.parseDouble(dimValue));
	                }
	                else {
	                	instance.setValue(trainSet.attribute(MultiDimClassifier.CLASS), dimValue);
	                }
	                trainSet.add(instance);
	            } 
	            else 
	            {
	                Instance instance = this.featureSelector.getInstanceFromDataPoint(dataPoint, testSet);
	                instance.setValue(testSet.attribute(MultiDimClassifier.CLASS), Instance.missingValue());
	                testSet.add(instance);
	                        
	                if(testDataPointSet != null)
	                	testDataPointSet.add(dataPoint); 
	            }
            }
            else
            {
	            if(!dimValue.equals(InputFileReader.UNKNOWN_STRING)){
	                Instance instance = this.featureSelector.getInstanceFromDataPoint(dataPoint, trainSet);
	                if (trainSet.attribute(MultiDimClassifier.CLASS).type() == Attribute.NUMERIC)
	                	instance.setValue(trainSet.attribute(MultiDimClassifier.CLASS), Double.parseDouble(dimValue));
	                else
	                	instance.setValue(trainSet.attribute(MultiDimClassifier.CLASS), dimValue);
	                trainSet.add(instance);
	            }             	
            }
        }
        
        Instances[] result = new Instances[2];
        result[0] = trainSet;
        result[1] = testSet;
        return result;
    }
    
    /**
     * Index of the data that is going to be written into an .arff file. The
     * index correspoinds to the number of the 'class' column in the input (xls)
     * file.
     * 
     * @param index
     * @param fileName
     * @param prefix
     * @param resetInstancesForFullSet
     * 
     */
    protected String writeARFFForDim(int index, 
    							String fileName, 
    							String prefix, 
    							boolean resetInstancesForFullset,
    							ProgressDialog pDialog) //resetInstancesForFullset == true, only when unlabeled data have been labeled and need
    {

    	if(resetInstancesForFullset)
    	{
            Instances[] datasets = this.getTrainInstances(index, null, null, true);
            Instances fullDataset = datasets[0];
            
	        String outputArff = MultiDimClassifier.ARFF + "/";
	        
	        if(!"".equals(prefix))
	        	outputArff = outputArff + prefix+"_";
	        
	        outputArff = outputArff+ this.inputFileReader.getNameForDim(index) + "_" + fileName;
	        if (pDialog != null) {
	        	pDialog.setIsIndeterminate(true);
	        	pDialog.setLabel("Writing ARFF file " + outputArff + " to disk...");
	        }
	        Logger logger = new Logger(outputArff);
	        logger.writeLine(fullDataset.toString());
	        logger.close();
	        System.out.println("wrote " + fullDataset.numInstances() + " to arff file: " + outputArff);
	        return outputArff;
    	}
    	else
    	{
	        String outputArff = MultiDimClassifier.ARFF + "/";
	        
	        if(!"".equals(prefix))
	        	outputArff = outputArff + prefix+"_";
	        
	        outputArff = outputArff+ this.inputFileReader.getNameForDim(index) + "_" + fileName;
	        if (pDialog != null) {
	        	pDialog.setIsIndeterminate(true);
	        	pDialog.setLabel("Writing ARFF file " + outputArff + " to disk...");
	        }
	        Logger logger = new Logger(outputArff);
	        logger.writeLine(this.trainSet.toString());
	        logger.close();
	        System.out.println("wrote " + this.trainSet.numInstances() + " to arff file: " + outputArff);
	        
	        if (testSet.numInstances() > 0) {
	            String outputTestArff = MultiDimClassifier.ARFF + "/";
	        
	            if(!"".equals(prefix))
	            	outputTestArff = outputTestArff + prefix+"_";
	            
	            outputTestArff = outputTestArff + 
	            				  this.inputFileReader.getNameForDim(index) +
	            				  "_"+fileName.replaceFirst("train","test");//"_test_" + prefix;
	  	        if (pDialog != null) {
	  	        	pDialog.setIsIndeterminate(true);
	  	        	pDialog.setLabel("Writing ARFF file " + outputTestArff + " to disk...");
	  	        }
	            Logger testLogger = new Logger(outputTestArff);
	            testLogger.writeLine(testSet.toString());
	            testLogger.close();
	            System.out.println("wrote " + this.testSet.numInstances() + " to arff file: " + outputTestArff);            
	        }
	        return outputArff;
    	}
    }
    
    /**
     * Writing the feature spaces into the MinorThird / SVM Light / GibbsLDA++ format 
     * 
     * @param index
     * @param fileName
     * @param prefix
     */

    protected void writeDatasetFileForDim(int index, String fileName, String prefix, 
    		ProgressDialog pDialog, int format, boolean includeSubpopInfo) {
        
        Iterator dpIter = dataPoints.iterator();
        String trainDataset = "";
        String testDataset = "";
        int trainInstId = 0;
        int testInstId = 0;
        String trainNewLine = "";
        String testNewLine = "";
        String fileExt = "";
        if (format == MultiDimClassifier.FORMAT_MINORTHIRD)
        	fileExt = MultiDimClassifier.FORMAT_MINORTHIRD_EXT;
        else if (format == MultiDimClassifier.FORMAT_GIBBSLDA)
        	fileExt = MultiDimClassifier.FORMAT_GIBBSLDA_EXT;
        else if (format == MultiDimClassifier.FORMAT_SVMLIGHT)
        	fileExt = MultiDimClassifier.FORMAT_SVMLIGHT_EXT;

        if (pDialog != null) {
        	pDialog.setMax(dataPoints.size());
        	pDialog.setValue(0);
        }
        FastVector atts = getAttributes(index);
        Instances dummy = new Instances("dummy", atts, 0);
        THashMap subPopHash = new THashMap();
        int progress = 0;
        int totalSize = dataPoints.size();
        while (dpIter.hasNext()) {
            // update progress bar
            progress++;
            if (pDialog != null) {
            	pDialog.setValue(progress);
            	pDialog.updateLabelWithSuffix("[instance " + progress + "/" + totalSize + "]");
            }
            DataPoint dataPoint = (DataPoint) dpIter.next();
            String currInst = "";
            String dimValue = dataPoint.getValueAtDim(index);
            // System.out.println(dimValue);
            boolean isReference = dataPoint.getIsReferenceAtDim(index);            
            if (isReference){//!dimValue.equals(InputFileReader.UNKNOWN_STRING)) {
                trainInstId++;
                currInst += trainNewLine;
                String subPop = dataPoint.getSubPopId().replaceAll("\\s+", "_");
                if (subPop.equals(""))
                	subPop = "" + trainInstId;
                if (format == MultiDimClassifier.FORMAT_MINORTHIRD)
                	currInst += this.featureSelector.getMTInstanceFromDataPoint(dataPoint, dummy, dimValue,
                		subPop, MultiDimClassifier.MTPREFIX);
                else if (format == MultiDimClassifier.FORMAT_GIBBSLDA)
                	currInst += this.featureSelector.getGibbsLDAPlusPlusInstanceFromDataPoint(dataPoint, dummy);
                else if (format == MultiDimClassifier.FORMAT_SVMLIGHT) {
                	if (Driver.isSkipClass())
                		currInst += this.featureSelector.getSVMLightInstanceFromDataPoint(dataPoint, dummy, 
                			includeSubpopInfo, subPopHash, "");
                	else
                		currInst += this.featureSelector.getSVMLightInstanceFromDataPoint(dataPoint, dummy, 
                    			includeSubpopInfo, subPopHash, dimValue);
                }
                trainNewLine = "\n";
                trainDataset += currInst;
            } else {
                testInstId++;
                currInst += testNewLine;
                String subPop = dataPoint.getSubPopId().replaceAll("\\s+", "_");
                if (subPop.equals(""))
                	subPop = "" + testInstId;
                if (format == MultiDimClassifier.FORMAT_MINORTHIRD)
                	currInst += this.featureSelector.getMTInstanceFromDataPoint(dataPoint, dummy, 
                			String.valueOf(InputFileReader.UNKNOWN), subPop, MultiDimClassifier.MTPREFIX);
                else if (format == MultiDimClassifier.FORMAT_GIBBSLDA)
                	currInst += this.featureSelector.getGibbsLDAPlusPlusInstanceFromDataPoint(dataPoint, dummy);
                else if (format == MultiDimClassifier.FORMAT_SVMLIGHT)
                	currInst += this.featureSelector.getSVMLightInstanceFromDataPoint(dataPoint, dummy, 
                			includeSubpopInfo, subPopHash, String.valueOf(InputFileReader.UNKNOWN));
                testNewLine = "\n";
                testDataset += currInst;
            }
        }

        if (pDialog != null)
        	pDialog.setIsIndeterminate(true);
        
        if (!trainDataset.equals("")) {
            String outputMT = MultiDimClassifier.ARFF + "/";
            
            if(!"".equals(prefix))
            	outputMT = outputMT + prefix + "_";
            
            outputMT = outputMT + this.inputFileReader.getNameForDim(index) + "_" + 
            	fileName.substring(0, fileName.lastIndexOf('.')) + fileExt;
            if (pDialog != null)
            	pDialog.setLabel("Writing file " + outputMT + " to disk...");
            Logger logger = new Logger(outputMT);
            if (format == MultiDimClassifier.FORMAT_GIBBSLDA)
            	logger.writeLine("" + trainInstId);
            logger.writeLine(trainDataset);
            logger.close();
            System.out.println("wrote " + trainInstId + " to file: " + outputMT);
            
            // write out the full list of simple features (covers only unigrams,
            // bigrams, pos-bigrams, punctuation)
            if (Driver.isWriteFeatures()) {
	            String featuresFile = MultiDimClassifier.ARFF + "/" + (prefix.equals("") ? "" : prefix + "_") + this.inputFileReader.getNameForDim(index) + "_" + 
	        		fileName.substring(0, fileName.lastIndexOf('.')) + MultiDimClassifier.FEATURES_EXT;
	            Logger feat = new Logger(featuresFile);
	            // skip the last attribute (class)
	            for (int i = 0; i < atts.size() - 1; i++) {
	            	feat.writeLine(((Attribute) atts.elementAt(i)).name());
	            }
	            feat.close();
            }
        }
        
        if (!testDataset.equals("")) {
            String outputMT = MultiDimClassifier.ARFF + "/";
            if(!"".equals(prefix))
            	outputMT = outputMT + prefix + "_";
            outputMT = outputMT + this.inputFileReader.getNameForDim(index) + "_test_" + 
            	fileName.substring(0, fileName.lastIndexOf('.')) + fileExt;
            if (pDialog != null)
            	pDialog.setLabel("Writing file " + outputMT + " to disk...");
            Logger logger = new Logger(outputMT);
            if (format == MultiDimClassifier.FORMAT_GIBBSLDA)
            	logger.writeLine("" + testInstId);
            logger.writeLine(testDataset);
            logger.close();
            System.out.println("wrote " + testInstId + " to file: " + outputMT);
        }

        if (pDialog != null)
        	pDialog.setIsIndeterminate(false);
        
    }

    /////////////////////////////////////////////////////////////
    ///
    /// Following methods are only for lab uses
    /// check the setting on Driver.SUPPLY_TESTING_DATASET
    ///
    ////////////////////////////////////////////////////////////
    
    protected void writeTestARFFForDim(int index, String fileName, String prefix, ProgressDialog pDialog) {
      // setting up test set structure, before writing it to the .arff file
      FastVector atts = getAttributes(index);
      // passing
      Instances testSet = new Instances(MultiDimClassifier.TEST_SET, atts, 0);
      testSet.setClassIndex(testSet.numAttributes() - 1);

      Iterator testDataPointIter = testDataPoints.iterator();

      int dataPointNum = 1;
      // splits data points into training/test set.
      while (testDataPointIter.hasNext()) {
          DataPoint dataPoint = (DataPoint) testDataPointIter.next();
          dataPointNum++;
          String dimValue = dataPoint.getValueAtDim(index);
          // System.out.println(dimValue);
          if (!dimValue.equals(InputFileReader.UNKNOWN_STRING)) {
              Instance instance = this.featureSelector.getInstanceFromDataPoint(dataPoint, testSet);
              if (testSet.attribute(MultiDimClassifier.CLASS).type() == Attribute.NUMERIC) {
              	instance.setValue(testSet.attribute(MultiDimClassifier.CLASS), Double.parseDouble(dimValue));
              } else {
              	instance.setValue(testSet.attribute(MultiDimClassifier.CLASS), dimValue);
              }
              testSet.add(instance);
          }
      }
      // Logger logger = new Logger(MultiDimClassifier.ARFF + "/" + prefix + "_" + this.inputFileReader.getNameForDim(index) + ".arff");
      String outputArff = MultiDimClassifier.ARFF + "/";
      if(!"".equals(prefix))
      	outputArff = outputArff + prefix+"_";
      outputArff = outputArff + this.inputFileReader.getNameForDim(index) + "_" + fileName;
      if (pDialog != null) {
      	pDialog.setIsIndeterminate(true);
      	pDialog.setLabel("Writing ARFF file " + outputArff + " to disk...");
      }
      Logger logger = new Logger(outputArff);
      logger.writeLine(testSet.toString());
      logger.close();
      System.out.println("wrote " + testSet.numInstances() + " to arff file: " + outputArff);
  }

    protected void writeTestDatasetFileForDim(int index, String fileName, String prefix, 
    		ProgressDialog pDialog, int format, boolean includeSubpopInfo) {
        
        Iterator dpIter = testDataPoints.iterator();
        String dataset = "";
        
        String fileExt = "";
        if (format == MultiDimClassifier.FORMAT_MINORTHIRD)
        	fileExt = MultiDimClassifier.FORMAT_MINORTHIRD_EXT;
        else if (format == MultiDimClassifier.FORMAT_GIBBSLDA)
        	fileExt = MultiDimClassifier.FORMAT_GIBBSLDA_EXT;
        else if (format == MultiDimClassifier.FORMAT_SVMLIGHT)
        	fileExt = MultiDimClassifier.FORMAT_SVMLIGHT_EXT;

        int instId = 1;
        if (pDialog != null) {
        	pDialog.setMax(dataPoints.size());
        	pDialog.setValue(0);
        }
        int progress = 0;
        int totalSize = testDataPoints.size();
        FastVector atts = getAttributes(index);
        Instances dummy = new Instances("dummy", atts, 0);
        THashMap subPopHash = new THashMap();
        while (dpIter.hasNext()) {
            // update progress bar
            progress++;
            if (pDialog != null) {
            	pDialog.setValue(progress);
            	pDialog.updateLabelWithSuffix("[instance " + progress + "/" + totalSize + "]");
            }
            DataPoint dataPoint = (DataPoint) dpIter.next();
            String currInst = "";
            String subPop = dataPoint.getSubPopId().replaceAll("\\s+", "_");
            if (subPop.equals(""))
            	subPop = "" + instId;
            String dimValue = dataPoint.getValueAtDim(index);
            // System.out.println(dimValue);
            if (!dimValue.equals(InputFileReader.UNKNOWN_STRING)) {
              if (format == MultiDimClassifier.FORMAT_MINORTHIRD)
              	currInst += this.featureSelector.getMTInstanceFromDataPoint(dataPoint, dummy, dimValue,
              		subPop, MultiDimClassifier.MTPREFIX);
              else if (format == MultiDimClassifier.FORMAT_GIBBSLDA)
              	currInst += this.featureSelector.getGibbsLDAPlusPlusInstanceFromDataPoint(dataPoint, dummy);
              else if (format == MultiDimClassifier.FORMAT_SVMLIGHT)
              	currInst += this.featureSelector.getSVMLightInstanceFromDataPoint(dataPoint, dummy, 
              			includeSubpopInfo, subPopHash, dimValue);
            }
            dataset += currInst;
            if (dpIter.hasNext()) {
                dataset += "\n";
                instId++;
            }
        }
        if (pDialog != null)
        	pDialog.setIsIndeterminate(true);
        
        // Logger logger = new Logger(MultiDimClassifier.ARFF + "/" + prefix + "_" + this.inputFileReader.getNameForDim(index) + ".arff");
        String outputMT = MultiDimClassifier.ARFF + "/";
        if(!"".equals(prefix))
        	outputMT = outputMT + prefix + "_";
        outputMT = outputMT + this.inputFileReader.getNameForDim(index) + "_" + 
        	fileName.substring(0, fileName.lastIndexOf('.')) + fileExt;
        if (pDialog != null)
        	pDialog.setLabel("Writing file " + outputMT + " to disk...");
        Logger logger = new Logger(outputMT);
        if (format == MultiDimClassifier.FORMAT_GIBBSLDA)
        	logger.writeLine("" + instId);
        logger.writeLine(dataset);
        logger.close();
        System.out.println("wrote " + instId + " to MT file: " + outputMT);
        
        if (pDialog != null)
        	pDialog.setIsIndeterminate(false);
    }
}

