/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package edu.cmu.taghelpertools.middle_layer;

import edu.cmu.taghelpertools.gui.Classifiers;
import edu.cmu.taghelpertools.gui.OptionsFrame;
import edu.cmu.taghelpertools.tokenizer.Tokenizer;
import edu.cmu.taghelpertools.middle_layer.extractor.FeatureSpaceMapBuilder;
import java.io.*;

import gnu.trove.THashMap;
import java.util.*;

import javax.swing.JOptionPane;

import weka.core.Instances;

/**
 * Maintains the user-specified feature selection and re-training settings. Each
 * file has its own OptionsModel. Feature selection options include: unigrams,
 * bigrams, POS bigrams, punctuation, sentence length, does the sentence contain
 * a non_stopword (a binary-valued indicator).
 * 
 * Language, evaluation and sampling settings are recorded here. This class
 * can be serialized as a file for later reuse.
 * 
 * Re-training settings determine
 * how predictions on the test set should be added to the test set during the
 * self-training process. The user can chose to (1) add a fixed number of
 * predictions, (2) a percentage of all predictions made, and (3) all
 * predictions with confidence above a user-defined threshold.
 * 
 * @author Jaime Arguello, Hao-Chuan Wang
 * 
 */
public class OptionsModel implements java.io.Serializable{
	
		public final static int DEFAULT_FREQ_THRESH = 5;
    public final static int DEFAULT_NUM_EXAMPLES = 5;
    public final static double DEFAULT_PERCENT = 0.10;
    public final static double DEFAULT_THRESH = 0.90;
    public static int DEFAULT_LANG = Tokenizer.LANG_ENGLISH;
    public final static String lang_setting_file = "default_lang_setting.txt";
    public static boolean enableChineseOption;    
    static
    {
        /*check Chinese option*/
        File f = new File("CKIPWS.dll");
        if(f.exists())
        	enableChineseOption = true;
        else
        	enableChineseOption = false;
        
    	try
    	{
    		BufferedReader rd = new BufferedReader(new FileReader(lang_setting_file));
    		String currentLine ="";
    		for(int i=0;(currentLine=rd.readLine())!=null; i++)
    		{
    			if(i==0)
    			{
    				if(currentLine.toLowerCase().trim().equals(Tokenizer.LANGS[Tokenizer.LANG_ENGLISH].toLowerCase()))
    				{
    					DEFAULT_LANG = Tokenizer.LANG_ENGLISH;
    				}
    				else if(currentLine.toLowerCase().trim().equals(Tokenizer.LANGS[Tokenizer.LANG_GERMAN].toLowerCase()))
    				{
    					DEFAULT_LANG = Tokenizer.LANG_GERMAN;
    				}
    				else if(currentLine.toLowerCase().trim().equals(Tokenizer.LANGS[Tokenizer.LANG_CHINESE].toLowerCase())
    						&& enableChineseOption)
    				{
    					DEFAULT_LANG = Tokenizer.LANG_CHINESE;
    				}
    				else
    				{
    					DEFAULT_LANG = Tokenizer.LANG_ENGLISH;    					
    				}
    			}
    		}
    	}
    	catch(Exception ex)
    	{
    		System.err.println("\nError on loading language default settings from the file: "+lang_setting_file+".");
    	}
    }
    
    // originally set to default values;    
    // values for exchange prediction with text and xml
    // values for action prediction without exchange, but with text and xml
    private boolean unigrams = true;
    private boolean bigrams = true;
    private boolean posBigrams = true;
    private boolean lineLength = true;
    private boolean containsNonStopword = true;
    private boolean puctuation = true;
    private boolean stem = true;
    private boolean removeStopwords = true;
    private boolean removeRareFeatures = true;
    private int frequencyThreshold = OptionsModel.DEFAULT_FREQ_THRESH;
    private boolean selfTrain = false;
    private weka.classifiers.Classifier classifier = new weka.classifiers.bayes.NaiveBayes();
    private String[] options4classifier = null;
    //followings are for self-training
    private int iterationMode = MultiDimClassifier.ADD_FIXED_NUM;
    private int fixedNum = OptionsModel.DEFAULT_NUM_EXAMPLES;
    private double thresh = OptionsModel.DEFAULT_THRESH;
    private double percent = OptionsModel.DEFAULT_PERCENT;
    
    private boolean useAdvFeature = false;
    
    private String arffFileName;
    
    //by Hao-Chuan Wang
    private boolean isUsingArff = true; // if false, then using MinorThird format
    private boolean outputFreqCounts = false; // if true, then frequency counts of features are
    																					// are output instead of the default 0/1 values
    private boolean subPop = false; // if true, then a "string" attribute is added to
    																// Weka ARFF file, as the sub-population identifier
    private String selectedLanguage = Tokenizer.LANGS[OptionsModel.DEFAULT_LANG];
    private String stopwordListPath = Tokenizer.STPS[OptionsModel.DEFAULT_LANG];
    
    //by Hao-Chuan wang
    private boolean isEvaluating = true;
    private String preprefix = "";
    
    //by Hao-Chuan Wang
    private THashMap featureMap;
    private TreeSet featureSet = null;
    private ArrayList nominalFeatureNames = null;
    private THashMap nominalFeatureValueSet = null;
    
    //by Hao-Chuan Wang, for advanced feature extraction
    private	ArrayList<DataPoint> preprocessedTrainingDataPoints = null;
    private InputFileReader ifr = null;
    private FeatureSpaceMapBuilder fsmBuilder = null;
    private OptionsFrame optionsFrame = null;
    
    //by Hao-Chuan Wang, for random splitting 
    private boolean subsetInput = false;
    private double subsetProportion = 0.5;
    
    //by Hao-Chuan Wang, for experimental comparison
    private File comparisonPath = null;
    
    public void setSubsetInput(boolean subsetInput)
    {
    	this.subsetInput = subsetInput;
    }
    
    public boolean isSubsetInput()
    {
    	return this.subsetInput;
    }
    
    public void setSubsetPortion(double portion)
    {
    	this.subsetProportion = portion;
    }
    
    public double getSubsetPortion()
    {
    	return this.subsetProportion;
    }
    
    public void setOptionsFrame(OptionsFrame of)
    {
    	this.optionsFrame = of;
    }
    
    public OptionsFrame getOptionsFrame()
    {
    	return this.optionsFrame;
    }
    
    public void setPreprocessed(ArrayList<DataPoint> dataPoints, InputFileReader ifr, FeatureSpaceMapBuilder fsmBuilder)
    {
    	this.preprocessedTrainingDataPoints = dataPoints;
    	this.ifr = ifr;
    	this.fsmBuilder = fsmBuilder;
    }
    
    public ArrayList<DataPoint> getPreprocessedTrainingDataPoints()    
    {
    	return this.preprocessedTrainingDataPoints;
    }
    
    public InputFileReader getInputFileReader()
    {
    	return this.ifr;
    }
    
    public FeatureSelector getFeatureSelector()
    {
    	if(this.fsmBuilder == null)
    		return null;
    	else
    		return this.fsmBuilder.getFeatureSelector();
    }
    
    public FeatureSpaceMapBuilder getFSMBuilder()
    {
    	return this.fsmBuilder;
    }
    
    public void clearPreprocessed()
    {
    	this.preprocessedTrainingDataPoints = null;
    	this.ifr = null;
    	this.fsmBuilder = null;
    	this.optionsFrame = null;
    }
    
    public boolean isComparison()
    {
    	return this.comparisonPath != null;
    }
    
    public void setComparisonPath(String path)
    {
    	if(this.comparisonPath == null)
    	{
	    	this.comparisonPath = new File(path);
	    	try
	    	{
	    		FileWriter writer  = new FileWriter(this.comparisonPath);
	    		String header = "Sampling Ratio, Kappa";
	    		writer.write(/*this.toString()+System.getProperty("line.separator")+*/
	    					header+System.getProperty("line.separator"));
//                JOptionPane.showMessageDialog(null, this.toString(), "Success", JOptionPane.INFORMATION_MESSAGE);	    		
	    		writer.close();
	    	}
	    	catch(Exception ex)
	    	{
	    		ex.printStackTrace();
	    		System.exit(0);
	    	}
    	}
    }
    
    public void setComparisonPath(File comparisonPath)
    {
    	this.comparisonPath = comparisonPath;
    }
    
    public File getComparisonPath()
    {
    	return this.comparisonPath;
    }
    
    public void appendStatToComparisonPath(String content)
    {
    	try
    	{
    		FileWriter writer  = new FileWriter(this.comparisonPath, true);
    		writer.write(content);    	
    		writer.close();
    	}
    	catch(Exception ex)
    	{
    		ex.printStackTrace();
    		System.exit(0);
    	}    	
    }
    
    
    public void setNominalFeatureValueSet(THashMap thm)
    {
    	this.nominalFeatureValueSet = thm;
    }
    
    public THashMap getNominalFeatureValueSet()
    {
    	return this.nominalFeatureValueSet;
    }
    
    
    public void setNominalFeatureNames(ArrayList al)
    {
    	this.nominalFeatureNames = al;
    }
    
    public ArrayList getNominalFeatureNames()
    {
    	return this.nominalFeatureNames;
    }

    public void setFeatureMap(THashMap featureMap)
    {
    	this.featureMap = featureMap;
    }

    public THashMap getFeatureMap()
    {
    	return this.featureMap;
    }

    public void setFeatureSet(TreeSet featureSet)
    {
    	this.featureSet= featureSet;
    }
        
    public TreeSet getFeatureSet()
    {
    	if(this.featureSet != null)
    		return this.featureSet;
    	else
    	{
	    	TreeSet ts = new TreeSet();
	    	Iterator itr = this.featureMap.keySet().iterator();
	    	while(itr.hasNext())
	    	{
	    		String tmpKey = itr.next().toString();
	    		ts.add(tmpKey);
	    		
	    	}
	    	
	    	return ts;
    	}
    }
    
    public void setOutputPrefix(String preprefix)
    {
    	this.preprefix = preprefix;
    }
    
    public String getOutputPrefix()
    {
    	return this.preprefix;
    }
    
    public String toString()
    {
    	return this.toString(this.getFeatureSelector());
    }
    
    public String toString(FeatureSelector featureSelector)
    {
    	String sp = System.getProperty("line.separator");    	
    	StringBuffer sb = new StringBuffer();
    	sb.append("=== Feature Selection Settings ===").append(sp).append(sp);
    	sb.append("(Tokenizer's language version: ").append(this.selectedLanguage).append(")").append(sp);
    	if(this.subsetInput)
    	{
    		sb.append("(Randomly sample input data= true").append(", Proportion=").append(this.subsetProportion).append(")").append(sp);
    	}
    	sb.append("Punctuation= ").append(this.puctuation).append(sp);
    	sb.append("Unigrams= ").append(this.unigrams).append(sp);
    	sb.append("Bigrams= ").append(this.bigrams).append(sp);
    	sb.append("POS Bigrams= ").append(this.posBigrams).append(sp);
    	sb.append("Line length= ").append(this.lineLength).append(sp);
    	sb.append("Remove stopwords= ").append(this.removeStopwords).append(sp);
    	sb.append("Stem= ").append(this.stem).append(sp);
    	sb.append("Contains non-stopwords= ").append(this.containsNonStopword).append(sp);
    	sb.append("Remove rare features= ").append(this.removeRareFeatures).append(sp);
    	if(this.removeRareFeatures)
    		sb.append("   -Freq. threshold for feature removal= ").append(this.frequencyThreshold).append(sp);
    	sb.append("User Defined= ").append(this.useAdvFeature).append(sp);
    	if(this.useAdvFeature)
    	{
    		/*if(this.fsmBuilder != null)*/
    		if(featureSelector != null)
    		{
    			Iterator itr = featureSelector.getGlobalUserDefinedFeatureSetIterator();
    			while(itr.hasNext())
    			{
    				String id = itr.next().toString();
    				sb.append("   "+id+": "+featureSelector.getlobalUserDefinedFeatureExpression(id)).append(sp);
    			}
    		}
    	}
    	
    	return sb.toString();
    }
    
    /**
     * @return the arffFileName
     */
    public String getArffFileName() {
        return arffFileName;
    }

    /**
     * @param arffFileName the arffFileName to set
     */
    public void setArffFileName(String arffFileName) {
        this.arffFileName = arffFileName;
    }

    public boolean isBigrams() {
        return bigrams;
    }

    public void setBigrams(boolean bigrams) {
        this.bigrams = bigrams;
    }

    public boolean isContainsNonStopword() {
        return containsNonStopword;
    }

    public void setContainsNonStopword(boolean containsNonStopword) {
        this.containsNonStopword = containsNonStopword;
    }

    public int getFixedNum() {
        return fixedNum;
    }

    public void setFixedNum(int fixedNum) {
        this.fixedNum = fixedNum;
    }

    public int getIterationMode() {
        return iterationMode;
    }

    public void setIterationMode(int iterationMode) {
        this.iterationMode = iterationMode;
    }

    public boolean isLineLength() {
        return lineLength;
    }

    public void setLineLength(boolean lineLength) {
        this.lineLength = lineLength;
    }

    public double getPercent() {
        return percent;
    }

    public void setPercent(double percent) {
        this.percent = percent;
    }

    public boolean isPosBigrams() {
        return posBigrams;
    }

    public void setPosBigrams(boolean posBigrams) {
        this.posBigrams = posBigrams;
    }

    public double getThresh() {
        return thresh;
    }

    public void setThresh(double thresh) {
        this.thresh = thresh;
    }

    public boolean isUnigrams() {
        return unigrams;
    }

    public void setUnigrams(boolean unigrams) {
        this.unigrams = unigrams;
    }

    public boolean isRemoveStopwords() {
        return removeStopwords;
    }

    public void setRemoveStopwords(boolean removeStopwords) {
        this.removeStopwords = removeStopwords;
    }

    public boolean useAdvFeatures()
    {
    	return this.useAdvFeature;
    }
    
    public void setUseAdvFeatures(boolean useAdv)
    {
    	this.useAdvFeature = useAdv;
    }
    
    public void setFrequencyThreshold(int frequencyThreshold) {
        this.frequencyThreshold = frequencyThreshold;
    }
    
    public int getFrequencyThreshold() {
        return frequencyThreshold;
    }
    
    public boolean isStem() {
        return stem;
    }

    public void setStem(boolean stem) {
        this.stem = stem;
    }
    
    public weka.classifiers.Classifier getClassifier()
    {
    	return this.classifier;
    }
    
    public void setClassifier(weka.classifiers.Classifier classifier)
    {
    	this.classifier = classifier;
    }
    
    public String[] getClassifierOptions()
    {
    	return this.options4classifier;
    }
    
    public void setClassifierOptions(String[] options)
    {
    	this.options4classifier = options;
    }

    public boolean isPuctuation() {
        return puctuation;
    }

    public void setPunctuation(boolean puctuation) {
        this.puctuation = puctuation;
    }

    public boolean isRemoveRareFeatures() {
        return this.removeRareFeatures;
    }

    public void setRemoveRareFeatures(boolean removeRareFeatures) {
        this.removeRareFeatures = removeRareFeatures;
    }

    public boolean isSelfTrain() {
        return this.selfTrain;
    }

    public void setSelfTraing(boolean selfTrain) {
        this.selfTrain = selfTrain;
    }
    
    //added by Hao-Chuan Wang, 200701xx
    public void setUsingArff(boolean isUsingArff)
    {
    	this.isUsingArff = isUsingArff;
    }
    
    public boolean isUsingArff()
    {
    	return this.isUsingArff;
    }
    
    public void setOutputFreqCounts(boolean outputFreqCounts) {
    	this.outputFreqCounts = outputFreqCounts;
    }
    
    public boolean isOutputFreqCounts() {
    	return this.outputFreqCounts;
    }
    
    public void setSubPop(boolean subPop) {
    	this.subPop = subPop;
    }
    
    public boolean isSubPop() {
    	return this.subPop;
    }
    
    //added by Hao-Chuan Wang, 20070117
    public void setLanguage(String lang, String stp)
    {
    	if(lang!=null)
    	{
    		this.selectedLanguage = lang;
    		this.stopwordListPath = stp;
    	}
    }
    
    public String getLanguage()
    {
    	return this.selectedLanguage;
    }
    
    public String getStopwordListPath()
    {
      return this.stopwordListPath;	
    }
    
    //by Hao-Chuan Wang
    public boolean isEvaluating()
    {
    	return this.isEvaluating;
    }
    
    public void setEvaluating(boolean eval)
    {
    	this.isEvaluating = eval;
    }
    
    
    public static void main(String[] args)
    {
    	File f = new File("runtht.bat");//CKISWS.dll");
    	System.out.print(f.exists());
    }
   
    
}