/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package edu.cmu.taghelpertools.middle_layer;

import java.io.*;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.swing.JOptionPane;

import org.upsam.grade.util.DLLChineseSegmenter;

import com.Ostermiller.util.CSVParser;

import jxl.Cell;
import jxl.Sheet;
import jxl.Workbook;
import jxl.WorkbookSettings;
import jxl.read.biff.BiffException;

import edu.cmu.taghelpertools.gui.Classifiers;
import edu.cmu.taghelpertools.gui.FileSelector;
import edu.cmu.taghelpertools.gui.ProgressDialog;
import edu.cmu.taghelpertools.gui.TestModelFrame;
import edu.cmu.taghelpertools.io.Logger;
import edu.cmu.taghelpertools.io.Reader;
import edu.cmu.taghelpertools.tokenizer.ChineseTokenizer;
import edu.cmu.taghelpertools.tokenizer.Tokenizer;
import edu.cmu.taghelpertools.util.GenericNode;
import edu.cmu.taghelpertools.util.GenericNodeSorter;
import edu.cmu.taghelpertools.util.HelperFunctions;
import gnu.trove.THashSet;

import weka.classifiers.Classifier;
import weka.classifiers.Evaluation;
import weka.core.Attribute;
import weka.core.FastVector;
import weka.core.Instance;
import weka.core.Instances;
import weka.classifiers.meta.*;
import gnu.trove.THashMap;

/**
 * This is the class for loading a previously trained model and test it by
 * supplying a new dataset. An evaluation report will be generated for displaying the 
 * performance.
 * 
 * When loading a trained model, associated resource files for feature space reconstruction
 * will also be loaded.  These resources all locate in subfolders in the CLASSIFIER folder.
 * 
 * @author haochuan
 *
 */
public class SavedModelEvaluator {

	private final static boolean DEAL_NOMINAL = true;

	private final static boolean DEBUG = false;

	private ArrayList extDataPoints = new ArrayList();

	private FeatureSelector featureSelector;

	private TreeMap<String, FeatureGlobalProperty> udf;
	
	private THashSet featureSet;

	private OptionsModel optionsModel;

	private Classifier classifier;

	private String dimName;

	private int dimIndex = -1;

	private FastVector dimLabels = new FastVector();

	private Instances testSet = null;

	private File testData;

	private String inputPath;

	private TestModelFrame gui;
	
	public SavedModelEvaluator(File savedModel, String dimName, TestModelFrame gui) {
		this.gui = gui;
		this.dimName = dimName;
		try {
			//initiate classifier			
			ObjectInputStream ois = new ObjectInputStream(new FileInputStream(
					savedModel));
			this.classifier = (Classifier) ois.readObject();
			ois.close();

			//initiate options model			
			ObjectInputStream ois2 = new ObjectInputStream(new FileInputStream(
					savedModel.getAbsolutePath().replaceAll("\\.model$", ".om")));
			this.optionsModel = (OptionsModel) ois2.readObject();
			ois2.close();

			//initiate instances
			ObjectInputStream ois3 = new ObjectInputStream(new FileInputStream(
					savedModel.getAbsolutePath().replaceAll("\\.model$", ".ins")));
			this.testSet = (Instances) ois3.readObject();
			ois3.close();
			
			//initiate features
			ObjectInputStream ois4 = new ObjectInputStream(new FileInputStream(
					savedModel.getAbsolutePath().replaceAll("\\.model$", ".feature")));
			TreeSet ts = (TreeSet) ois4.readObject();
			this.optionsModel.setFeatureSet(ts);
			ois4.close();

			//initiate nominal feature names
			ObjectInputStream ois5 = new ObjectInputStream(new FileInputStream(
					savedModel.getAbsolutePath().replaceAll("\\.model$", ".nfn")));
			ArrayList al = (ArrayList) ois5.readObject();
			this.optionsModel.setNominalFeatureNames(al);
			ois5.close();

			//initiate nominal feature value set
			ObjectInputStream ois6 = new ObjectInputStream(new FileInputStream(
					savedModel.getAbsolutePath().replaceAll("\\.model$", ".nfvs")));
			THashMap thm = (THashMap) ois6.readObject();
			this.optionsModel.setNominalFeatureValueSet(thm);
			ois6.close();

			// generating a hash set of the training set features, used to
			// check the test point features
			this.featureSet = new THashSet(this.optionsModel.getFeatureSet());
			
			//initiate user defined feature expressions
			ObjectInputStream ois7 = new ObjectInputStream(new FileInputStream(
					savedModel.getAbsolutePath().replaceAll("\\.model$", ".udf")));
			TreeMap<String, FeatureGlobalProperty> udf = (TreeMap<String, FeatureGlobalProperty>) ois7
					.readObject();
			this.udf = udf;
			ois7.close();

			// TODO must recheck this place, whether we have to set this boolean as
			// TRUE?
			this.featureSelector = new FeatureSelector(optionsModel, this.udf, true);
			this.featureSelector.setFeatureMap(this.optionsModel.getFeatureMap());
			
			// initiate dimension labels and min/max counts for line length
			boolean flag = false;
			String currentLine = "";
			BufferedReader rd = new BufferedReader(new FileReader(savedModel
					.getAbsolutePath().replaceAll("\\.model$", ".dic")));
			for (int i = 0; (currentLine = rd.readLine()) != null; i++) {
				if (flag == true) {
					this.dimLabels.addElement(currentLine.trim());
				} else if (currentLine.startsWith("###LABELS###")) {
					flag = true;
				} else if (currentLine.startsWith("max_count")) {
					int index = currentLine.indexOf("=");
					int max = Integer.parseInt(currentLine.substring(index + 1));
					this.featureSelector.setMaxCount(max);
				} else if (currentLine.startsWith("min_count")) {
					int index = currentLine.indexOf("=");
					int min = Integer.parseInt(currentLine.substring(index + 1));
					this.featureSelector.setMinCount(min);
				}
			}
		} catch (Exception ex) {
			ex.printStackTrace();
			JOptionPane.showMessageDialog(null,
					"Error on loading the saved model. Please check.", "TERMINAL ERROR",
					JOptionPane.ERROR_MESSAGE);

			System.exit(0);
		}
	}

	public EvalDataPoint getDataPointForServer(String dimValue, String text,
			ArrayList nominalFeatures) {

		String newText = HelperFunctions.replacePuncWithSpecialTokens(text, optionsModel.getLanguage());
		EvalDataPoint dp = new EvalDataPoint(dimValue, newText, featureSelector,
				nominalFeatures, this.featureSet, true,
				this.optionsModel.useAdvFeatures());

		return dp;
	}

	public int setUp(ProgressDialog pDialog, File f) {
		this.testData = f;
		Reader input = new Reader(this.testData.getAbsolutePath());
		System.out.println("reading file " + input.getFilename() + "...");

		// passes the input file and gets back an iterator of DataPoints. Also,
		// when the data point is instantiated, it's featues are added to the
		// featureSelector featureSet. In other words, the feature selector
		// feature set contains the union of features in the dataPoints.
		InputFileReader ifr = new InputFileReader();
		
		Iterator iter = null;
		try {
			iter = ifr.readFile(input, pDialog, this.featureSelector, this.optionsModel.useAdvFeatures(), this);
		} catch (Exception e) {
			e.printStackTrace();
		}
		
		if (iter == null)
			return 1;

		// adds each DataPoint's individual text (the utterance) to a text list.
		while (iter.hasNext()) {
			EvalDataPoint dataPoint = (EvalDataPoint) iter.next();
			extDataPoints.add(dataPoint);
		}

		// print out number of datapoints found in the input file
		System.out.println("found " + extDataPoints.size() + " datapoints" + "\n");

		return 0;
	}


	public void run(ProgressDialog pDialog) {

		pDialog.setIsIndeterminate(false);
		pDialog.setTitle("DIM: " + this.dimName);
		pDialog.setValue(0);
		pDialog.setMax(this.extDataPoints.size());

		HashSet testDataPointSet = new HashSet();
		Iterator extDataPointIter = extDataPoints.iterator();

		int dataPointNum = 1;

		while (extDataPointIter.hasNext()) {
			pDialog.setValue(dataPointNum);
			EvalDataPoint extDataPoint = (EvalDataPoint) extDataPointIter.next();
			dataPointNum++;
			String dimValue = extDataPoint.getValueAtDim();
			boolean tpp = extDataPoint.isTruePredictionPresent();

			if (tpp) {
				Instance instance = this.featureSelector.getInstanceFromDataPoint(
						extDataPoint, testSet);
				// System.out.println("DIM_VALUE = "+ dimValue);
				
				// we know the true class of the datapoint, we are going to
				// evaluate the model using this test data
				if (testSet.attribute(MultiDimClassifier.CLASS).type() == Attribute.NUMERIC)
					instance.setValue(testSet.attribute(MultiDimClassifier.CLASS), Double.parseDouble(dimValue));
				else
					instance.setValue(testSet.attribute(MultiDimClassifier.CLASS), dimValue);
				testSet.add(instance);
			}

			// whether or not we know the true class of the data point, we are
		  // going to predict the output using the model and print output
		  testDataPointSet.add(extDataPoint);
		}

		// check if evaluation can be done and needs to be done, do the required
		if (testSet.numInstances() > 0) {
			if (this.optionsModel.isEvaluating()) {
//				if (this.classifier.getClass().getName().equals(
//						AttributeSelectedClassifier.class.getName()))
//					this.writeARFFForDimAttSelection();//this.testData);
//				else
//					this.writeARFFForDim();//this.testData);
				String tmpHead = "*Evaluating the saved model by using the supplied test data " +
						"(only the instances with known true predictions)";
				try {
					this.exportEvaluation(tmpHead, testSet, pDialog);
				} catch (Exception ex) {
					JOptionPane.showMessageDialog(null, "Cannot evaluate "
							+ this.classifier.getClass().getName() + ": " + ex.getMessage(),
							"TERMINAL ERROR", JOptionPane.ERROR_MESSAGE);
				}
			}
			// write the ARFF file for the entire test file
			// (including both instances with/without true prediction
			this.writeARFFForDim();
		} else {
			// the options model specified that evaluation should be done, but none
			// of the test instances have true predictions
			// TODO: this seems to be a weird user flow, need to think about a good solution
//			if (this.optionsModel.isEvaluating())
//				System.out.println("WARNING: Cannot evaluate saved model as test data " +
//						"does not contain true predictions. You can safely ignore this message.");
		}
		
		// /////////////////////////
		// applying the classifier to the entire set
		// /////////////////////////
		Iterator testDataPointIter = testDataPointSet.iterator();
		while (testDataPointIter.hasNext()) {
			EvalDataPoint dataPoint = (EvalDataPoint) testDataPointIter.next();
			dataPoint = runClassifier(dataPoint);
		}

		pDialog.setLabel("Writing results to the OUTPUT folder");
		pDialog.setIsIndeterminate(true);
		String path = MultiDimClassifier.OUTPUT + "/";
		String prefix = this.optionsModel.getOutputPrefix();
		if (!prefix.equals(""))
			prefix += "_";
		path = path + prefix + "SUPP_Output_" + this.dimName + "_";
		path = path + this.testData.getName().replace('.', '#').split("#")[0]
				+ ".csv";
		this.writeOutput(new File(path), this.testData.getName().endsWith(".csv"));
		this.end(pDialog);
		return;
	}

	public EvalDataPoint runClassifier(EvalDataPoint dataPoint) {
		Instance instance = this.featureSelector.getInstanceFromDataPoint(
				dataPoint, testSet);
		instance.setDataset(testSet);
		instance.setClassMissing();

		// holds double[] of confidence values for all labels this
		// classifier is classifying into. The index with the largest
		// double will be the predicted class.
		double[] dist;
		try {
			dist = this.classifier.distributionForInstance(instance);

			if (testSet.attribute(MultiDimClassifier.CLASS).type() == Attribute.NUMERIC) {
				// for regression, the zeroth value in the array is the prediction
				dataPoint.setPredictedValueAtDim("" + dist[0]);
				dataPoint.setScoreForDim(-1);
			} else {
				// find the index with the largest double value.
				double max = (double) Integer.MIN_VALUE;
				int max_index = Integer.MIN_VALUE;
				for (int j = 0; j < dist.length; j++) {
					if (dist[j] > max) {
						max = dist[j];
						max_index = j;
					}
				}
  			String predictedLabel = this.dimLabels.elementAt(max_index).toString();
  			dataPoint.setPredictedValueAtDim(predictedLabel);
  			dataPoint.setScoreForDim(max);
			}
			//            gs.addGenericNode(new GenericNode(dataPoint, max));
		} catch (Exception ex) {
			ex.printStackTrace();
			System.err.println("Error while classifying test instance <"
					+ (dataPoint.getValueAtDim().equals(
							InputFileReader.UNKNOWN_STRING) ? "?" : dataPoint
							.getValueAtDim()) + "," + dataPoint.getText()
					+ ">: Please check.");
		}

		return dataPoint;
	}

	public void writeARFFForDimAttSelection() {//File testData) {
		// setting up training set, before writting it to the .arff file
		ArrayList selected = this
				.getSelectedFeatures((AttributeSelectedClassifier) this.classifier);
		//System.out.println("HERE----->"+selected);
		FastVector atts = new FastVector();
		Iterator featureIter = this.optionsModel.getFeatureSet().iterator();
		while (featureIter.hasNext()) {
			String tmpAtt = featureIter.next().toString();
			if (selected.contains(tmpAtt)) {
				Attribute att = new Attribute(tmpAtt);
				atts.addElement(att);
			}
		}
		if (optionsModel.isLineLength()) {
			Attribute att = new Attribute(FeatureSelector.LINE_LENGTH_FLAG);
			atts.addElement(att);
		}

		// add nominal features to the arff dataset structure
		if (EvalDataPoint.nominalFeatureNames != null) {
			for (int i = 0; i < EvalDataPoint.nominalFeatureNames.size(); i++) {
				String nominalAtt = (String) EvalDataPoint.nominalFeatureNames.get(i);
				FastVector nominalAttValues = (FastVector) EvalDataPoint.nominalFeatureValueSet
						.get(nominalAtt);
				if (nominalAttValues != null && nominalAttValues.size() > 0) {
					Attribute att = new Attribute(nominalAtt, nominalAttValues);
					atts.addElement(att);
				}
			}
		}

		// adds the class labels to the attribute set.
		atts.addElement(new Attribute(MultiDimClassifier.CLASS, this.dimLabels));

		// test set (consisting of the instances whose class is not known)
		Instances testSet = new Instances(MultiDimClassifier.TEST_SET, atts, 0);
		testSet.setClassIndex(testSet.numAttributes() - 1);

		Iterator dataPointIter = this.extDataPoints.iterator();

		int dataPointNum = 1;
		// splits data points into training/test set.
		while (dataPointIter.hasNext()) {
			EvalDataPoint dataPoint = (EvalDataPoint) dataPointIter.next();
			dataPointNum++;
			String dimValue = dataPoint.getValueAtDim();
			// System.out.println(dimValue);
			Instance instance = this.featureSelector.getInstanceFromDataPoint(
					dataPoint, testSet);
			if (!dimValue.equals(InputFileReader.UNKNOWN_STRING)) {
				if (testSet.attribute(MultiDimClassifier.CLASS).type() == Attribute.NUMERIC)
					instance.setValue(testSet.attribute(MultiDimClassifier.CLASS), Double.parseDouble(dimValue));
				else
					instance.setValue(testSet.attribute(MultiDimClassifier.CLASS), dimValue);
				testSet.add(instance);
			} else {
				instance.setValue(testSet.attribute(MultiDimClassifier.CLASS), Instance
						.missingValue());
				testSet.add(instance);

			}
		}
		//        String outputArff = testData.getAbsolutePath().replace('.', '#').split("#")[0]+".arff";
		String path = MultiDimClassifier.OUTPUT + "/";
		String prefix = this.optionsModel.getOutputPrefix();
		if (!prefix.equals(""))
			prefix += "_";
		path = path + prefix + "SUPP_Features_AttSelected_" + this.dimName + "_";
		path = path + this.testData.getName().replace('.', '#').split("#")[0]
				+ ".arff";

		Logger logger = new Logger(path);
		logger.writeLine(testSet.toString());
		logger.close();
		System.out.println("wrote " + testSet.numInstances() + " to arff file: "
				+ path);

	}

	public void writeARFFForDim() {
		// remove the data instances, just keep the ARFF structure
		this.testSet.delete();
		Iterator dataPointIter = this.extDataPoints.iterator();

		int dataPointNum = 1;
		// splits data points into training/test set.
		while (dataPointIter.hasNext()) {
			EvalDataPoint dataPoint = (EvalDataPoint) dataPointIter.next();
			dataPointNum++;
			String dimValue = dataPoint.getValueAtDim();
			// System.out.println(dimValue);
			Instance instance = this.featureSelector.getInstanceFromDataPoint(
					dataPoint, this.testSet);
			if (!dimValue.equals(InputFileReader.UNKNOWN_STRING)) {
				if (testSet.attribute(MultiDimClassifier.CLASS).type() == Attribute.NUMERIC)
					instance.setValue(testSet.attribute(MultiDimClassifier.CLASS), Double.parseDouble(dimValue));
				else
					instance.setValue(testSet.attribute(MultiDimClassifier.CLASS), dimValue);
				this.testSet.add(instance);
			} else {
				instance.setValue(this.testSet.attribute(MultiDimClassifier.CLASS),
						Instance.missingValue());
				this.testSet.add(instance);

			}
		}
		// String outputArff = testData.getAbsolutePath().replace('.', '#').split("#")[0]+".arff";
		String path = MultiDimClassifier.OUTPUT + "/";
		String prefix = this.optionsModel.getOutputPrefix();
		if (!prefix.equals(""))
			prefix += "_";
		path = path + prefix + "SUPP_Features_" + this.dimName + "_";
		path = path + this.testData.getName().replace('.', '#').split("#")[0]
				+ ".arff";

		Logger logger = new Logger(path);
		logger.writeLine(this.testSet.toString());
		logger.close();
		System.out.println("wrote " + this.testSet.numInstances()
				+ " to arff file: " + path);

	}

	private void exportEvaluation(String headline, Instances testSet,
			ProgressDialog pDialog) throws Exception {
		try {
			StringBuffer evalReport = new StringBuffer();
			String sp = System.getProperty("line.separator");
			Evaluation eTest = new Evaluation(testSet);

			evalReport.append(headline).append(sp).append(sp);
			pDialog.setLabel("Evaluating the Model (specified TEST dataset)");
			pDialog.setIsIndeterminate(true);
			eTest.evaluateModel(this.classifier, testSet);

			evalReport.append("Dataset: ").append(
					this.testData.getName().replace('.', '#').split("#")[0] + sp);
			evalReport.append("Language of the dataset: ").append(
					this.optionsModel.getLanguage()).append(sp);
			evalReport.append("Dimension: ").append(this.dimName).append(sp);
			//    		evalReport.append("Num. of Instances: ").append(testSet.numInstances()).append(sp);
			//    		evalReport.append("Num. of Attributes: ").append(testSet.numAttributes()).append(sp);
			evalReport.append("Classifier: " + this.classifier.getClass()).append(sp);
			evalReport.append("Classifier parameters: ").append(sp);
			for (int i = 0; i < this.classifier.getOptions().length; i++) {
				evalReport.append(this.classifier.getOptions()[i] + " ");
			}
			evalReport.append(sp).append(sp);
			evalReport.append(this.classifier.toString().replaceAll("\n", sp))
					.append(sp);
			evalReport.append(this.optionsModel.toString(this.featureSelector))
					.append(sp);
			evalReport.append(eTest.toSummaryString(true).replaceAll("\n", sp))
					.append(sp);
			if (!(testSet.attribute(MultiDimClassifier.CLASS).type() == Attribute.NUMERIC))
				evalReport.append(eTest.toMatrixString().replaceAll("\n", sp)).append(sp);

			String path = MultiDimClassifier.OUTPUT + "/";
			String prefix = this.optionsModel.getOutputPrefix();
			if (!prefix.equals(""))
				prefix += "_";
			path = path + prefix + "SUPP_Eval_" + this.dimName + "_";
			path = path + this.testData.getName().replace('.', '#').split("#")[0] + ".txt"; //+".txt";

			FileWriter writer = new FileWriter(path);
			writer.write(evalReport.toString());
			writer.flush();
			writer.close();
			System.out.println("wrote evaluation report to TXT file: " + path);
		} catch (Exception ex) {
			//    		ex.printStackTrace();
			throw ex;
		}

	}

	private ArrayList getSelectedFeatures(AttributeSelectedClassifier asc) {
		ArrayList al = new ArrayList();
		String content = asc.toString();
		BufferedReader rd = new BufferedReader(new StringReader(content));
		String currentLine = "";
		try {
			while ((currentLine = rd.readLine()) != null) {
				if (currentLine.startsWith("@attribute")) {
					String[] values = currentLine.split(" ");
					String att = values[1].trim();
					if (!att.equals("CLASS")) {
						al.add(att);
					}
				}
			}
		} catch (Exception ex) {
			//ex.printStackTrace();
			JOptionPane.showMessageDialog(null, "Error on processing: "
					+ ex.getMessage() + ". Please check.", "TERMINAL ERROR",
					JOptionPane.ERROR_MESSAGE);

		}
		return al;
	}

	public void writeOutput(File output, boolean csvOutput) {
  	String specific_sep = csvOutput ? "," : "&##&";
  	String escapeForDoubleQuote = csvOutput ? "\\" : "\"";
		Logger logger = new Logger(output.getAbsolutePath());
		String line = "";

		line += (this.dimName + " (actual)" + specific_sep + this.dimName + " (predicted)"
				 + specific_sep + "confidence"  + specific_sep + InputFileReader.TEXT);
		logger.writeLine(line);
		Iterator dataPointIter = this.extDataPoints.iterator();
		while (dataPointIter.hasNext()) {
			EvalDataPoint dataPoint = (EvalDataPoint) dataPointIter.next();
			line = "";
			String actualValue = (dataPoint.getValueAtDim().equals(InputFileReader.UNKNOWN_STRING) ? "?" : dataPoint.getValueAtDim());
			line += (actualValue + specific_sep + dataPoint.getPredictedValueAtDim() + 
					specific_sep
					+ (dataPoint.getScoreForDim() == -1 ? "N/A" : 
						dataPoint.getScoreForDim()) 
					+ specific_sep);
      if (csvOutput) {
        // text always goes in double quotes in the CSV
        // insert the correct escape character for the " (double quote)
        line += '"' + dataPoint.getText().replaceAll("\"", escapeForDoubleQuote) + '"';
      } else {
      	// no enclosing in double quotes for XLS
      	// no escaping the double quotes
      	line += dataPoint.getText();
      }
			logger.writeLine(line);
		}
		logger.close();

		if (!csvOutput) {
  		//convert CSV to XLS
  		Reader tmpCSV = new Reader(output.getAbsolutePath());
  		String path = InputFileReader.convertCSVToXLS(tmpCSV, specific_sep);
  		tmpCSV.close();
  		output.delete();
  		System.out.println("wrote labeling results to XLS file: " + path);
		} else {
			System.out.println("wrote labeling results to CSV file: " + output.getAbsolutePath());
		}
	}

	private void end(ProgressDialog pDialog) {
		pDialog.dispose();
		System.out.println("\nDone!");
		JOptionPane.showMessageDialog(null, "done!", "Success",
				JOptionPane.INFORMATION_MESSAGE);
		this.gui.dispose();
	}
	
	public String getDimName() {
		return this.dimName;
	}
	
	public THashSet getFeatureSet() {
		return this.featureSet;
	}

	public FastVector getDimLabels() {
		return this.dimLabels;
	}
}
