/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package edu.cmu.taghelpertools.middle_layer.extractor;

import java.util.*;

/**
 * This class servers as an "inverted index" recording the position and length for basic features associated with 
 * each datapoint. (unigrams, bigrams, POS bigrams).
 * User defined features are also stored in the class.  
 * 
 * @author haochuan
 *
 */

public class FeatureSpaceMap implements java.io.Serializable{ //inverted index
	//private TreeMap<String, PositionLengthPair> ordinaryFeatures;	// for ordinary unigrams, bigrams, POS bigrams, this is the inverted index
	//private TreeMap<String, String> syntheticFeatureExpressions;
	//private TreeMap<String, PositionLengthPair> syntheticFeatureValues; // should replace this line with Integer
	
	public TreeMap<String, PositionLengthPair> ordinaryFeatures;	// for ordinary unigrams, bigrams, POS bigrams, this is the inverted index
	public TreeMap<String, String> syntheticFeatureExpressions;
	public TreeMap<String, PositionLengthPair> syntheticFeatureValues; // should replace this line with Integer
	
	public static final int NOT_COUNTABLE = -99;
	
	public FeatureSpaceMap()
	{
		this.ordinaryFeatures = new TreeMap<String, PositionLengthPair>();
		this.syntheticFeatureExpressions = new TreeMap<String, String>();
		this.syntheticFeatureValues = new TreeMap<String, PositionLengthPair>();
	}
//	public FeatureSpaceMap(TreeMap<String, String> synExp, TreeMap<String, PositionLengthPair> synVal)
//	{
//		this.ordinaryFeatures = new TreeMap<String, PositionLengthPair>();
//		this.syntheticFeatureExpressions = new TreeMap<String, String>();
//		this.syntheticFeatureValues = new TreeMap<String, PositionLengthPair>();
//		
//		this.syntheticFeatureExpressions.putAll(synExp);
//		this.syntheticFeatureValues.putAll(synVal);
//	}
		
	public Iterator getOrdKeyIterator()
	{
		return this.ordinaryFeatures.keySet().iterator();
	}
	
	public Iterator getSynKeyIterator()
	{
		return this.syntheticFeatureValues.keySet().iterator();
	}
	
	public void removeSyhtheticFeature(String featureName)
	{
		//System.out.println("FSM - removeSyhtheticFeature1:  " + this.syntheticFeatureExpressions.size());
		//System.out.println("FSM - removeSyhtheticFeature1:  " + this.syntheticFeatureValues.size());
		this.syntheticFeatureExpressions.remove(featureName);
		this.syntheticFeatureValues.remove(featureName);
		//System.out.println("FSM - removeSyhtheticFeature2:  " + this.syntheticFeatureExpressions.size());
		//System.out.println("FSM - removeSyhtheticFeature2:  " + this.syntheticFeatureValues.size());
	}
	
	public void removeSyhtheticFeatureIfNotMatched(String featureName, 
												   String expression)
	{
		if(this.syntheticFeatureValues.containsKey(featureName))
		{
			if(!expression.equals(this.syntheticFeatureExpressions.get(featureName)))
			{
				this.syntheticFeatureExpressions.remove(featureName);
				this.syntheticFeatureValues.remove(featureName);
			}
		}
	}
	
	public boolean addSyntheticFeature(String featureName,
					String expression,
					int position,
					int length)
	{
		return addSyntheticFeature(featureName, expression, position, length, null);
	}
	
	public boolean addSyntheticFeature(String featureName,
									   String expression,
									   int position,
									   int length,
									   OperationList opList)
	{
		/*to be implemented*/	
		boolean result;
		if(this.syntheticFeatureValues.containsKey(featureName) && expression.equals(this.syntheticFeatureExpressions.get(featureName)))
		{
			result = true;			
			PositionLengthPair tmpPair = this.syntheticFeatureValues.get(featureName);
			tmpPair.addPair(position, length);	
			System.out.print("---> Same: "+expression);				
		}
		else if(this.syntheticFeatureValues.containsKey(featureName) && !expression.equals(this.syntheticFeatureExpressions.get(featureName)))
		{
			result = true;
			PositionLengthPair tmpPair = new PositionLengthPair(position, length, opList);
			this.syntheticFeatureExpressions.put(featureName, expression);	
			this.syntheticFeatureValues.put(featureName, tmpPair);			
			System.out.print("---> Replaced: "+expression);			
		}
		else
		{
			result = false;
			PositionLengthPair tmpPair = new PositionLengthPair(position, length, opList);
			this.syntheticFeatureExpressions.put(featureName, expression);				
			this.syntheticFeatureValues.put(featureName, tmpPair);
			System.out.print("---> New: "+expression);			
		}
		
		if(opList !=null)
			System.out.println("  "+opList.toString());
		else
			System.out.println("");
			
		return result;		
	}
	
	public boolean addOrdFeature(String featureName, 
								  int position, 
								  int length,
								  boolean isPOS) // length: unigram-1, bigram-2
	{
		boolean result;
		if(this.ordinaryFeatures.containsKey(featureName))
		{
			result = true;
			PositionLengthPair tmpPair = this.ordinaryFeatures.get(featureName);
			tmpPair.addPair(position, length);				
		}
		else
		{
			result = false;
			PositionLengthPair tmpPair = new PositionLengthPair(position, length, isPOS);
			this.ordinaryFeatures.put(featureName, tmpPair);
		}
		return result;
	}
	
	public PositionLengthPair getOrdFeature(String featureName)
	{
		return this.ordinaryFeatures.get(featureName);
	}
	
	public PositionLengthPair getSynFeature(String featureName)
	{
		return this.syntheticFeatureValues.get(featureName);
	}
	
	public OperationList getOplist(String featureName)
	{
		return this.getSynFeature(featureName).getOperationList();
	}
	
	public boolean isNonCountable(String featureName)
	{
		PositionLengthPair tmp = this.ordinaryFeatures.get(featureName);
		if(tmp!=null)
		{
			return tmp.isNonCountable();
		}
		else
		{
//			System.out.println("\n\n\n ---->here0");			
			return false;
		}
	}	
	
	public boolean isPOS(String featureName)
	{
		PositionLengthPair tmp = this.ordinaryFeatures.get(featureName);
		if(tmp!=null)
		{
			return tmp.isPOS();
		}
		else
		{
//			System.out.println("\n\n\n ---->here0");			
			return false;
		}
	}
	
	public boolean isBigram(String featureName)
	{
		PositionLengthPair tmp = this.ordinaryFeatures.get(featureName);
		if(tmp!=null)
		{
			return tmp.isBigram();
		}
		else
		{
//			System.out.println("\n\n\n ---->here1");
			return false;
		}
	}	
	
	public boolean hasOrigFeature(String featureName)
	{
		return this.ordinaryFeatures.containsKey(featureName);
	}
	
	public int getNumOccurance(String featureName)
	{
		if(this.ordinaryFeatures.containsKey(featureName))
		{
			PositionLengthPair tmpPair = this.ordinaryFeatures.get(featureName);			
			return tmpPair.getNumOccurance();
		}
		else
		{
			return -1;
		}
	}
	
	/**
	 *  A wrapper for recording position and length information associated with features 
	 * @author haochuan
	 *
	 */
	class PositionLengthPair
	{		
		private ArrayList<Integer> positionList = new ArrayList<Integer>();
		private ArrayList<Integer> lengthList = new ArrayList<Integer>();
		private boolean isPOS;
		private boolean isBigram;
		private boolean isNonCountable;
		private OperationList operationList = null; // for synthetic feature only
		
		PositionLengthPair(int position, int length, OperationList opList)
		{
			this(position, length);
			this.operationList = opList;
		}
		PositionLengthPair(int position, int length, boolean isPOS)
		{
			this.isPOS = isPOS;				
			if(position==FeatureSpaceMap.NOT_COUNTABLE || length == FeatureSpaceMap.NOT_COUNTABLE)
				this.isNonCountable = true;
			
			if(length == 2)
				this.isBigram = true;
			else
				this.isBigram = false;		
			
			this.positionList.add(new Integer(position));
			this.lengthList.add(new Integer(length));			
		}
		
		PositionLengthPair(int position, int length)
		{
			this(position, length, false);
		}		
		
		void addPair(int position, int length)
		{
			if(!this.positionList.contains(position))
			{
				this.positionList.add(new Integer(position));
				this.lengthList.add(new Integer(length));
			}
			
		}
		
		public OperationList getOperationList()
		{
			return this.operationList;
		}
		
		public boolean isNonCountable()
		{
			return this.isNonCountable;
		}
		
		public boolean isPOS()
		{
			return this.isPOS;
		}
		
		public boolean isBigram()
		{
			return this.isBigram;
		}
		
		public boolean occurAt(int position)
		{
			for(int i=0; i<this.positionList.size(); i++)
			{
				if(this.positionList.get(i).intValue() == position)
					return true;
			}
			return false;
		}
		
		public int occurAt(int position, boolean returnIndex)
		{
			for(int i=0; i<this.positionList.size(); i++)
			{
				if(this.positionList.get(i).intValue() == position)
					return i;
			}
			return -1;
		}		

		public int occurAfter(int position, boolean returnIndex)
		{
			for(int i=0; i<this.positionList.size(); i++)
			{
				if(this.positionList.get(i).intValue() >= position)
					return i;
			}
			return -1;
		}			
		
		public ArrayList<Integer> getPositionList()
		{
			return this.positionList;
		}
		
		public ArrayList<Integer> getLengthList()
		{
			return this.lengthList;
		}
		
		int getNumOccurance()
		{
			return this.positionList.size();
		}
	
	}	
}
