/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * Java command line parsercourtesy of John Lloyd
 */
package edu.cmu.taghelpertools.middle_layer.util;

import edu.cmu.taghelpertools.middle_layer.*;
import edu.cmu.taghelpertools.tokenizer.Tokenizer;

import java.util.*;

import weka.core.*;
import argparser.*;
import gnu.trove.*;
import java.io.*;

/**
 * SimpleFeatureSpaceBuilder is a simple implementation that provides 
 * methods for fulfilling some general needs for programmers to build feature space
 * by supplying a list of texts or a .dic file.
 * 
 * Example:
 * <pre><code>
 * ...
 * 
 * SimpleFeatureSpaceBuilder simpleBuilder = new SimpleFeatureSapceBuilder(); 
 * simpleBuilder.build(texts);  //texts is a list
 * Iterator itr = simpleBuilder.getBinaryFeatureIterator(text);
 *  
 * ...
 * </code></pre>
 * 
 * @author Hao-Chuan Wang
 * 
 */

public class SimpleFeatureSpaceBuilder {
	
	private OptionsModel optionsModel;
	private FeatureSelector featureSelector;
	private ArrayList allAtts;
	private ArrayList numeticAtts;
	private ArrayList dataPoints; 
	private boolean localOptions = false;

    /**
     * Instantiating a SimpleFeatureSapceBuilder object with customized feature selection options
     * 
     * @param options -- a list of feature selection options separated by spaces
     * <h3>options:</h3>
     * <pre>
     * -stop	--remove stopwords 
     * -stem	--do stemming
     * -rr		--remove rare words
     * -rt (integer value)	
     * 		--set the threshold of rare word removal to an integer value, 
     * 		specifying this option automatically turns on -rr
     * -f {uni, bi, posbi, punc, ll, cns}  
     *		--set whether to include following features: ungrams, bigrams, 
     *		pos bigrams, punctuation, line length, and containing non-stopwords
     *-lang {eng, ger, chi}
     *		--set the language of the dataset, default language setting can be specified in
     *		default_lang_setting.txt (English, Germany, or Chinese(extra module required; 
     *		the Chinese module is licensed separately) )
     *</pre>
     *<p>
     * <h3>example (remove stopwords, do stemming, set language to German, include unigram, 
     * 				bigram and pos bigram features):</h3>
     * <code>-stop -stem -lang ger -f uni -f bi -f posbi</code>
     * 
     * <p>
     * Also see edu.cmu.taghelpertools.middle_layer.Tester for a concrete example
     */	
	public SimpleFeatureSpaceBuilder(String options)
	{
  		ArgParser argparser = new ArgParser("");

		BooleanHolder stop = new BooleanHolder(false);
		BooleanHolder stem = new BooleanHolder(false);
		BooleanHolder removeRare = new BooleanHolder(false);
		IntHolder rareThresh = new IntHolder(0);
		Vector featVec = new Vector(7);
		argparser.addOption(
				"-f,--feature %s {uni,bi,posbi,punc,ll,cns} #Features to use", featVec);
		argparser.addOption(
				"-stop,--stopwords %v #Remove stopwords or functional words from list of features", stop);
		argparser.addOption(
				"-stem,--stemwords %v #Reduce words to morphological root forms", stem);
		argparser.addOption(
				"-rr,--removerare %v #Remove rare features (occurring fewer number of times than a threshold)", removeRare);
		argparser.addOption(
				"-rt,--rarethresh %d #The threshold used for removing rare features (specifying this option automatically turns on -rr)",
				rareThresh);
		StringHolder lang = new StringHolder("eng");		
		argparser.addOption(
				"-lang,--language %s {eng, ger, chi} #The language of the dataset)",
				lang);	
		
		String[] args = options.split(" ");
		String[] unmatched =
			argparser.matchAllArgs(args, 0, argparser.EXIT_ON_ERROR);
		
		this.optionsModel = new OptionsModel();
		this.optionsModel.setUnigrams(false);
		this.optionsModel.setBigrams(false);
		this.optionsModel.setPunctuation(false);
		this.optionsModel.setPosBigrams(false);
		this.optionsModel.setLineLength(false);
		this.optionsModel.setContainsNonStopword(false);		
		this.optionsModel.setEvaluating(false);
		this.optionsModel.setUsingArff(true);		
		
		for (int i = 0; i < featVec.size(); i++) {
			if (((StringHolder) featVec.get(i)).value.equalsIgnoreCase("uni")) 
			{
				this.optionsModel.setUnigrams(true);
			} 
			else if (((StringHolder) featVec.get(i)).value.equalsIgnoreCase("bi")) 
			{
				this.optionsModel.setBigrams(true);
			} 
			else if (((StringHolder) featVec.get(i)).value
					.equalsIgnoreCase("punc")) 
			{
				this.optionsModel.setPunctuation(true);
			} 
			else if (((StringHolder) featVec.get(i)).value
					.equalsIgnoreCase("posbi")) 
			{
				this.optionsModel.setPosBigrams(true);
			} 
			else if (((StringHolder) featVec.get(i)).value.equalsIgnoreCase("ll")) 
			{
				this.optionsModel.setLineLength(true);
			} 
			else if (((StringHolder) featVec.get(i)).value
					.equalsIgnoreCase("cns")) 
			{
				this.optionsModel.setContainsNonStopword(true);
			}
		}
		
		// if a rare threshold is provided, turn on remove-rare flag
		if (rareThresh.value > 0) {
			removeRare.value = true;
		}
		
		// if remove-rare flag is turned on and rare threshold is not
		// specified, set it to default value of 5
		if (removeRare.value == true && rareThresh.value == 0) {
			rareThresh.value = 5;
		}
		
		this.optionsModel.setRemoveStopwords(stop.value);
		this.optionsModel.setStem(stem.value);
		this.optionsModel.setRemoveRareFeatures(removeRare.value);
		this.optionsModel.setFrequencyThreshold(rareThresh.value);	
		
		//set up the language
		if(lang.value.equalsIgnoreCase("eng"))
		{
			this.optionsModel.setLanguage(Tokenizer.LANGS[Tokenizer.LANG_ENGLISH], 
										  Tokenizer.STPS[Tokenizer.LANG_ENGLISH]);
		}
		else if(lang.value.equalsIgnoreCase("ger"))
		{
			this.optionsModel.setLanguage(Tokenizer.LANGS[Tokenizer.LANG_GERMAN], 
					  					  Tokenizer.STPS[Tokenizer.LANG_GERMAN]);			
		}
		else if(lang.value.equalsIgnoreCase("chi"))
		{
			this.optionsModel.setLanguage(Tokenizer.LANGS[Tokenizer.LANG_CHINESE], 
					  					  Tokenizer.STPS[Tokenizer.LANG_CHINESE]);			
		}
		
		this.featureSelector = new FeatureSelector(optionsModel);
		this.localOptions = true;
	}
	
    /**
     * Instantiating a SimpleFeatureSapceBuilder object with all default options for feature selection:
     * deafult feature selection options:
     * <p>
     * <pre>
     * punctuation, 
     * ungrams, 
     * bigrams, 
     * POS bigrams, 
     * line length, 
     * contains non-stopwords, 
     * remove rare features(threshold=5), 
     * remove storwords=true, 
     * stemming=true 
     * </pre>
     */	
	public SimpleFeatureSpaceBuilder()
	{
		this.optionsModel = new OptionsModel();
		this.featureSelector = new FeatureSelector(optionsModel);		
	}
	
	/**     
     * Building an attribute space from a dic file
     * <p>
     * @param dicFilePath -- the path of the input dictionary file
     * @return all the attributes extracted from the texts
     * 			(including binary and numeric attributes)
     */		
	public ArrayList build(String dicFilePath)
	{
		File dicFile = new File(dicFilePath);
		this.featureSelector = new FeatureSelector(dicFile);
		
		if(!this.localOptions)
		{
			this.optionsModel = this.featureSelector.getOptionsModel();
		}
		else
		{
			this.featureSelector.setOptionsModel(this.optionsModel);
		}
		
		this.allAtts = new ArrayList();
        this.numeticAtts = new ArrayList();
        
        Iterator featureIter = this.featureSelector.getFeatureSetIterator();
        while (featureIter.hasNext()) 
        {
        	this.allAtts.add(featureIter.next().toString());
        }

        if (optionsModel.isLineLength()) 
        {
        	this.allAtts.add(FeatureSelector.LINE_LENGTH_FLAG);
        	this.numeticAtts.add(FeatureSelector.LINE_LENGTH_FLAG);     
        }
        
        return this.allAtts;        
	}
	
	/**     
     * Output the built attribute space to a dictionary file, which can be reloaded and reused later
     * <p>
     * @param dicFilePath -- the path of the input dictionary file
     */		
	public void writeDicFile(String dicFilePath)
	{
		this.featureSelector.writeDicFile(dicFilePath, null);
	}
	
	/**
	 * Print out the feature selection options and also return the string that has been printed
	 * <p>
	 * @return String that has been printed
	 */
	public String printoutOptions()
	{
		String output = this.optionsModel.toString();
		System.out.println(output);
		return output;
	}
		
    	
    /**
     * Building an attribute space from the list of texts where attirbutes should be extracted from
     * 
     * @param texts -- a list of string
     * @return all the attributes extracted from the texts
     * 			(including binary and numeric attributes)
     */	
	public ArrayList build(ArrayList<String> texts)
	{	
        this.dataPoints = new ArrayList();
        this.allAtts = new ArrayList();
        this.numeticAtts = new ArrayList();        
        
        Iterator textsItr = texts.iterator();
        while(textsItr.hasNext())
        {
        	DataPoint dp = new  DataPoint(textsItr.next().toString(), this.featureSelector);  
        	
        	int numTokens = Tokenizer.getInstance(featureSelector.getOptionsModel())
        							.tokenizeIncStopword(dp.getText()).getNumTokens();
        	if (numTokens > featureSelector.getMaxCount())
        	{
        		featureSelector.setMaxCount(numTokens);
        	}
        	if (numTokens < featureSelector.getMinCount()) 
        	{
        		featureSelector.setMinCount(numTokens);
        	}
        	
        	this.dataPoints.add(dp);
        }     

        if (this.optionsModel.isRemoveRareFeatures()) {
            this.featureSelector.removeRareFeatures();
        }
        
        Iterator featureIter = this.featureSelector.getFeatureSetIterator();
        while (featureIter.hasNext()) 
        {
        	this.allAtts.add(featureIter.next().toString());
//            Attribute att = new Attribute(featureIter.next().toString());
//            this.atts.addElement(att);
        }

        if (optionsModel.isLineLength()) 
        {
        	this.allAtts.add(FeatureSelector.LINE_LENGTH_FLAG);
        	this.numeticAtts.add(FeatureSelector.LINE_LENGTH_FLAG);
//            Attribute att = new Attribute(FeatureSelector.LINE_LENGTH_FLAG);
//            this.atts.addElement(att);            
        }
        
        return this.allAtts;
	}	

	
	/**
     * Retrieve the binary dimensions of the attribute space
     * 
     * @return java.uti.Iterator -- the iterator of binary attributes
     */
	public Iterator getBinaryFeatureIterator()
	{
        Iterator featureIter = this.featureSelector.getFeatureSetIterator();
        return featureIter;
	}
	
	/**
     * Retrieve the binary dimensions of the attribute space
     * 
     * @return java.uti.ArrayList -- the list of binary attributes
     */
	public ArrayList getBinaryFeatures()
	{
        Iterator featureIter = this.featureSelector.getFeatureSetIterator();
        ArrayList list = new ArrayList();
        while(featureIter.hasNext())
        {
        	list.add(featureIter.next().toString());
        }
        return list;
	}	
	
	/**     
     * Retrieve the binary attributes which are true for the given text
     * 
     * @return java.uti.Iterator -- the iterator of binary attributes
     */
	public Iterator getBinaryFeatureIterator(String text)
	{
		if(this.featureSelector == null)
			return null;
		
		TreeSet featureSet = this.optionsModel.getFeatureSet();
		THashSet tmp = new THashSet(featureSet);
		THashSet tmpFeatureSet = this.featureSelector.getTestPointFeatureSet(text, tmp);
		return tmpFeatureSet.iterator();
	}
	
	/**     
     * Retrieving a list of names of all existent numeric attributes
     * 
     * @return java.uti.Iterator -- the iterator of numeric attributes
     */
	public Iterator getNumericFeatureIterator()
	{
		return this.numeticAtts.iterator();
	}
	
	/**     
     * Retrieving a list of names of all existent numeric attributes
     * 
     * @return java.util.ArrayList -- the list of numeric attributes
     */	
	public ArrayList getNumericFeatures()
	{
		return this.numeticAtts;
	}	
	
	/**     
     * Retrieving values of numeric features for the given text. 
     * 
     * Note that currently the only possible numeric feature is line length.
     * The value of line length returned by this method is normalized against
     * the maximum and minimum values of line length observed in the corpus (if available). 
     * 
     * @return the list of values (wrapped by Double) in which the order is 
     * consistent with the provided list of feature names. 
     */	
	public ArrayList<Double> getNumericFeatureValues (String text, ArrayList featureNames)
	{
		if(this.featureSelector == null)
			return null;
				
		ArrayList<Double> result = new ArrayList<Double>();
		
		for(int i=0; i<featureNames.size(); i++)
		{
			String tmpName = (String)featureNames.get(i);
			
			if(tmpName != null)
			{
				if(tmpName.equals(FeatureSelector.LINE_LENGTH_FLAG))
				{
					double linelength= this.featureSelector.getLineLength(text);
					result.add(new Double(linelength));
				}
				else
				{
					result.add(null);
				}
			}
			else
			{
				result.add(null);
			}
		}
		
		return result;
	}
		
	
}
