/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package edu.cmu.taghelpertools.tokenizer;
import edu.cmu.taghelpertools.middle_layer.*;
import edu.cmu.taghelpertools.middle_layer.extractor.FeatureSpaceMap;

/**
 *  This is an abstraction of tokenizer that is language-independent. Each specialized version for particular languages should 
 *  extend this class and implement those abstract methods. 
 *  
 * @author Hao-Chuan Wang
 * 
 */
public abstract class AbstractTokenizer {

    public static final String BOL = "BOL";
    public static final String EOL = "EOL";
    
    public static String taggedResult=null;

    /**
     * Extracts set of unigrams, while stemming and removing stopwords.
     * 
     * @param line -
     *            input String to be tokenized.
     * @return - tokens
     */
    public abstract Tokens lemmatizeExcStopword(String line);
    public abstract Tokens lemmatizeExcStopword(String line, FeatureSpaceMap fsm);    

    /**
     * Extracts set of unigrams, while stemming. Stopwords are not removed.
     * 
     * @param line -
     *            input String to be tokenized.
     * @return - tokens
     */
    public abstract Tokens lemmatizeIncStopword(String line) ;
    public abstract Tokens lemmatizeIncStopword(String line, FeatureSpaceMap fsm) ;    

    /**
     * Extracts set of unigrams. Does not stem or remove stopwords.
     * 
     * @param line -
     *            input String to be tokenized.
     * @return - tokens
     */
    public abstract Tokens tokenizeIncStopword(String line);
    public abstract Tokens tokenizeIncStopword(String line, FeatureSpaceMap fsm);    
    /**
     * Extracts set of unigrams. No stemming is done. Stopwords are removed.
     * 
     * @param line -
     *            input String to be tokenized.
     * @return - tokens
     */
    public abstract Tokens tokenizeExcStopword(String line);
    public abstract Tokens tokenizeExcStopword(String line, FeatureSpaceMap fsm);    
    
    /**
     * Extracts bigrams from the text. Constituent unigrams are not stemmed and
     * stopwords are not removed. A <BOL> (beginning of line) and <EOL> end of
     * line flag is added to the text. Thus, a text with words A B C will return
     * four bigrams: <BOL>_A, A_B, B_C, C_<EOL>
     * 
     * @param line -
     *            input String to be tokenized.
     * @return - tokens
     */
    public abstract Tokens tokenizeBigramsIncStopwords(String line);
    public abstract Tokens tokenizeBigramsIncStopwords(String line, FeatureSpaceMap fsm);    

    /**
     * Extracts bigrams from the text. Constituent unigrams are stemmed, but
     * stopwords are not removed. A <BOL> (beginning of line) and <EOL> end of
     * line flag is added to the text. Thus, a text with words A B C will return
     * four bigrams: <BOL>_A, A_B, B_C, C_<EOL>
     * 
     * @param line -
     *            input String to be tokenized.
     * @return - tokens
     */
    public abstract Tokens lemmatizeBigramsIncStopwords(String line) ;
    public abstract Tokens lemmatizeBigramsIncStopwords(String line, FeatureSpaceMap fsm) ;

    /**
     * Extracts bigrams from the text. Constituent unigrams are not stemmed and
     * stopwords are removed. A <BOL> (beginning of line) and <EOL> (end of
     * line) flag is added to the text. Thus, a text with words A B C will
     * return four bigrams: <BOL>_A, A_B, B_C, C_<EOL>. Bigrams that include a
     * stopword as either of its two constituents are not included in Tokens.
     * 
     * @param line -
     *            input String to be tokenized.
     * @return - tokens
     */
    public abstract Tokens tokenizeBigramsExcStopwords(String line);
    public abstract Tokens tokenizeBigramsExcStopwords(String line, FeatureSpaceMap fsm);
    /**
     * Extracts bigrams from the text. Constituent unigrams are stemmed and
     * stopwords are removed. A <BOL> (beginning of line) and <EOL> end of line
     * flag is added to the text. Thus, a text with words A B C will return four
     * bigrams: <BOL>_A, A_B, B_C, C_<EOL>. Bigrams that include a stopword as
     * either of its two constituents are not included in Tokens.
     * 
     * @param line -
     *            input String to be tokenized.
     * @return - tokens
     */
    public abstract Tokens lemmatizeBigramsExcStopwords(String line);
    public abstract Tokens lemmatizeBigramsExcStopwords(String line, FeatureSpaceMap fsm);
   
    /**
     * Extracts POS bigrams from input text using the Stanford POS tagger.
     * 
     * @param line
     *            to be processed
     * @return - Tokens with POS bigrams
     */
    public abstract Tokens getPOSBigrams(String line);
    public abstract Tokens getPOSBigrams(String line, FeatureSpaceMap fsm);    

}
