/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package edu.cmu.taghelpertools.tokenizer;

import java.util.Iterator;
import java.util.Set;
import java.util.StringTokenizer;
import javax.swing.JOptionPane;
import edu.cmu.taghelpertools.middle_layer.*;
import edu.cmu.taghelpertools.middle_layer.extractor.FeatureSpaceMap;

import org.upsam.grade.util.DLLChineseSegmenter;

/**
 * Chinese normalizer for tokenizing and segmenting Chinese words.
 * It employs Academia Sinica's Chinese word segmenter.
 * 
 * @author haochuan
 *
 */
public class ChineseTokenizer extends AbstractTokenizer{
    static String STOPWORD_FILENAME = Tokenizer.STPS[Tokenizer.LANG_CHINESE];  	

    protected ChineseTokenizer() {
    } 

    public static String removeParentheses(String original)
    {
        String processed = original.replaceAll("[]\\(][@\\d\\s\\w]*[^\\)]",""); 
        return processed;
    }
    
    public static String normalizeWhiteSpace(String original)
    {
        String processed = original.replaceAll("[@\\s]+"," ");  
        return processed;
    }
    
    public static String normalizePunctuation(String original)
    {
    	String processed = original.replaceAll("A", ",")
    							   .replaceAll("C",".")
    							   .replaceAll("u", "\"")
    							   .replaceAll("v", "\"")
    							   .replaceAll("I", "!")
    							   .replaceAll("H", "?")
    							   .replaceAll("G", ":")
    							   .replaceAll("F", ";")
    							   .replaceAll("B", ".");
    	return processed;
    }

    /**
     * Extracts set of unigrams, while stemming and removing stopwords.
     * Note: Chiense version --- no stemming issue
     * 
     * @param line -
     *            input String to be tokenized.
     * @return - tokens
     */
    public Tokens lemmatizeExcStopword(String line) {
        return tokenizeExcStopword(line, null);
    }
    
    public Tokens lemmatizeExcStopword(String line, FeatureSpaceMap fsm) {
        return tokenizeExcStopword(line, fsm);
    }    

    /**
     * Extracts set of unigrams, while stemming. Stopwords are not removed.
     * 
     * @param line -
     *            input String to be tokenized.
     * @return - tokens
     */
    public Tokens lemmatizeIncStopword(String line) {
        return tokenizeIncStopword(line, null);
    }
    
    public Tokens lemmatizeIncStopword(String line, FeatureSpaceMap fsm) {
        return tokenizeIncStopword(line, fsm);
    }    

    /**
     * Extracts set of unigrams. Does not stem or remove stopwords.
     * 
     * @param line -
     *            input String to be tokenized.
     * @return - tokens
     */
    public Tokens tokenizeIncStopword(String line) {
    	return tokenizeIncStopword(line, null);
    }
    public Tokens tokenizeIncStopword(String line, FeatureSpaceMap fsm) {
        Tokens tokens = new Tokens();
        if (line == null)
            return tokens;
        else {
            if (line.trim().length() == 0) {
                return tokens;
            }
        }
//        line = line.trim();
//        String result = DLLChineseSegmenter.doChineseSeg(line);    
//        result = ChineseTokenizer.normalizeWhiteSpace(result);        
//        //no pos
//        result = ChineseTokenizer.removeParentheses(result); 
        System.out.println(/*result*/line);
        String[] split = /*result*/line.split(" ");
        for (int i = 0; i < split.length; i++) {
            String term = split[i];
            if (term.length() > 0) {
                tokens.addToken(term, term);
                if(fsm!=null)
                	fsm.addOrdFeature(term, i, 1, false);                  
            }
        }
//        System.out.println("1. ------------------->Number of tokens:"+tokens.getNumTokens());
//        try
//        {
//        System.in.read();
//        }catch(Exception e){}
        return tokens;
    }

    /**
     * Extracts set of unigrams. No stemming is done. Stopwords are removed.
     * 
     * @param line -
     *            input String to be tokenized.
     * @return - tokens
     */
    public Tokens tokenizeExcStopword(String line) { 
    	return tokenizeExcStopword(line, null);
    }
    public Tokens tokenizeExcStopword(String line, FeatureSpaceMap fsm) {       
        Tokens tokens = new Tokens();
        if (line == null)
            return tokens;
        else {
            if (line.trim().length() == 0) {
                return tokens;
            }
        }
//        line = line.trim();
//        String result = DLLChineseSegmenter.doChineseSeg(line);    
//        result = ChineseTokenizer.normalizeWhiteSpace(result);
//        
//        //no pos
//        result = ChineseTokenizer.removeParentheses(result);            
        System.out.println(/*result*/line);                
        String[] split = /*result*/line.split(" ");
        for (int i = 0; i < split.length; i++) {
            String term = split[i];
            if (term.length() > 0) {
            	//NOTE: Chinese-- no stemming
                if (StopwordManager.getInstance(ChineseTokenizer.STOPWORD_FILENAME).isStopword(term)) {
                    tokens.addToken(Token.EMPTY, term);
                } 
                else 
                {
                    tokens.addToken(term, term);
                    if(fsm!=null)
                    	fsm.addOrdFeature(term, i, 1, false);                      
                }
            }
        }
//        System.out.println("2. ------------------->Number of tokens:"+tokens.getNumTokens()); 
//        try
//        {
//        System.in.read();
//        }catch(Exception e){}
        return tokens;        
    }

    /**
     * Extracts bigrams from the text. Constituent unigrams are not stemmed and
     * stopwords are not removed. A <BOL> (beginning of line) and <EOL> end of
     * line flag is added to the text. Thus, a text with words A B C will return
     * four bigrams: <BOL>_A, A_B, B_C, C_<EOL>
     * 
     * @param line -
     *            input String to be tokenized.
     * @return - tokens
     */
    public Tokens tokenizeBigramsIncStopwords(String line) {
    	return tokenizeBigramsIncStopwords(line, null);
    }
    public Tokens tokenizeBigramsIncStopwords(String line, FeatureSpaceMap fsm) {
        Tokens tokens = this.tokenizeIncStopword(line);
        Iterator tIter = tokens.getIterator();
        Tokens bigramTokens = new Tokens();
        String lastUni = ChineseTokenizer.BOL;
        int count =0;
        while (tIter.hasNext()) {
            Token token = (Token) tIter.next();
            String term = token.getIndexToken();
            String bigram = lastUni + "_" + term;
            bigramTokens.addToken(bigram, bigram);
            if(fsm!=null)
            	fsm.addOrdFeature(bigram, count, 2, false);             
            lastUni = term;
            if (!tIter.hasNext()) {
                bigram = lastUni + "_" + ChineseTokenizer.EOL;
                bigramTokens.addToken(bigram, bigram);
                if(fsm!=null)
                	fsm.addOrdFeature(bigram, count, 2, false);  
            }
            count++;
        }
        return bigramTokens;
    }

    /**
     * Extracts bigrams from the text. Constituent unigrams are stemmed, but
     * stopwords are not removed. A <BOL> (beginning of line) and <EOL> end of
     * line flag is added to the text. Thus, a text with words A B C will return
     * four bigrams: <BOL>_A, A_B, B_C, C_<EOL>
     * 
     * @param line -
     *            input String to be tokenized.
     * @return - tokens
     */
    public Tokens lemmatizeBigramsIncStopwords(String line) {
    	return lemmatizeBigramsIncStopwords(line, null);
    }
    public Tokens lemmatizeBigramsIncStopwords(String line, FeatureSpaceMap fsm) {
        Tokens tokens = this.lemmatizeIncStopword(line);
        Iterator tIter = tokens.getIterator();
        Tokens bigramTokens = new Tokens();
        String lastUni = ChineseTokenizer.BOL;
        int count=0;
        while (tIter.hasNext()) {
            Token token = (Token) tIter.next();
            String term = token.getIndexToken();
            String bigram = lastUni + "_" + term;
            bigramTokens.addToken(bigram, bigram);
            if(fsm!=null)
            	fsm.addOrdFeature(bigram, count, 2, false);            
            lastUni = term;
            if (!tIter.hasNext()) {
                bigram = lastUni + "_" + ChineseTokenizer.EOL;
                bigramTokens.addToken(bigram, bigram);
                if(fsm!=null)
                	fsm.addOrdFeature(bigram, count, 2, false);                
            }
            count++;
        }
        return bigramTokens;
    }

    /**
     * Extracts bigrams from the text. Constituent unigrams are not stemmed and
     * stopwords are removed. A <BOL> (beginning of line) and <EOL> (end of
     * line) flag is added to the text. Thus, a text with words A B C will
     * return four bigrams: <BOL>_A, A_B, B_C, C_<EOL>. Bigrams that include a
     * stopword as either of its two constituents are not included in Tokens.
     * 
     * @param line -
     *            input String to be tokenized.
     * @return - tokens
     */
    public Tokens tokenizeBigramsExcStopwords(String line) {
    	return tokenizeBigramsExcStopwords(line, null);
    }
    public Tokens tokenizeBigramsExcStopwords(String line, FeatureSpaceMap fsm) {
        Tokens tokens = this.tokenizeExcStopword(line);
        Iterator tIter = tokens.getIterator();
        Tokens bigramTokens = new Tokens();
        String lastUni = ChineseTokenizer.BOL;
        int count =0;
        while (tIter.hasNext()) {
            Token token = (Token) tIter.next();
            String term = token.getIndexToken();
            if (lastUni.length() == 0) {
                if (!term.equals(Token.EMPTY)) {
                    lastUni = term;
                }
            } 
            else 
            {
                if (!term.equals(Token.EMPTY)) {
                    String bigram = lastUni + "_" + term;
                    bigramTokens.addToken(bigram, bigram);
                    if(fsm!=null)
                    	fsm.addOrdFeature(bigram, count, 2, false);                        
                    lastUni = term;
                } else {
                    lastUni = "";
                }
            }
            count++;
        }
        return bigramTokens;
    }

    /**
     * Extracts bigrams from the text. Constituent unigrams are stemmed and
     * stopwords are removed. A <BOL> (beginning of line) and <EOL> end of line
     * flag is added to the text. Thus, a text with words A B C will return four
     * bigrams: <BOL>_A, A_B, B_C, C_<EOL>. Bigrams that include a stopword as
     * either of its two constituents are not included in Tokens.
     * 
     * @param line -
     *            input String to be tokenized.
     * @return - tokens
     */
    public Tokens lemmatizeBigramsExcStopwords(String line) {
    	return  lemmatizeBigramsExcStopwords(line, null);
    }
    public Tokens lemmatizeBigramsExcStopwords(String line, FeatureSpaceMap fsm) {
        Tokens tokens = this.lemmatizeExcStopword(line);
        Iterator tIter = tokens.getIterator();
        Tokens bigramTokens = new Tokens();
        String lastUni = ChineseTokenizer.BOL;
        int count = 0;
        while (tIter.hasNext()) {
            Token token = (Token) tIter.next();
            String term = token.getIndexToken();
            if (lastUni.length() == 0) {
                if (!term.equals(Token.EMPTY)) {
                    lastUni = term;
                }
            } else {
                if (!term.equals(Token.EMPTY)) {
                    String bigram = lastUni + "_" + term;
                    bigramTokens.addToken(bigram, bigram);
                    if(fsm!=null)
                    	fsm.addOrdFeature(bigram, count, 2, false);                      
                    lastUni = term;
                } else {
                    lastUni = "";
                }
            }
            count++;
        }
        return bigramTokens;
    }

    public Tokens getPOSBigrams(String line) {
    	return getPOSBigrams(line, null);
    }
    public Tokens getPOSBigrams(String line, FeatureSpaceMap fsm) {
    	Tokens tokens = new Tokens();
    	if (line == null)
    		return tokens;
    	else {
    		if (line.trim().length() == 0) {
    			return tokens;
    		}
    	}
    	line = line.trim();
    	String taggedString = DLLChineseSegmenter.doChineseSeg(line);     
    	taggedString = ChineseTokenizer.normalizeWhiteSpace(taggedString);
//        System.out.println(taggedString);
    	String[] split = taggedString.split(" ");
    	String prevPos = ChineseTokenizer.BOL;
    	
    	for (int i = 0; i < split.length; i++) {
    		String pos = split[i].split("\\(")[1].replace(')',' ').trim();
    		String bigram = prevPos + "_" + pos;
    		tokens.addToken(bigram, bigram);
            if(fsm!=null)
            	fsm.addOrdFeature(bigram, i, 2, true);      		
    		prevPos = pos;
    		if (i == split.length - 1) {
    			bigram = prevPos + "_" + ChineseTokenizer.EOL;
    			tokens.addToken(bigram, bigram);
                if(fsm!=null)
                	fsm.addOrdFeature(bigram, i, 2, true);      			
    		}
    	}

    	return tokens;
    } 
    
    /**
     * Generating regularized strings by putting a space between each string token
     * @param tokens
     * @param startIndex
     * @param endIndex
     * @return -Regularized String
     */
    private static String getRegularizedString(String[] tokens, int startIndex, int endIndex)
    {
    	StringBuffer sb = new StringBuffer();
    	
    	for(int i=0; i<tokens.length; i++)
    	{
    		if(i>= startIndex && i<=endIndex)
    			sb.append(tokens[i]).append(" ");
    	}
    	
    	return sb.toString();
    }

    public static void main(String args[]) {

        String test = " >>(2) Do women have souls in Islam? >> >>People have said here that some Muslims say that women do not have >>souls   I must admit I have never heard of such a view being held by >>Muslims of any era   I have heard of some Christians of some eras >>holding this viewpoint  but not Muslims   Are you sure you might not be >>confusing Christian history with Islamic history? >> >  >Yes  it is supposed to have been a predominant view in the Turkish >Caliphate  >   I am not aware of any  Turkish Caliphate  viewpoint on this  Can you reference?  However  I found a quote due to Imam Ali  whom the Shias follow:   Men  never obey your women in any way whatsoever  Never let them give their advice on any matter whatsoever  even those of everyday life  Indeed  allow them freely to give advice on anything and they will fritter away one s wealth and disobey the wishes of the owner of this wealth    We see them without religion  when  alone  they are left to their own devices  they are lacking in both pity and virtue when their carnal desires are at stake  It is easy to enjoy them  but they cause great anxiety  The most virtious among them are libertines  But the most corrupt are whores  Only those of them whom age has deprived of any charm are untainted by vice  They have three qualities particular to miscreants  they complain of being oppressed  whereas it is they who oppress  they make oaths  whereas they are lying  they pretend to refuse men s solicitations  whereas they desire them most ardently  Let us beg the help of God to emerge victorious from their evil deeds  And preserve us in any case from their good ones    (Quote from Mas ud al-Qanawi  ref  A  Bouhdiba  Sexuality in Islam   p  118)   I wouldn t consider this quote as being exemplary of the Islamic (TM) viewpoint though   For all we know  the prophet s cousin and the Fourth Khalif Hazret-i Ali may have said this after a frustrating  night with a woman   Selim Guncer  -- Selim E  Guncer               | Jaca negra  luna grande  CSSER-ASU                     | y aceitunas en mi alforja  (602)-965-4096                | Aunque sepa los caminos guncer@enuxha eas asu edu     | yo nunca llegare a Cordoba   (FGL)";
        System.out.println(test.split("\\W+").length);
        Iterator iter = Tokenizer.getInstance("English").getPOSBigrams(test).getIterator();
        while (iter.hasNext()) {
            Token tNode = (Token) iter.next();
            System.out.println(tNode.getIndexToken());
        }


    }
}
