/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package edu.cmu.taghelpertools.tokenizer;

import java.util.regex.*;

public class GermanStemmer{
    // German-specific constants
    public static final String vowels = "aeiouy";

    private static final Pattern p = Pattern.compile(".*?[" + vowels + "][^" + vowels + "]");
    private int r1_begin, r1_length, r2_begin, r2_length;

    // implementation of Snowball's German stemming algorithm, under BSD license
    // (http://snowball.tartarus.org/algorithms/german/stemmer.html)

    public String stem(String term){
	term.toLowerCase();
	//  -> ss
	term.replace("", "ss");
	// vuv -> vUv, vyv -> vYv
	String matchVowel = "([" + vowels + "])";
	term.replaceAll(matchVowel + "u$1", "$1U$1");
        term.replaceAll(matchVowel + "y$1", "$1Y$1");
	// find R1 and R2 bounds
	r1_begin = 0;
	r2_begin = 0;
	Matcher m = p.matcher(term);
	if(m.lookingAt()){
	    r1_begin = m.end();
	    m = p.matcher(term.substring(r1_begin));
	    if(m.lookingAt()){
		r2_begin = m.end() + r1_begin;
	    }
	}
	r1_length = term.length() - r1_begin;
	r2_length = term.length() - r2_begin;
	// NOT IMPLEMENTED: region before R1 has at least 3 letters

	term = suffixOne(term);
	term = suffixTwo(term);
	term = stepThree(term);

	/*  turn U and Y back into lower case, and remove the umlaut accent from a, o and u. */
	term.toLowerCase();
	term.replace('', 'a');
	term.replace('', 'o');
	term.replace('', 'u');

//	System.out.println("Stemming--->"+term);
	return term;
    }

    private static final String[] suffixes_1_a = {"ern", "em", "en", "er", "es"};
    private static final String[] suffixes_1_b = {"bs", "ds", "fs", "gs", "hs", "ks", "ls", "ms", "ns", "rs", "ts"};

    private String suffixOne(String term){
	/* Search for the longest among the following suffixes,
           (a) e   em   en   ern   er   es
           (b) s (preceded by a valid s-ending) 
           and delete if in R1. */
	for(int i = 0; i < suffixes_1_a.length; i++){
	    if(suffixes_1_a[i].length() <= r1_length){
		if(term.endsWith(suffixes_1_a[i])){
		    r1_length -= suffixes_1_a[i].length();
		    r2_length -= suffixes_1_a[i].length();
		    return term.substring(0, term.length() - suffixes_1_a[i].length());
		}
	    }
	}
	for(int i = 0; i < suffixes_1_b.length; i++){
	    if(suffixes_1_b[i].length() <= r1_length + 1){
		if(term.endsWith(suffixes_1_b[i])){
		    r1_length -= suffixes_1_b[i].length();
		    r2_length -= suffixes_1_b[i].length();
		    return term.substring(0, term.length() - 1);
		}
	    }
	}
	// and e
	if(term.endsWith("e")){
	    r1_length -= 1;
	    r2_length -= 1;
	    return term.substring(0, term.length() - 1);
	}
	return term;
    }

    private static final String[] suffixes_2_a = {"est", "en", "er"};
    private static final String[] suffixes_2_b = {"bst", "dst", "fst", "gst", "hst", "kst", 
						  "lst", "mst", "nst", "tst"};

    private String suffixTwo(String term){
	/* Search for the longest among the following suffixes,
	   (a) en   er   est
	   (b) st (preceded by a valid st-ending, itself preceded by at least 3 letters) 
	   and delete if in R1. */
	for(int i = 0; i < suffixes_2_a.length; i++){
	    if(suffixes_2_a[i].length() <= r1_length){
		if(term.endsWith(suffixes_2_a[i])){
		    r1_length -= suffixes_2_a[i].length();
		    r2_length -= suffixes_2_a[i].length();
		    return term.substring(0, term.length() - suffixes_2_a[i].length());
		}
	    }
	}
	for(int i = 0; i < suffixes_2_b.length; i++){
	    if(suffixes_2_b[i].length() <= r1_length + 1){
		if(term.endsWith(suffixes_2_b[i]) && (term.length() > 5)){
		    r1_length -= suffixes_2_b[i].length();
		    r2_length -= suffixes_2_b[i].length();
		    return term.substring(0, term.length() - 2);
		}
	    }
	}
	return term;
    }

    private static final String[] suffixes_3 = {"keit", "lich", "heit", "isch",
						"end", "ung",
						"ig", "ik"};

    private String stepThree(String term){
	// apply only first found (this only works because no ending is a substring of another)
	for(int i = 0; i < suffixes_3.length; i++){
	    if(term.endsWith(suffixes_3[i])){
		if(suffixes_3[i].equals("end") ||
		   suffixes_3[i].equals("ung")){
		    /* end   ung
		       delete if in R2 
		       if preceded by ig, delete if in R2 and not preceded by e */
		    if(suffixes_3[i].length() <= r2_length){
			r1_length -= suffixes_3[i].length();
			r2_length -= suffixes_3[i].length();
			term = term.substring(0, term.length() - suffixes_3[i].length());
			if((2 < r2_length) && term.endsWith("ig") && !term.endsWith("eig")){
			    r1_length -= 2;
			    r2_length -= 2;
			    term = term.substring(0, term.length() - 2);
			}
		    }
		    return term;
		}else if(suffixes_3[i].equals("ig") ||
			 suffixes_3[i].equals("ik") ||
			 suffixes_3[i].equals("isch")){
		    /* ig   ik   isch
		       delete if in R2 and not preceded by e */
		    if((suffixes_3[i].length() <= r2_length) &&
		       !term.endsWith("e" + suffixes_3[i])){
			r1_length -= suffixes_3[i].length();
			r2_length -= suffixes_3[i].length();
			term = term.substring(0, term.length() - suffixes_3[i].length());
		    }
		    return term;
		}else if(suffixes_3[i].equals("lich") ||
			 suffixes_3[i].equals("heit")){
		    /* lich   heit
		       delete if in R2 
		       if preceded by er or en, delete if in R1 */
		    if(suffixes_3[i].length() <= r2_length){
			r1_length -= suffixes_3[i].length();
			r2_length -= suffixes_3[i].length();
			term = term.substring(0, term.length() - suffixes_3[i].length());
			if((2 <= r1_length) && (term.endsWith("er") || term.endsWith("en"))){
			    r1_length -= 2;
			    r2_length -= 2;
			    term = term.substring(0, term.length() - 2);
			}
		    }
		    return term;
		}else if(suffixes_3[i].equals("keit")){
		    /* keit
		       delete if in R2 
		       if preceded by lich or ig, delete if in R2 */
		    if(4 <= r2_length){
			r1_length -= 4;
			r2_length -= 4;
			term = term.substring(0, term.length() - 4);
			if((2 < r2_length) && term.endsWith("ig")){
			    r1_length -= 2;
			    r2_length -= 2;
			    term = term.substring(0, term.length() - 2);
			}else if((4 < r2_length) && term.endsWith("lich")){
			    r1_length -= 4;
			    r2_length -= 4;
			    term = term.substring(0, term.length() - 4);
			}
		    }
		    return term;
		}else{
		    System.out.println("SERIOUS PROBLEM IN STEMMER.");
		}
	    }
	}
	return term;
    }
}