/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package edu.cmu.taghelpertools.tokenizer;

import java.io.*;
import java.util.*;
import javax.swing.JOptionPane;
import edu.cmu.taghelpertools.middle_layer.*;
import edu.cmu.taghelpertools.middle_layer.extractor.FeatureSpaceMap;
import edu.cmu.taghelpertools.util.HelperFunctions;
 

public class GermanTokenizer extends AbstractTokenizer{
    private static final String BOL = "BOL";
    private static final String EOL = "EOL";
    static String STOPWORD_FILENAME = Tokenizer.STPS[Tokenizer.LANG_GERMAN]; 
    private static final boolean DEBUG = false;
    
    private static GermanPOSTagger tagger;
    static {
        try {
            tagger = new GermanPOSTagger("etc/de-tiger-pos", null);
        } catch (Exception e) {
            JOptionPane.showMessageDialog(null, "Could not find model file in location " + '"' + "<ROOT>/de-tiger-pos" + '"', "ERROR", JOptionPane.ERROR_MESSAGE);
            System.exit(0);
        }
    }

    protected GermanTokenizer() {
    }

    // ehuber: German tokenize
    public LinkedList germanTokenize(String line) {
        // first break up by whitespace. we don't care to keep whitespace as
        // tokens.
        LinkedList l = new LinkedList();
        StringTokenizer st = new StringTokenizer(line);
        while (st.hasMoreTokens()) {
            // then break up by things we want to keep.
            String a = st.nextToken();
            // System.out.println("before inner: " + a);
            StringTokenizer stinner = new StringTokenizer(a, ".,;:\"'()!?/\\-&$%", true);
            while (stinner.hasMoreTokens()) {
                l.add(stinner.nextToken());
            }
        }
        return l;
    }

    // ehuber: canonize, tokenize, lemmatize, remove stopwords    
    public Tokens lemmatizeExcStopword(String line) 
    {
    	return this.lemmatizeExcStopword(line, null);
    }
    
    public Tokens lemmatizeExcStopword(String line, FeatureSpaceMap fsm) {
        // sanity check here
        Tokens tokens = new Tokens();
        if (line == null)
            return tokens;
        else {
            if (line.trim().length() == 0) {
                return tokens;
            }
        }
        line = line.trim().toLowerCase();
        // canonize
        // for (int i = 0; i < line.length() - 1; i++) {
        // System.out.println("canonizer loop: " + line.substring(i, i + 2) + "
        // (" + (int) line.charAt(i) + ", " + (int) line.charAt(i + 1) + ")");
        // }
        /*
         * if(line.indexOf("ö") == -1){ System.out.println("HEY!!! this has NO
         * oe: " + line); }
         */
        // tokenize, reject if stopword. otherwise, stem.
        LinkedList l = germanTokenize(line);
        GermanStemmer stemmer = new GermanStemmer();
        int count = 0; //for FSM        
        while (!l.isEmpty()) {
            String term = this.trimNonLetterChars((String) l.removeFirst());
            if (term.length() > 0) {
                boolean isPreStemmingStop = false;
                if (StopwordManager.getInstance(GermanTokenizer.STOPWORD_FILENAME).isStopword(term)) {
                    isPreStemmingStop = true;
                }
                String stem = stemmer.stem(term);
                if (StopwordManager.getInstance(GermanTokenizer.STOPWORD_FILENAME).isStopword(stem) || isPreStemmingStop) {
                    tokens.addToken(Token.EMPTY, term);
                } 
                else {
                    tokens.addToken(stem, term);
                    if(fsm!=null)
                    	fsm.addOrdFeature(stem, count, 1, false);
                }
            }
            count++;
        }
        return tokens;
    }

    // ehuber: canonize, tokenize, lemmatize, leave in stopwords
    public Tokens lemmatizeIncStopword(String line) {
    	return lemmatizeIncStopword(line, null);
    }    
    
    public Tokens lemmatizeIncStopword(String line, FeatureSpaceMap fsm) {
        // canonization step here
        Tokens tokens = new Tokens();
        if (line == null)
            return tokens;
        else {
            if (line.trim().length() == 0) {
                return tokens;
            }
        }
        line = line.trim().toLowerCase();
        GermanStemmer stemmer = new GermanStemmer();
        StringTokenizer st = new StringTokenizer(line);
        int count = 0;
        while (st.hasMoreTokens()) {
            String term = this.trimNonLetterChars(st.nextToken()).trim();
            if (term.length() > 0) {
                String stem = stemmer.stem(term);
                tokens.addToken(stem, term);
                if(fsm!=null)
                	fsm.addOrdFeature(stem, count, 1, false);                
            }
            count++;
        }
        return tokens;
    }

    // ehuber: canonize, tokenize, leave in stopwords
    public Tokens tokenizeIncStopword(String line) {
    	return tokenizeIncStopword(line, null);
    }
    
    public Tokens tokenizeIncStopword(String line, FeatureSpaceMap fsm) {
        // canonization step here
        Tokens tokens = new Tokens();
        if (line == null)
            return tokens;
        else {
            if (line.trim().length() == 0) {
                return tokens;
            }
        }
        line = line.trim().toLowerCase();
        StringTokenizer st = new StringTokenizer(line);
        int count = 0;
        while (st.hasMoreTokens()) {
            String term = this.trimNonLetterChars(st.nextToken()).trim();
            if (term.length() > 0) {
                tokens.addToken(term, term);
                if(fsm!=null)
                	fsm.addOrdFeature(term, count, 1, false);                 
            }
            count++;
        }
        return tokens;
    }

    // ehuber: canonize, tokenize, remove stopwords
    public Tokens tokenizeExcStopword(String line) {
    	return tokenizeExcStopword(line, null);
    }
    
    public Tokens tokenizeExcStopword(String line, FeatureSpaceMap fsm) {
        // canonization step here
        Tokens tokens = new Tokens();
        if (line == null)
            return tokens;
        else {
            if (line.trim().length() == 0) {
                return tokens;
            }
        }
        line = line.trim().toLowerCase();
        StringTokenizer st = new StringTokenizer(line);
        int count = 0;
        while (st.hasMoreTokens()) {
            String term = this.trimNonLetterChars(st.nextToken()).trim();
            if (term.length() > 0) {
                if (!StopwordManager.getInstance(GermanTokenizer.STOPWORD_FILENAME).isStopword(term)) {
                    tokens.addToken(term, term);
                    if(fsm!=null)
                    	fsm.addOrdFeature(term, count, 1, false);                     
                } else {
                    tokens.addToken(Token.EMPTY, term);
                }
            }
            count++;
        }
        return tokens;
    }

    // ehuber: canonize, tokenize, leave in stopwords -> bigrams
    public Tokens tokenizeBigramsIncStopwords(String line) {
    	return tokenizeBigramsIncStopwords(line, null);
    }
    
    public Tokens tokenizeBigramsIncStopwords(String line, FeatureSpaceMap fsm) {
        // canonization step here
        Tokens tokens = this.tokenizeIncStopword(line);
        Iterator tIter = tokens.getIterator();
        Tokens bigramTokens = new Tokens();
        String lastUni = GermanTokenizer.BOL;
        int count = 0;
        while (tIter.hasNext()) {
            Token token = (Token) tIter.next();
            String term = token.getIndexToken();
            String bigram = lastUni + "_" + term;
            bigramTokens.addToken(bigram, bigram);
            if(fsm!=null)
            	fsm.addOrdFeature(bigram, count, 2, false);             
            lastUni = term;
            if (!tIter.hasNext()) {
                bigram = lastUni + "_" + GermanTokenizer.EOL;
                bigramTokens.addToken(bigram, bigram);
                if(fsm!=null)
                	fsm.addOrdFeature(bigram, count, 2, false);                 
            }
            count++;
        }
        return bigramTokens;
    }

    //TODO
    // ehuber: canonize, tokenize, lemmatize, leave in stopwords -> bigrams
    public Tokens lemmatizeBigramsIncStopwords(String line) {
    	return lemmatizeBigramsIncStopwords(line, null);
    }
    
    public Tokens lemmatizeBigramsIncStopwords(String line, FeatureSpaceMap fsm) {
        Tokens tokens = this.lemmatizeIncStopword(line);
        Iterator tIter = tokens.getIterator();
        Tokens bigramTokens = new Tokens();
        String lastUni = GermanTokenizer.BOL;
        int count = 0;
        while (tIter.hasNext()) {
            Token token = (Token) tIter.next();
            String term = token.getIndexToken();
            String bigram = lastUni + "_" + term;
            bigramTokens.addToken(bigram, bigram);
            if(fsm!=null)
            	fsm.addOrdFeature(bigram, count, 2, false);               
            lastUni = term;
            if (!tIter.hasNext()) {
                bigram = lastUni + "_" + GermanTokenizer.EOL;
                bigramTokens.addToken(bigram, bigram);
                if(fsm!=null)
                	fsm.addOrdFeature(bigram, count, 2,false);                   
            }
            count++;
        }
        return bigramTokens;
    }

    // ehuber: canonize, tokenize, remove stopwords -> bigrams
    public Tokens tokenizeBigramsExcStopwords(String line) {
    	return tokenizeBigramsExcStopwords(line, null);
    }
    
    public Tokens tokenizeBigramsExcStopwords(String line, FeatureSpaceMap fsm) {
        Tokens tokens = this.tokenizeExcStopword(line);
        Iterator tIter = tokens.getIterator();
        Tokens bigramTokens = new Tokens();
        String lastUni = GermanTokenizer.BOL;
        int count =0;
        while (tIter.hasNext()) {
            Token token = (Token) tIter.next();
            String term = token.getIndexToken();
            if (lastUni.length() == 0) {
                if (!term.equals(Token.EMPTY)) {
                    lastUni = term;
                }
            } else {
                if (!term.equals(Token.EMPTY)) {
                    String bigram = lastUni + "_" + term;
                    bigramTokens.addToken(bigram, bigram);
                    if(fsm!=null)
                    	fsm.addOrdFeature(bigram, count, 2, false);                       
                    lastUni = term;
                } else {
                    lastUni = "";
                }
            }
            count++;
        }
        return bigramTokens;
    }

    // ehuber: canonize, tokenize, lemmatize, remove stopwords -> bigrams
    public Tokens lemmatizeBigramsExcStopwords(String line) {
    	return lemmatizeBigramsExcStopwords(line, null);
    }
    
    public Tokens lemmatizeBigramsExcStopwords(String line, FeatureSpaceMap fsm) {
        Tokens tokens = this.lemmatizeExcStopword(line);
        Iterator tIter = tokens.getIterator();
        Tokens bigramTokens = new Tokens();
        String lastUni = GermanTokenizer.BOL;
        int count = 0;
        while (tIter.hasNext()) {
            Token token = (Token) tIter.next();
            String term = token.getIndexToken();
            if (lastUni.length() == 0) {
                if (!term.equals(Token.EMPTY)) {
                    lastUni = term;
                }
            } else {
                if (!term.equals(Token.EMPTY)) {
                    String bigram = lastUni + "_" + term;
                    bigramTokens.addToken(bigram, bigram);
                    if(fsm!=null)
                    	fsm.addOrdFeature(bigram, count, 2, false);                      
                    lastUni = term;
                } else {
                    lastUni = "";
                }
            }
            count++;
        }
        return bigramTokens;
    }

    public String trimNonLetterChars(String term) {
        return term;
    }

    // ehuber: tokenize, canonize, get pos tags -> bigrams
    public Tokens getPOSBigrams(String line) {
    	return getPOSBigrams(line, null);
    }
    
    public Tokens getPOSBigrams(String line, FeatureSpaceMap fsm) {
    	//Hao-Chuan: a logger for debugging purpose
    	StringBuffer logger = new StringBuffer();
    	String sp = System.getProperty("line.separator");
    	
        // ehuber: tokenize & join
        String linecopy = line;
        line = "";
        for (Iterator tok = germanTokenize(linecopy).iterator(); tok.hasNext();)
            line += tok.next() + " ";
        // ehuber: canonize

        // ehuber: carry out rest of operation
        Tokens tokens = new Tokens();
        String taggedString = "";
        try {
            // dont print out these debug strings...
//            System.out.println("line after tokenization = " + line);
             logger.append("Line after tokenization = " + line+sp);
             taggedString = tagger.tag(line);
             AbstractTokenizer.taggedResult = taggedString;
            // System.out.println("line after tagging = " + taggedString);
             logger.append("Line after tagging = " + taggedString+sp);
        } 
        catch (Exception e) {
            int result = JOptionPane.showConfirmDialog(null, "Could not parse line: " + line + ". Would you like to terminate?");
            if (result == 0) {
                System.exit(0);
            }
        }
        String[] split = taggedString.split(" ");

	        String prevPos = GermanTokenizer.BOL;
	        int count=0;
	        for (int i = 0; i < split.length; i++) {
	        	String pos=null;	        	
	        	try{
	        		pos = split[i].split("\\/")[1];
	        	}
	        	catch(Exception ex)
	        	{
	        		break;
	        	}
	        	
	            String bigram = prevPos + "_" + pos;
	            tokens.addToken(bigram, bigram);
	            if(fsm!=null)
	            	fsm.addOrdFeature(bigram, count, 2, true);               
	        	  logger.append("Bigram added: "+bigram+sp);            
	            prevPos = pos;
	            if (i == split.length - 1) {
	                bigram = prevPos + "_" + GermanTokenizer.EOL;
	                tokens.addToken(bigram, bigram);
	                if(fsm!=null)
	                	fsm.addOrdFeature(bigram, count, 2, true);                   
	          	  	logger.append("Bigram added: "+bigram+sp);                
	            }
	            count++;
	        }
       
        //output the logger
	   if(DEBUG)
	   {
	        try{
	            String path = "POS_bigram_log.txt";
	            FileWriter writer  = new FileWriter(path);
	            writer.write(logger.toString());
	            writer.flush();
	            writer.close();
		    }
		    catch(Exception ex)
		    {
		            ex.printStackTrace();
		    }
	   }
        //
        return tokens;
    }

    public static void main(String args[]) {

        String test = "Jenny, �ckern, Rohit, Emil, Jaime, Moses, and Pam, " + " " + " " + "We need to find a way to integrate trained classifiers with TuTalk.  I'd " + "like to find the most painless way of doing this possible.  I'd like to " + "see this happen by the summerschool if at all possible. " + "I think Jaime is going to be able to set up the TagHelper training setup " + "to generate everything needed to configure the TagHelper server to " + "classify spans of text using the trained classifiers.  If that can be " + "accomplished, then I think from the TuTalk authoring environment, we " + "should be able to read in that config file so that authors have the choice " + "either to link in a trained classifier by name from a menu. The tricky " + "part is integrating the trained classifiers into the TuTalk backend. " + " " + "What makes this complicated is that I don't think it will ever be the case " + "that an author would choose to completely swap out the current TuTalk "
                + "pattern matching �ckern �ckern �ckernin favor of a classifier since it would be overkill for " + "the majority of input processing.  It is only necessary for certain " + "questions. " + " " + "So I think what would be easiest, if it is possible, would be to insert a " + "layer in between the interface and the TuTalk backend so that if the " + "current goal is one where we are using a classifier, we run the classifier " + "over the input and then insert the classification result into the " + "text that we pass back to the TuTalk backend.  What's tricky is that the " + "TuTalk student interface would need to know which classifiers if any are " + "relevant for the current question.  We would also have a concept defined " + "for the classifier with the set of possible classification results.  That " + "way, the regular TuTalk pattern matcher could make a decision based on " + "which code is matched. " + " " + "[Moses, this is where you come in - Rohit, we also need you for this] "
                + "I think the easiest solution for integration with the TuTalk backend would " + "be to insert a field in the initiation where we can insert the name of a " + "classifier.  If that initiation is presented to the student, the backend " + "should pass that classifier name to the student interface.  Then the " + "student interface knows that question will use that classifier.  It should " + "be possible to list more than one.  If we simply insert the classification " + "reaults into the text, leaving the original text in tact, then we can " + "combine classification and pattern matching for the same answer, which I " + "think would be the best combination. " + " " + "I'd like to have this working for the summerschool because it sounds like " + "people who take the TuTalk classes will also take the TagHelper classes - " + "so then it would be good if the two were formally integrated by the " + "summerschool.  Let me know what you all think. " + " " + " " + "Carolyn ";


        Iterator iter = Tokenizer.getInstance(Tokenizer.LANGS[Tokenizer.LANG_GERMAN]).lemmatizeBigramsIncStopwords(test).getIterator();
        while (iter.hasNext()) {
            Token token = (Token) iter.next();
            if (!token.getIndexToken().equals(Token.EMPTY)) {
                System.out.println(token);
            }
        }
    }
}
