/**
 * Created on 02/09/2006
 */
package edu.cmu.taghelpertools.tokenizer;

/**
 * Improvement on the implementation of the Snowball's Spanish stemming algorithm, under BSD license
 * (http://snowball.tartarus.org/algorithms/spanish/stemmer.html)
 * @author Ismael Pascual Nieto
 * @author
 */
public class SpanishStemmer {
    private static final String vowels = "aeiou";
    
    private static final String[] PRONOUN_ATTACHED={"me","se","sela","selo","selas","selos","la","le","lo","las","les","los","nos"};
    private static final String[] BEF_PRONOUN_ATTACHED={"indo","ndo","r","r","r","iendo","ando","ar","er","ir","yendo"};
    private static final String[] LOG_SUFFIXES= {"loga","logas","logia","logias"};
    private static final String[] ENTE_SUFFIXES= {"encia","encias"};
    private static final String[] IDAD_SUFFIXES= {"idad", "idades"};
    private static final String[] IVA_SUFFIXES=  {"iva","ivo","ivas","ivos"};
    private static final String[] Y_VERB_SUFFIXES= {"ya","ye","yan","yen","yeron","yendo","yo","y","yas","yes","yais","yamos"};
    private static final String[] E_VERB_SUFFIXES= {"en","es","is","eis","emos"};
    private static final String[] A_VERB_SUFFIXES= {"aran","aras","arn","ars","arais","ara","aris","aramos","aremos","ar","ar","eran","eras","ern","ers","erais","era","eris","eramos","eremos","er","er","iran","iras","irn","irs","irais","ira","iris","iramos","iremos","ir","ir","aba","ada","ida","a","ara","iera","ad","ed","id","ase","iese","aste","iste","an","aban","an","aran","ieran","asen","iesen","aron","ieron","ado","ido","ando","iendo","i","ar","er","ir","abas","adas","idas","as","aras","ieras","ases","ieses","s","is","abais","ais","arais","ierais","aseis","ieseis","asteis","isteis","ados","idos","bamos","amos","imos","ramos","iramos","isemos","semos","amos","as"};
    private static final String[] RESIDUAL_SUFFIXES={"os","a","o","","",""};
    private static final String[] E_RESIDUAL_SUFFIXES={"e",""};
    private static final String[] FIRST_SUFFIXES={"ero", "era", "anza","anzas","ico","ica","icos","icas","ismo","ismos","able","ables","ible","ibles","ista","istas","oso","osa","osos","osas","amiento","amientos","imiento","imientos", "simo", "simo", "simos", "simas"};
    
    private String regionR1, regionR2, regionRV;
    
	/**
	 * 
	 */
	public SpanishStemmer() {
	}

	/**
	 * Returns the stem of a Spanish word.
	 * @param word the word to get the stem
	 * @return the stem of the word
	 */
	public String stem(String word){
		// Prepare the process
		word=word.toLowerCase();
		getRegions(word);
		
		// Process the word
		word=removeAttachedPronouns(word);
		word=removeSuffixes(word);
		word=removeResidualSuffixes(word);
		word=unescapeAccents(word);
		
		return word;
	}
	
	/**
	 * Method to remove the pronouns attached to a word
	 * @param word the word
	 * @return the word without attached pronouns
	 */
	private String removeAttachedPronouns(String word) {
		
		for (String pronoun: PRONOUN_ATTACHED) {
			// If the word ends with a pronoun
			if (regionRV.endsWith(pronoun)) {
				String RVs=regionRV.substring(0, regionRV.length()-pronoun.length());
				
				// If the pronoun is preceeded by a particular set of in-fixes...
				for (String befAtt: BEF_PRONOUN_ATTACHED) {
					if (RVs.endsWith(befAtt)) {
						if (befAtt.equals("yendo")) {
							int indx=word.indexOf(befAtt);
							if (indx>-1 && word.charAt(indx)=='u') {
								word=unescapeAccents(word.substring(0,word.length()-pronoun.length()));
								getRegions(word);
								return word;
							}
						} else {
							word=unescapeAccents(word.substring(0, word.length()-pronoun.length()));
							getRegions(word);
							return word;
						}
					}
				}
			}
		}
		
		return word;
	}
	
	/**
	 * Removes a set of common suffixes from a word
	 * @param word the word
	 * @return the word without the suffixes
	 */
	private String removeSuffixes(String word) {
		for (String suffix: FIRST_SUFFIXES) {
			if (regionR2.endsWith(suffix)) {
				word=word.substring(0, word.length()-suffix.length());
				shortenRegions(suffix.length());
				return word;
			}
		}
		
		for (String suffix: LOG_SUFFIXES) {
			if (regionR2.endsWith(suffix)) {
				// Replace suffix by "log", the first three letters of it
				word=word.substring(0, word.length()-suffix.length()+3);
				shortenRegions(suffix.length()+3);
				return word;
				
			}
		}
		
		for (String suffix: ENTE_SUFFIXES) {
			if (regionR2.endsWith(suffix)) {
				// Replace suffix by "ente"
				word=word.substring(0, word.length()-suffix.length())+"ente";
				shortenRegions(suffix.length()+3);
				return word;
				
			}
		}
		
		if (regionR2.endsWith("amente")) {
			String str=word.substring(0,word.length()-"amente".length());
			if (str.endsWith("os") || str.endsWith("ic") || str.endsWith("ad")) {
				word=str;
				shortenRegions("mente".length());
			} else if (str.endsWith("iv") && str.contains("a")) {
				word=str;
				shortenRegions("amente".length());
				return word;
			}
		}
		
		
		if (regionR2.endsWith("mente")) {
			String str=word.substring(0,word.length()-"mente".length());
			if (str.endsWith("ante") || str.endsWith("able") || str.endsWith("ible")) {
				word=str;
				shortenRegions("mente".length());
				return word;
				
			}
		}
		
		for (String suffix: IDAD_SUFFIXES) {
			if (regionR2.endsWith(suffix)) {
				String str=word.substring(0,word.length()-suffix.length());
				if (str.endsWith("abil")) {
					word=str;
					shortenRegions(suffix.length());
				} else if (str.endsWith("ic") || str.endsWith("iv") || str.endsWith("il")) {
					word=str;
					shortenRegions(suffix.length());
					return word;
					
				}
			}
		}
			
		for (String suffix: IVA_SUFFIXES) {
			if (regionR2.endsWith(suffix)) {
				String str=word.substring(0,word.length()-suffix.length());
				if (str.endsWith("at")) {
					word=str;
					shortenRegions(suffix.length());
					return word;
				}
			}
		}
		
		return removeVerbSuffixes(word);
	}
	
	/**
	 * Removes verb suffixes from a word
	 * @param word the word
	 * @return the word without verb suffixes
	 */
	private String removeVerbSuffixes(String word) {
		for (String suffix: Y_VERB_SUFFIXES) {
			if (regionRV.endsWith(suffix)) {
				String str=word.substring(0,word.length()-suffix.length());
				if (str.endsWith("u")) {
					word=str;
					shortenRegions(suffix.length());
					return word;
				}
			}
		}
		
		for (String suffix: E_VERB_SUFFIXES) {
			if (regionRV.endsWith(suffix)) {
				String str=word.substring(0,word.length()-suffix.length());
				if (str.endsWith("gu")) {
					word=word.substring(0,word.length()-suffix.length()+1);
					shortenRegions(suffix.length());
					return word;
				}
			}
		}
		
		for (String suffix: A_VERB_SUFFIXES) {
			if (regionRV.endsWith(suffix)) {
				word=word.substring(0,word.length()-suffix.length());
				shortenRegions(suffix.length());
			}
		}
		
		
		return word;
	}
	
	private String removeResidualSuffixes(String word) {
		for (String suffix: RESIDUAL_SUFFIXES) {
			if (regionRV.endsWith(suffix)) {
				word=word.substring(0,word.length()-suffix.length());
				shortenRegions(suffix.length());
				return word;
			}
		}
		
		for (String suffix: E_RESIDUAL_SUFFIXES) {
			if (regionRV.endsWith(suffix)) {
				String str=word.substring(0,word.length()-suffix.length());
				if (str.endsWith("gu")) {
					word=word.substring(0,word.length()-suffix.length()+1);
					shortenRegions(suffix.length());
					return word;
				}
			}
		}		
		return word;
	}
	
	private void shortenRegions(int number) {
		if (regionR1.length()>number) {
			regionR1=regionR1.substring(0, regionR1.length()-number);
		} else {
			regionR1="";
		}

		if (regionR2.length()>number) {
			regionR2=regionR2.substring(0, regionR2.length()-number);
		} else {
			regionR2="";
		}
		
		if (regionRV.length()>number) {
			regionRV=regionRV.substring(0, regionRV.length()-number);
		} else {
			regionRV="";
		}
		
		
	}
	
	
	/**
	 * Change the characters with tilde in a string by its version
	 * without tildes.
	 * @param str the string
	 * @return the string without tildes
	 */
	private String unescapeAccents(String str) {
		int l=str.length();
		StringBuffer sb=new StringBuffer(l);
		char c;
		for (int i=0; i<l; i++) {
			c=str.charAt(i);
			if (c=='') sb.append("a");
			else if (c=='') sb.append("e");
			else if (c=='') sb.append("i");
			else if (c=='') sb.append("o");
			else if (c=='') sb.append("u");
			else if (c=='') sb.append("u");
			else sb.append(c);
		}
		return sb.toString();
	}
	
	/**
	 * Get the Snowball regions in a word
	 * @param word
	 */
	private void getRegions(String word) {
		int i=0, l=word.length();
		// Find the first non-vowel after a vowel
		while(i<l && vowels.indexOf(word.charAt(i))<0) i++;
		while(i<l && vowels.indexOf(word.charAt(i))>-1) i++;
		
		i++; // R1 starts after the non-vowel
		if (i>=l) { 
			regionR1="";
		} else {
			regionR1=word.substring(i);
		}		
		
		// Find the first non-vowel
		i=0; l=regionR1.length();
		while(i<l && vowels.indexOf(regionR1.charAt(i))<0) i++;
		while(i<l && vowels.indexOf(regionR1.charAt(i))>-1) i++;
		
		i++; // R1 starts after the non-vowel
		if (i>=l) {
			regionR2="";
		} else {
			regionR2=regionR1.substring(i);
		}
		
		// Get region RV
		i=0; l=word.length();
		if (2<l && vowels.indexOf(word.charAt(1))<0) {
			// Find the next vowel
			i=2;
			while(i<l && vowels.indexOf(word.charAt(i))>-1) i++;
			if (i>=l) {
				regionRV="";
			} else {
				regionRV=word.substring(i);
			}
			
		} else if (2<l && vowels.indexOf(word.charAt(0))>-1 && vowels.indexOf(word.charAt(1))>-1){
			// Find the next vowel
			i=2;
			while(i<l && vowels.indexOf(word.charAt(i))<0) i++;
			if (i>=l) {
				regionRV="";
			} else {
				regionRV=word.substring(i);
			}			
		} else if (l>3) {
			regionRV=word.substring(3);
		} else {
			regionRV=word;
		}
	}
	
	public static void main (String argv[]) {
		SpanishStemmer stemmer=new SpanishStemmer();
		String stem=stemmer.stem("toro");
		System.out.println("R1: "+stemmer.regionR1);
		System.out.println("R2: "+stemmer.regionR2);
		System.out.println("RV: "+stemmer.regionRV);
		System.out.println();
		System.out.println("STEM: "+ stem);
		
	}
}
