/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package edu.cmu.taghelpertools.tokenizer;

import java.io.IOException;
import java.util.Iterator;

import javax.swing.JOptionPane;

import latoan.pos.brill.BrillTagger;
import latoan.pos.brill.tokens.POSToken;

import edu.cmu.taghelpertools.middle_layer.extractor.FeatureSpaceMap;

/**
 * <p>TagHelper Tokeniser for Spanish. 
 * <p> Created in the <a href="www.eps.uam.es">Escuela Politcnica Superior</a> of the 
 *     <a href="www.uam.es">Universidad Autnoma de Madrid</a>.
 *      
 * 
 * @author Ismael Pascual Nieto (ismael.pascual at uam.es)
 * @author Diana Rosario Prez Marn (diana.perez at uam.es)
 * 
 */
public class SpanishTokenizer extends AbstractTokenizer {
	
    static String STOPWORD_FILENAME = Tokenizer.STPS[Tokenizer.LANG_SPANISH];
    
	/** The POS Tagger */
    private static BrillTagger tagger;
    
    /** The token splitter */
	private static SpanishScanner scanner = new SpanishScanner();
	
	/** The stemmer */
	private static SpanishStemmer stemmer = new SpanishStemmer();
	
    static {
    	
        try {
            tagger = new BrillTagger(
            				"etc/spanish-pos/FIRSTGUESS",
            				"etc/spanish-pos/LEXICON",
            				"etc/spanish-pos/LEXICALRULEFILE",
            				"etc/spanish-pos/BIGRAMS",
            				"etc/spanish-pos/CONTEXTUALRULEFILE"
            				);  
        } catch (Exception e) {
        	e.printStackTrace();
            JOptionPane.showMessageDialog(null, "Could not find model files in directory " + '"' + "<ROOT>/spanish-pos" + '"', "ERROR", JOptionPane.ERROR_MESSAGE);
            System.exit(0);
        }
    }	
	
    /* 
     * Make the constructor non-visible to non-derived classes
     */
    protected SpanishTokenizer() {
    }
	

	/* (non-Javadoc)
	 * @see edu.cmu.taghelpertools.tokenizer.AbstractTokenizer#lemmatizeExcStopword(java.lang.String)
	 */
	@Override
	public Tokens lemmatizeExcStopword(String line) {
    	return this.lemmatizeExcStopword(line, null);
	}

	/* (non-Javadoc)
	 * @see edu.cmu.taghelpertools.tokenizer.AbstractTokenizer#lemmatizeExcStopword(java.lang.String, edu.cmu.taghelpertools.middle_layer.extractor.FeatureSpaceMap)
	 */
	@Override
	public Tokens lemmatizeExcStopword(String line, FeatureSpaceMap fsm) {
		Tokens tokens=new Tokens();
		try {
			scanner.reset(line);
			POSToken token;
			String term, stem;
			int count=0;
			StopwordManager StpWdManager=StopwordManager.getInstance(SpanishTokenizer.STOPWORD_FILENAME);
			
			// Get the tokens from the scanner. The punctuation token is directly marked
			// for POS in the scanner (and also the numbers).
			while ((token=scanner.nextToken())!=null) {
				term=token.getText();
				stem=stemmer.stem(term);
				
				if (StpWdManager.isStopword(term.toLowerCase()) || isPunctuation(token.getPOS()) || StpWdManager.isStopword(stem)) {
					tokens.addToken(Token.EMPTY, term);
					
				} else {
					tokens.addToken(stem, term);
					if(fsm!=null)  fsm.addOrdFeature(stem, count, 1, false);
				}
				
				count++;
			}
			
		} catch (IOException e) {
			System.err.println("Error in tokenizeIncStopword: ");
			e.printStackTrace();
		}
		return tokens;
	}

	/* (non-Javadoc)
	 * @see edu.cmu.taghelpertools.tokenizer.AbstractTokenizer#lemmatizeIncStopword(java.lang.String)
	 */
	@Override
	public Tokens lemmatizeIncStopword(String line) {
    	return lemmatizeIncStopword(line, null);
	}

	/* (non-Javadoc)
	 * @see edu.cmu.taghelpertools.tokenizer.AbstractTokenizer#lemmatizeIncStopword(java.lang.String, edu.cmu.taghelpertools.middle_layer.extractor.FeatureSpaceMap)
	 */
	@Override
	public Tokens lemmatizeIncStopword(String line, FeatureSpaceMap fsm) {
		Tokens tokens=new Tokens();
		try {
			scanner.reset(line);
			POSToken token;
			String term, stem;
			int count=0;
			
			while ((token=scanner.nextToken())!=null) {
				term=token.getText();
				stem=stemmer.stem(term);
				
				tokens.addToken(stem, term);
				
				if(fsm!=null)  fsm.addOrdFeature(stem, count, 1, false);
				
				count++;
			}
			
		} catch (IOException e) {
			System.err.println("Error in tokenizeIncStopword: ");
			e.printStackTrace();
		}
		return tokens;
	}

	/* (non-Javadoc)
	 * @see edu.cmu.taghelpertools.tokenizer.AbstractTokenizer#lemmatizeBigramsExcStopwords(java.lang.String)
	 */
	@Override
	public Tokens lemmatizeBigramsExcStopwords(String line) {
    	return lemmatizeBigramsExcStopwords(line, null);
	}

	/* (non-Javadoc)
	 * @see edu.cmu.taghelpertools.tokenizer.AbstractTokenizer#lemmatizeBigramsExcStopwords(java.lang.String, edu.cmu.taghelpertools.middle_layer.extractor.FeatureSpaceMap)
	 */
	@Override
	public Tokens lemmatizeBigramsExcStopwords(String line, FeatureSpaceMap fsm) {
        Tokens tokens = this.lemmatizeExcStopword(line);
        Iterator tIter = tokens.getIterator();
        Tokens bigramTokens = new Tokens();
        String lastUni = SpanishTokenizer.BOL;
        int count = 0;
        while (tIter.hasNext()) {
            Token token = (Token) tIter.next();
            String term = token.getIndexToken();
            if (lastUni.length() == 0) {
                if (!term.equals(Token.EMPTY)) {
                    lastUni = term;
                }
            } else {
                if (!term.equals(Token.EMPTY)) {
                    String bigram = lastUni + "_" + term;
                    bigramTokens.addToken(bigram, bigram);
                    if(fsm!=null)
                    	fsm.addOrdFeature(bigram, count, 2, false);                      
                    lastUni = term;
                } else {
                    lastUni = "";
                }
            }
            count++;
        }
        return bigramTokens;
	}

	/* (non-Javadoc)
	 * @see edu.cmu.taghelpertools.tokenizer.AbstractTokenizer#lemmatizeBigramsIncStopwords(java.lang.String)
	 */
	@Override
	public Tokens lemmatizeBigramsIncStopwords(String line) {
    	return lemmatizeBigramsIncStopwords(line, null);
	}

	/* (non-Javadoc)
	 * @see edu.cmu.taghelpertools.tokenizer.AbstractTokenizer#lemmatizeBigramsIncStopwords(java.lang.String, edu.cmu.taghelpertools.middle_layer.extractor.FeatureSpaceMap)
	 */
	@Override
	public Tokens lemmatizeBigramsIncStopwords(String line, FeatureSpaceMap fsm) {
        Tokens tokens = this.lemmatizeIncStopword(line);
        Iterator tIter = tokens.getIterator();
        Tokens bigramTokens = new Tokens();
        String lastUni = SpanishTokenizer.BOL;
        int count = 0;
        while (tIter.hasNext()) {
            Token token = (Token) tIter.next();
            String term = token.getIndexToken();
            String bigram = lastUni + "_" + term;
            bigramTokens.addToken(bigram, bigram);
            if (fsm!=null) fsm.addOrdFeature(bigram, count, 2, false);
            	
            lastUni = term;
            if (!tIter.hasNext()) {
                bigram = lastUni + "_" + SpanishTokenizer.EOL;
                bigramTokens.addToken(bigram, bigram);
                if (fsm!=null) fsm.addOrdFeature(bigram, count, 2, false);                  
            }
            count++;
        }
        return bigramTokens;
	}
	

	/* (non-Javadoc)
	 * @see edu.cmu.taghelpertools.tokenizer.AbstractTokenizer#tokenizeExcStopword(java.lang.String)
	 */
	@Override
	public Tokens tokenizeExcStopword(String line) {
		return tokenizeExcStopword(line, null);
	}

	/* (non-Javadoc)
	 * @see edu.cmu.taghelpertools.tokenizer.AbstractTokenizer#tokenizeExcStopword(java.lang.String, edu.cmu.taghelpertools.middle_layer.extractor.FeatureSpaceMap)
	 */
	@Override
	public Tokens tokenizeExcStopword(String line, FeatureSpaceMap fsm) {
		Tokens tokens=new Tokens();
		try {
			scanner.reset(line);
			POSToken token;
			String term;
			int count=0;
			StopwordManager StpWdManager=StopwordManager.getInstance(SpanishTokenizer.STOPWORD_FILENAME);
			
			// Get the tokens from the scanner. The punctuation token is directly marked
			// for POS in the scanner (and also the numbers).
			while ((token=scanner.nextToken())!=null) {
				term=token.getText();
				
				if (StpWdManager.isStopword(term.toLowerCase()) || isPunctuation(token.getPOS())) {
					tokens.addToken(Token.EMPTY, term);
				} else {
					tokens.addToken(term, term);
					if(fsm!=null)  fsm.addOrdFeature(term, count, 1, false);
				}
				
				count++;
			}
			
		} catch (IOException e) {
			System.err.println("Error in tokenizeIncStopword: ");
			e.printStackTrace();
		}
		return tokens;
	}
	
	/**
	 * Determines if a string has punctuation in its first position.
	 * @param pos the string
	 * @return true if it contains punctuation characters
	 */
	private boolean isPunctuation(String pos) {
		if (pos==null) return false;
		if (Character.isLetter(pos.charAt(0))) return false;
		return true;
	}

	/* (non-Javadoc)
	 * @see edu.cmu.taghelpertools.tokenizer.AbstractTokenizer#tokenizeIncStopword(java.lang.String)
	 */
	@Override
	public Tokens tokenizeIncStopword(String line) {
		return tokenizeIncStopword(line, null);
	}

	/* (non-Javadoc)
	 * @see edu.cmu.taghelpertools.tokenizer.AbstractTokenizer#tokenizeIncStopword(java.lang.String, edu.cmu.taghelpertools.middle_layer.extractor.FeatureSpaceMap)
	 */
	@Override
	public Tokens tokenizeIncStopword(String line, FeatureSpaceMap fsm) {
		Tokens tokens=new Tokens();
		try {
			scanner.reset(line);
			POSToken token;
			String term;
			int count=0;
			
			while ((token=scanner.nextToken())!=null) {
				term=token.getText();
				tokens.addToken(term, term);
				
				if(fsm!=null)  fsm.addOrdFeature(term, count, 1, false);
				
				count++;
			}
			
		} catch (IOException e) {
			System.err.println("Error in tokenizeIncStopword: ");
			e.printStackTrace();
		}
		return tokens;
	}

	/* (non-Javadoc)
	 * @see edu.cmu.taghelpertools.tokenizer.AbstractTokenizer#tokenizeBigramsExcStopwords(java.lang.String)
	 */
	@Override
	public Tokens tokenizeBigramsExcStopwords(String line) {
    	return tokenizeBigramsExcStopwords(line, null);
	}

	/* (non-Javadoc)
	 * @see edu.cmu.taghelpertools.tokenizer.AbstractTokenizer#tokenizeBigramsExcStopwords(java.lang.String, edu.cmu.taghelpertools.middle_layer.extractor.FeatureSpaceMap)
	 */
	@Override
	public Tokens tokenizeBigramsExcStopwords(String line, FeatureSpaceMap fsm) {
        Tokens tokens = this.tokenizeExcStopword(line);
        Iterator tIter = tokens.getIterator();
        Tokens bigramTokens = new Tokens();
        String lastUni = SpanishTokenizer.BOL;
        int count =0;
        while (tIter.hasNext()) {
            Token token = (Token) tIter.next();
            String term = token.getIndexToken();
            if (lastUni.length() == 0) {
                if (!term.equals(Token.EMPTY)) {
                    lastUni = term;
                }
            } else {
                if (!term.equals(Token.EMPTY)) {
                    String bigram = lastUni + "_" + term;
                    bigramTokens.addToken(bigram, bigram);
                    if (fsm!=null) fsm.addOrdFeature(bigram, count, 2, false);                       
                    lastUni = term;
                } else {
                    lastUni = "";
                }
            }
            count++;
        }
        return bigramTokens;
	}

	/* (non-Javadoc)
	 * @see edu.cmu.taghelpertools.tokenizer.AbstractTokenizer#tokenizeBigramsIncStopwords(java.lang.String)
	 */
	@Override
	public Tokens tokenizeBigramsIncStopwords(String line) {
    	return tokenizeBigramsIncStopwords(line, null);
    }

	/* (non-Javadoc)
	 * @see edu.cmu.taghelpertools.tokenizer.AbstractTokenizer#tokenizeBigramsIncStopwords(java.lang.String, edu.cmu.taghelpertools.middle_layer.extractor.FeatureSpaceMap)
	 */
	@Override
	public Tokens tokenizeBigramsIncStopwords(String line, FeatureSpaceMap fsm) {
        // canonization step here
        Tokens tokens = this.tokenizeIncStopword(line);
        Iterator tIter = tokens.getIterator();
        Tokens bigramTokens = new Tokens();
        String lastUni = SpanishTokenizer.BOL;
        int count = 0;
        while (tIter.hasNext()) {
            Token token = (Token) tIter.next();
            String term = token.getIndexToken();
            String bigram = lastUni + "_" + term;
            bigramTokens.addToken(bigram, bigram);
            if (fsm!=null) fsm.addOrdFeature(bigram, count, 2, false);             
            lastUni = term;
            if (!tIter.hasNext()) {
                bigram = lastUni + "_" + SpanishTokenizer.EOL;
                bigramTokens.addToken(bigram, bigram);
                if(fsm!=null) fsm.addOrdFeature(bigram, count, 2, false);                 
            }
            count++;
        }
        return bigramTokens;
	}
	
	
	/* (non-Javadoc)
	 * @see edu.cmu.taghelpertools.tokenizer.AbstractTokenizer#getPOSBigrams(java.lang.String)
	 */
	@Override
	public Tokens getPOSBigrams(String line) {
    	return getPOSBigrams(line, null);
	}

	/* (non-Javadoc)
	 * @see edu.cmu.taghelpertools.tokenizer.AbstractTokenizer#getPOSBigrams(java.lang.String, edu.cmu.taghelpertools.middle_layer.extractor.FeatureSpaceMap)
	 */
	@Override
	public Tokens getPOSBigrams(String line, FeatureSpaceMap fsm) {
		Tokens tokens=new Tokens();
		
		try {
			scanner.reset(line);
			
			java.util.List<POSToken> sentence;
			String prevPos=SpanishTokenizer.BOL, pos;
			String bigram;
			int count=0;
			while ((sentence=scanner.nextSentenceFragment(60))!=null) {
				// Tag the sentence with POS tags
				tagger.tag(sentence);
				
				for (POSToken token: sentence) {
					pos=token.getPOS();
					
					bigram=replacePunctuationBySpecials(prevPos)+"_"+replacePunctuationBySpecials(pos);
					
					tokens.addToken(bigram, bigram);
                    if (fsm!=null) fsm.addOrdFeature(bigram, count, 2, true); 

                    count++;
                    
					// The previous POS tag is now the current
					prevPos=pos;
				}
			}
			
			// Add the last token
			pos=EnglishTokenizer.EOL;
			bigram=replacePunctuationBySpecials(prevPos)+"_"+replacePunctuationBySpecials(pos);
			tokens.addToken(bigram, bigram);
            if (fsm!=null) fsm.addOrdFeature(bigram, count, 2, true); 
			
		} catch (IOException e) {
			System.err.println("Error in tokenizeIncStopword: ");
			e.printStackTrace();
			
            int result = JOptionPane.showConfirmDialog(null, "Could not parse line: " + line + ". Would you like to terminate?");
            if (result == 0) {
                System.exit(0);
            }			
			
		}
		
		return tokens;
	}
	
	/**
	 * Method to replace the punctuation characters in the bigrams by other
	 * special string sequences.
	 * @param text
	 * @return
	 */
	private String replacePunctuationBySpecials(String text) {
		text = text.replaceAll("<", "<lessthan>");
		text = text.replaceAll(">","<greaterthan>");
		text = text.replaceAll(",", "<comma>");
		text = text.replaceAll("\\.", "<period>");
		text = text.replaceAll(";", "<semicolon>");
		text = text.replaceAll(":", "<colon>");
		text = text.replaceAll("-", "<dash>");
		text = text.replaceAll("\\?", "<questionmark>");
		text = text.replaceAll("!", "<exclamationmark>");
		text = text.replaceAll("\\*", "<asterisk>");
		text = text.replaceAll("&", "<ampersand>");
		text = text.replaceAll("@", "<atsign>");
		text = text.replaceAll("#", "<numbersign>");
		text = text.replaceAll("%", "<percentsign>");
		text = text.replaceAll("\\(", "<openparen>");
		text = text.replaceAll("\\)", "<closeparen>");
		text = text.replaceAll("\\{", "<opencurly>");
		text = text.replaceAll("\\}", "<closecurly>");
		text = text.replaceAll("\"", "<doublequote>");
		text = text.replaceAll("\\$", "<dollar>");
		text = text.replaceAll("\\\\", "<backslash>");
		text = text.replaceAll("\\/", "<forwardslash>");
		text = text.replaceAll("\\+", "<plussign>");
		text = text.replaceAll("\\~", "<tilde>");
		text = text.replaceAll("\\^", "<carat>");
		text = text.replaceAll("\\[", "<leftsquareparen>");
		text = text.replaceAll("\\]", "<rightsquareparen>");
		text = text.replaceAll("_", "<underline>");
		
		return text;
	}

	public static void main (String argv[]) {
		SpanishTokenizer toker=new SpanishTokenizer();
		String line="La nia come patatas.";
		Tokens tokens=toker.getPOSBigrams(line);
		
		Iterator it=tokens.getIterator();
		while(it.hasNext()){
			Token tok=(Token)it.next();
			System.out.println(tok.getSurfaceToken()+"\t\t"+tok.getIndexToken());
		}
		
	}
}
