/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


package edu.cmu.taghelpertools.util;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.TreeMap;

/**
 * A generic data structure that can be used to keep a set of arbirary objects
 * (GenericNodes) sorted. GenericNodes hold an arbitrary object and a value
 * double. In a GenericNodeSorter, the GenericNodes are maintained in descending
 * order of value. Two elements with the same value can co-exist (i.e., the one
 * added later does not replace the one added previously.
 * 
 * @author Jaime Arguello
 */
public class GenericNodeSorter {

    private TreeMap treeMap;
    private static GenericNodeSorter instance = new GenericNodeSorter();


    private GenericNodeSorter() {
        treeMap = new TreeMap(new GenericNodeComparator());
    }

    public static GenericNodeSorter getNewInstance() {
        return new GenericNodeSorter();
    }


    public static GenericNodeSorter getInstance() {
        instance.clear();
        return instance;
    }

    public int getSize() {
        return treeMap.size();
    }


    public void clear() {
        treeMap.clear();
    }

    public void addGenericNode(GenericNode gNode) {
        treeMap.put(gNode, null);
    }


    public Iterator getGenericNodes() {
        return treeMap.keySet().iterator();
    }


    public GenericNode getTopGenericNode() {
        return (GenericNode) treeMap.firstKey();
    }


    public GenericNode getButtomGenericNode() {
        return (GenericNode) treeMap.lastKey();
    }


    public GenericNode[] getTopGenericNodes(int max) {
        GenericNode[] nodes;
        if (max < treeMap.size())
            nodes = new GenericNode[max];
        else
            nodes = new GenericNode[treeMap.size()];
        Iterator i = treeMap.keySet().iterator();
        int x = 0;
        while (i.hasNext() && x < max) {
            nodes[x] = (GenericNode) i.next();
            x++;
        }
        return nodes;
    }


    public GenericNode[] getAllGenericNodes() {
        GenericNode[] nodes = new GenericNode[treeMap.size()];
        Iterator i = treeMap.keySet().iterator();
        int index = 0;
        while (i.hasNext()) {
            nodes[index] = (GenericNode) i.next();
            index++;
        }
        return nodes;
    }

    public GenericNode[] getAllGenericNodesReversed() {
        GenericNode[] nodes = new GenericNode[treeMap.size()];
        Object[] objs = treeMap.keySet().toArray();
        int index = 0;
        for (int i = objs.length - 1; i >= 0; i--) {
            nodes[index] = (GenericNode) objs[i];
            index++;
        }
        return nodes;
    }


    public boolean replaceGenericNode(double weight, GenericNode gNode) {
        GenericNode[] gNodes = this.getAllGenericNodes();
        treeMap.clear();
        boolean found = false;
        for (int i = 0; i < gNodes.length; i++) {
            if (gNodes[i].getWeight() == weight) {
                if (found) {
                    System.out.println("more than one segment with the same id!");
                    System.exit(0);
                }
                this.addGenericNode(gNode);
                found = true;
            } else {
                this.addGenericNode(gNodes[i]);
            }
        }
        return found;
    }

    public String[] getTopTerms(int max) {
        String[] terms;
        if (max < treeMap.size())
            terms = new String[max];
        else
            terms = new String[treeMap.size()];
        Iterator i = treeMap.keySet().iterator();
        int x = 0;
        while (i.hasNext() && x < max) {
            terms[x] = ((GenericNode) i.next()).getO().toString();
            x++;
        }
        return terms;
    }

    public String toString() {
        String result = "";
        GenericNode[] gNodes = this.getAllGenericNodes();
        for (int i = 0; i < gNodes.length; i++) {
            GenericNode gNode = gNodes[i];
            result += (gNode.getO() + "=" + gNode.getWeight() + "\n");
        }
        return result;
    }


    public static void main(String ags[]) {
        GenericNodeSorter gNodeSorter = GenericNodeSorter.getInstance();
        gNodeSorter.addGenericNode(new GenericNode(new String("1"), 2.0));
        gNodeSorter.addGenericNode(new GenericNode(new String("2"), 3.0));
        gNodeSorter.addGenericNode(new GenericNode(new String("3"), 4.0));
        gNodeSorter.addGenericNode(new GenericNode(new String("4"), 5.0));
        gNodeSorter.addGenericNode(new GenericNode(new String("5"), 6.0));
        gNodeSorter.addGenericNode(new GenericNode(new String("6"), 7.0));
        gNodeSorter.addGenericNode(new GenericNode(new String("7"), 10.0));

        gNodeSorter.replaceGenericNode(6, new GenericNode("10", 6.0));

        GenericNode[] gNodes = gNodeSorter.getAllGenericNodes();
        for (int i = 0; i < gNodes.length; i++)
            System.out.println(gNodes[i]);

    }
}
